//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include "effect.h"
#include "parser.h"
#include "log.h"

using namespace nsVDShaderParser;

void VDShaderDisassembleFunction(IVDShaderLogOutput& out, const VDShaderFunction& func, const VDShaderNameTable *nameTable) {
	out.WriteLogOutput("//\n");
	out.WriteLogOutput("// Generated by VirtualDub Shader Compiler\n");
	out.WriteLogOutput("//\n");
	out.WriteLogOutput("// Registers:\n");
	out.WriteLogOutput("//\n");
	out.WriteLogOutput("//   Name              Reg   Size\n");
	out.WriteLogOutput("//   ----------------- ----- ----\n");

	for(VDShaderFunction::Parameters::const_iterator it(func.mParameters.begin()), itEnd(func.mParameters.end()); it!=itEnd; ++it) {
		const VDShaderVariableDesc& var = *it;
		const char *name = nameTable ? nameTable->GetName(var.mNameToken) : "<unknown>";

		switch(var.mRegister & kVDShaderRegTypeMask) {
		case kVDShaderRegV0:
			VDShaderWriteLogOutputF(out, "//   %-17s v%-4d %d\n", name, var.mRegister & kVDShaderRegIndexMask, kRowCountForType[var.mType]);
			break;
		case kVDShaderRegS0:
			VDShaderWriteLogOutputF(out, "//   %-17s s%-4d %d\n", name, var.mRegister & kVDShaderRegIndexMask, 1);
			break;
		default:
			VDShaderWriteLogOutputF(out, "//   %-17s <unknown>\n", name);
			break;
		}
	}

	int samplerIdx = 0;
	for(VDShaderFunction::SamplerBindings::const_iterator it(func.mSamplerBindings.begin()), itEnd(func.mSamplerBindings.end()); it != itEnd; ++it, ++samplerIdx) {
		const VDShaderSamplerBinding& binding = *it;
		const char *name = nameTable ? nameTable->GetName(binding.mNameToken) : "<unknown>";

		VDShaderWriteLogOutputF(out, "//   %-17s s%-4d %d\n", name, samplerIdx, 1);
	}

	int numConstantBindings = func.mConstantBindings.size();
	for(int i = 0; i < numConstantBindings; ++i) {
		const VDShaderConstantBinding& binding = func.mConstantBindings[i];
		int nameToken = binding.mNameToken;

		const char *name = nameTable ? nameTable->GetName(nameToken) : "<unknown>";
		VDShaderWriteLogOutputF(out, "//   %-17s c%-4d %d\n", name, binding.mRegister & kVDShaderRegIndexMask, binding.mSize);
	}

	out.WriteLogOutput("\n");

	static const char *const kInsnName[]={
		"nop",
		"abs",
		"add",
		"cmp",
		"crs",
		"dp2add",
		"dp3",
		"dp4",
		"exp",
		"frc",
		"log",
		"lrp",
		"m3x2",
		"m3x3",
		"m3x4",
		"m4x3",
		"m4x4",
		"mad",
		"max",
		"min",
		"mov",
		"mul",
		"nrm",
		"pow",
		"rcp",
		"rsq",
		"sincos",
		"sub",
		"texkill",
		"texld",
		"texldb",
		"texldp"
	};

	struct local {
		static void print_reg(IVDShaderLogOutput& out, int reg) {
			switch(reg & kVDShaderRegTypeMask) {
				case kVDShaderRegR0:
					VDShaderWriteLogOutputF(out, "r%d", reg & kVDShaderRegIndexMask);
					break;
				case kVDShaderRegC0:
					VDShaderWriteLogOutputF(out, "c%d", reg & kVDShaderRegIndexMask);
					break;
				case kVDShaderRegO0:
					VDShaderWriteLogOutputF(out, "o%d", reg & kVDShaderRegIndexMask);
					break;
				case kVDShaderRegS0:
					VDShaderWriteLogOutputF(out, "s%d", reg & kVDShaderRegIndexMask);
					break;
				case kVDShaderRegV0:
					VDShaderWriteLogOutputF(out, "v%d", reg & kVDShaderRegIndexMask);
					break;
				default:
					out.WriteLogOutput("<invalid>");
					break;
			}
		}
	};

	int index = 0;
	for(VDShaderParserFunction::Constants::const_iterator it(func.mConstants.begin()), itEnd(func.mConstants.end()); it!=itEnd; ++it, ++index) {
		const VDShaderConstant& c = *it;

		VDShaderWriteLogOutputF(out, "    def c%d, %g, %g, %g, %g\n", index, c.v[0], c.v[1], c.v[2], c.v[3]);
	}

	for(VDShaderParserFunction::Insns::const_iterator it(func.mInsns.begin()), itEnd(func.mInsns.end()); it!=itEnd; ++it) {
		const VDShaderInsn& insn = *it;

		VDShaderWriteLogOutputF(out, "    %s%s ", kInsnName[insn.mOp], insn.mModifiers & kVDShaderInsnMod_Saturate ? "_sat" : "");

		local::print_reg(out, insn.mDestRegister);

		if (insn.mWriteMask != 0xF) {
			out.WriteLogOutput(".");
			if (insn.mWriteMask == 0x0)
				out.WriteLogOutput("-");
			else {
				if (insn.mWriteMask & 1)
					out.WriteLogOutput("x");
				if (insn.mWriteMask & 2)
					out.WriteLogOutput("y");
				if (insn.mWriteMask & 4)
					out.WriteLogOutput("z");
				if (insn.mWriteMask & 8)
					out.WriteLogOutput("w");
			}
		}

		for(int i=0; i<4; ++i) {
			const VDShaderInsnSource& src = insn.mSources[i];

			if (!src.mRegister)
				break;

			out.WriteLogOutput(", ");

			if (src.mModifiers & kVDShaderMod_Negate)
				out.WriteLogOutput("-");

			local::print_reg(out, src.mRegister);

			static const char kSwizzleChars[]={'x', 'y', 'z', 'w'};

			unsigned char srcmask = (kInsnComponentMasks[insn.mOp] >> (4 * i)) & 15;

			switch(insn.mOp) {
				case kVDShaderOp_dp2add:
				case kVDShaderOp_dp3:
				case kVDShaderOp_dp4:
				case kVDShaderOp_texld:
				case kVDShaderOp_rcp:
				case kVDShaderOp_rsq:
				case kVDShaderOp_frc:
				case kVDShaderOp_log:
				case kVDShaderOp_exp:
					break;
				default:
					srcmask &= insn.mWriteMask;
			}

			srcmask = (unsigned char)(((((unsigned)srcmask * 0x08040201) & 0x40100401) * 0x03030303) >> 24);

			if (!((src.mSwizzle ^ 0x00) & srcmask))
				out.WriteLogOutput(".x");
			else if (!((src.mSwizzle ^ 0x55) & srcmask))
				out.WriteLogOutput(".y");
			else if (!((src.mSwizzle ^ 0xAA) & srcmask))
				out.WriteLogOutput(".z");
			else if (!((src.mSwizzle ^ 0xFF) & srcmask))
				out.WriteLogOutput(".w");
			else if (((src.mSwizzle ^ 0xE4) & srcmask)) {
				out.WriteLogOutput(".");
				VDShaderWriteLogOutputF(out, "%c", kSwizzleChars[src.mSwizzle&3]);
				if (srcmask >= 0x4) {
					VDShaderWriteLogOutputF(out, "%c", kSwizzleChars[(src.mSwizzle>>2)&3]);
					if (srcmask >= 0x10) {
						VDShaderWriteLogOutputF(out, "%c", kSwizzleChars[(src.mSwizzle>>4)&3]);
						if (srcmask >= 0x40)
							VDShaderWriteLogOutputF(out, "%c", kSwizzleChars[(src.mSwizzle>>6)&3]);
					}
				}
			}
		}

		out.WriteLogOutput("\n");
	}

	out.WriteLogOutput("\n");

	VDShaderWriteLogOutputF(out, "// approximately %d instruction slots used\n", func.mInsns.size());
}

void VDShaderDisassembleEffect(IVDShaderLogOutput& out, const VDShaderEffect& eff) {
	const VDShaderNameTable& nameTable = eff.NameTable();

	out.WriteLogOutput("// variables\n\n");
	int varCount = eff.GetVariableCount();
	for(int i=0; i<varCount; ++i) {
		const VDShaderVariable& var = *eff.GetVariableByIndex(i);
		const char *type = VDShaderGetTypeName(var.mDesc.mType);
		const char *name = nameTable.GetName(var.mDesc.mNameToken);
		const char *isExtern = var.mDesc.mbExtern ? "extern " : "";
		const char *isStatic = var.mDesc.mbStatic ? "static " : "";

		if (var.mDesc.mSemanticToken) {
			const char *sema = nameTable.GetName(var.mDesc.mSemanticToken);
			VDShaderWriteLogOutputF(out, "%s%s%s %s : %s", isStatic, isExtern, type, name, sema);
		} else {
			VDShaderWriteLogOutputF(out, "%s%s%s %s", isStatic, isExtern, type, name);
		}

		const VDShaderAnnotations& annos = var.mDesc.mAnnotations;
		if (annos.mCount) {
			out.WriteLogOutput(" <\n");

			for(int i=0; i<annos.mCount; ++i) {
				const VDShaderVariable& anno = *eff.GetAnnotationByIndex(annos.mStart + i);
				const char *name = nameTable.GetName(anno.mDesc.mNameToken);
				const char *type = VDShaderGetTypeName(anno.mDesc.mType);

				VDShaderWriteLogOutputF(out, "\t%s %s", type, name);

				switch(anno.mDesc.mType) {
					case kTypeBool:
						VDShaderWriteLogOutputF(out, " = %s;\n", eff.GetBool(&anno) ? "true" : "false");
						break;
					case kTypeFloat:
						VDShaderWriteLogOutputF(out, " = %g;\n", eff.GetFloat(&anno));
						break;
					case kTypeString:
						VDShaderWriteLogOutputF(out, " = \"%s\";\n", eff.GetString(&anno));
						break;
					default:
						out.WriteLogOutput(";\n");
				}
			}

			out.WriteLogOutput(">");
		}


		out.WriteLogOutput(";\n");
	}

#if 0
	out.WriteLogOutput("\n// functions\n\n");

	int functionCount = eff.GetFunctionCount();
	for(int i=0; i<functionCount; ++i) {
		const VDShaderFunction& func = *eff.GetFunctionByIndex(i);

		VDShaderDisassembleFunction(out, func, nameTable);
	}
#endif

	int techniqueCount = eff.GetTechniqueCount();
	for(int i=0; i<techniqueCount; ++i) {
		const VDShaderTechnique& tech = *eff.GetTechniqueByIndex(i);

		if (int techName = tech.GetNameToken())
			VDShaderWriteLogOutputF(out, "\ntechnique %s", nameTable.GetName(techName));
		else
			out.WriteLogOutput("\ntechnique ");

		const VDShaderAnnotations& annos = tech.GetAnnotations();
		if (annos.mCount) {
			out.WriteLogOutput("<\n");

			for(int i=0; i<annos.mCount; ++i) {
				const VDShaderVariable& anno = *eff.GetAnnotationByIndex(annos.mStart + i);
				const char *name = nameTable.GetName(anno.mDesc.mNameToken);
				const char *type = VDShaderGetTypeName(anno.mDesc.mType);

				VDShaderWriteLogOutputF(out, "\t%s %s", type, name);

				switch(anno.mDesc.mType) {
					case kTypeBool:
						VDShaderWriteLogOutputF(out, " = %s;\n", eff.GetBool(&anno) ? "true" : "false");
						break;
					case kTypeFloat:
						VDShaderWriteLogOutputF(out, " = %g;\n", eff.GetFloat(&anno));
						break;
					case kTypeString:
						VDShaderWriteLogOutputF(out, " = \"%s\";\n", eff.GetString(&anno));
						break;
					default:
						out.WriteLogOutput(";\n");
				}
			}

			out.WriteLogOutput("> ");
		}

		out.WriteLogOutput("{\n");

		int passCount = tech.GetPassCount();
		for(int j=0; j<passCount; ++j) {
			const VDShaderPass& pass = *tech.GetPassByIndex(j);

			if (int passName = pass.GetNameToken())
				VDShaderWriteLogOutputF(out, "\tpass %s ", nameTable.GetName(passName));
			else
				out.WriteLogOutput("\tpass ");

			const VDShaderAnnotations& annos = pass.GetAnnotations();
			if (annos.mCount) {
				out.WriteLogOutput("<\n");

				for(int i=0; i<annos.mCount; ++i) {
					const VDShaderVariable& anno = *eff.GetAnnotationByIndex(annos.mStart + i);

					const char *name = nameTable.GetName(anno.mDesc.mNameToken);
					const char *type = VDShaderGetTypeName(anno.mDesc.mType);

					VDShaderWriteLogOutputF(out, "\t\t%s %s", type, name);

					switch(anno.mDesc.mType) {
						case kTypeBool:
							VDShaderWriteLogOutputF(out, " = %s;\n", eff.GetBool(&anno) ? "true" : "false");
							break;
						case kTypeFloat:
							VDShaderWriteLogOutputF(out, " = %g;\n", eff.GetFloat(&anno));
							break;
						case kTypeString:
							VDShaderWriteLogOutputF(out, " = \"%s\";\n", eff.GetString(&anno));
							break;
						default:
							out.WriteLogOutput(";\n");
					}
				}

				out.WriteLogOutput("\t> ");
			}

			out.WriteLogOutput("{\n");

			int psindex = pass.GetPixelShaderIndex();
			if (psindex >= 0) {
				out.WriteLogOutput("\t\tPixelShader = asm {\n");

				const VDShaderFunction& func = *eff.GetFunctionByIndex(psindex);
				VDShaderDisassembleFunction(out, func, &nameTable);

				out.WriteLogOutput("\t\t};\n");
			}

			out.WriteLogOutput("\t}\n");
		}

		out.WriteLogOutput("}\n");
	}
}
