//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include "effect.h"

///////////////////////////////////////////////////////////////////////////////
namespace {
	unsigned hash(const char *s, size_t len) {
		unsigned h = 0;

		while(len--) {
			h = (h * 33) ^ (unsigned)*s++;
		}

		return h;
	}
}

bool VDShaderConstant::operator==(const VDShaderConstant& c) const {
	return v[0] == c.v[0]
		&& v[1] == c.v[1]
		&& v[2] == c.v[2]
		&& v[3] == c.v[3];
}

bool VDShaderConstant::operator!=(const VDShaderConstant& c) const {
	return v[0] != c.v[0]
		|| v[1] != c.v[1]
		|| v[2] != c.v[2]
		|| v[3] != c.v[3];
}

size_t VDShaderConstantHash::operator()(const VDShaderConstant& c) const {
	union { float f; int i; } k[4]={
		c.v[0], c.v[1], c.v[2], c.v[3],
	};

	return k[0].i + k[1].i + k[2].i + k[3].i;
}

///////////////////////////////////////////////////////////////////////////////

VDShaderFunction::VDShaderFunction() 
	: mParameterRegisterCount(0)
	, mRegisterCount(0)
	, mSamplerCount(0)
{
}

VDShaderFunction::~VDShaderFunction() {
}

bool VDShaderFunction::IsConstantRegisterBound(int index) const {
	int r = index + kVDShaderRegC0;

	for(ConstantBindings::const_iterator it(mConstantBindings.begin()), itEnd(mConstantBindings.end()); it != itEnd; ++it) {
		const VDShaderConstantBinding& binding = *it;

		if ((uint32)(r - binding.mRegister) < (uint32)binding.mSize)
			return !binding.mbExtern;
	}

	return false;
}

///////////////////////////////////////////////////////////////////////////////

VDShaderNameTable::VDShaderNameTable() {
	mNameHeap.push_back(0);
}

VDShaderNameTable::~VDShaderNameTable() {
}

int VDShaderNameTable::GetName(const char *s, size_t len) const {
	unsigned h = hash(s, len);

	std::pair<NameMap::const_iterator, NameMap::const_iterator> range(mNameMap.equal_range(h));

	for(; range.first != range.second; ++range.first) {
		int offset = range.first->second;

		const char *t = &mNameHeap[offset];
		size_t len2 = strlen(t);
		if (len == len2 && !memcmp(s, t, len))
			return offset;
	}

	return 0;
}

int VDShaderNameTable::AddName(const char *s, size_t len) {
	unsigned h = hash(s, len);

	std::pair<NameMap::const_iterator, NameMap::const_iterator> range(mNameMap.equal_range(h));

	for(; range.first != range.second; ++range.first) {
		int offset = range.first->second;

		const char *t = &mNameHeap[offset];
		size_t len2 = strlen(t);
		if (len == len2 && !memcmp(s, t, len))
			return offset;
	}

	int pos = (int)mNameHeap.size();

	mNameHeap.insert(mNameHeap.end(), s, s+len);
	mNameHeap.push_back(0);
	mNameMap.insert(NameMap::value_type(h, pos));
	return pos;
}

///////////////////////////////////////////////////////////////////////////////
VDShaderPass::VDShaderPass() {
	mAnnotations.mStart = 0;
	mAnnotations.mCount = 0;
}

VDShaderPass::~VDShaderPass() {
}

///////////////////////////////////////////////////////////////////////////////
VDShaderTechnique::VDShaderTechnique() {
	mAnnotations.mStart = 0;
	mAnnotations.mCount = 0;
}

VDShaderTechnique::~VDShaderTechnique() {
	while(!mPasses.empty()) {
		delete mPasses.back();
		mPasses.pop_back();
	}
}

///////////////////////////////////////////////////////////////////////////////
VDShaderEffect::VDShaderEffect() {
}

VDShaderEffect::~VDShaderEffect() {
	while(!mFunctions.empty()) {
		delete mFunctions.back();
		mFunctions.pop_back();
	}

	while(!mTechniques.empty()) {
		delete mTechniques.back();
		mTechniques.pop_back();
	}
}

const VDShaderVariable *VDShaderEffect::GetVariableByName(int nameToken) const {
	if (!nameToken)
		return NULL;

	for(Variables::const_iterator it(mVariables.begin()), itEnd(mVariables.end()); it != itEnd; ++it) {
		const VDShaderVariable& var = *it;

		if (nameToken == var.mDesc.mNameToken)
			return &var;
	}

	return NULL;
}

const VDShaderVariable *VDShaderEffect::GetAnnotationByName(const VDShaderAnnotations& annos, int nameToken) const {
	if (!nameToken)
		return NULL;

	for(int i=0; i<annos.mCount; ++i) {
		const VDShaderVariable& anno = mAnnotations[(annos.mStart - 1) + i];

		if (anno.mDesc.mNameToken == nameToken)
			return &anno;
	}

	return NULL;
}

bool VDShaderEffect::GetBool(const VDShaderVariable *var) const {
	if (!var)
		return false;

	if (var->mDesc.mType != nsVDShaderParser::kTypeBool)
		return false;

	if (!var->mValueToken)
		return false;

	return mDataStorage[var->mValueToken - 1] != 0;
}

float VDShaderEffect::GetFloat(const VDShaderVariable *var) const {
	if (!var)
		return 0;

//	if (var->mDesc.mType == nsVDShaderParser::kTypeInt)
//		return (float)GetInt(var);

	if (var->mDesc.mType != nsVDShaderParser::kTypeFloat)
		return 0;

	if (!var->mValueToken)
		return 0;

	float v;
	memcpy(&v, &mDataStorage[var->mValueToken - 1], sizeof(float));
	return v;
}

int VDShaderEffect::GetInt(const VDShaderVariable *var) const {
	if (!var)
		return 0;

	if (var->mDesc.mType == nsVDShaderParser::kTypeFloat)
		return (int)GetFloat(var);

//	if (var->mDesc.mType != nsVDShaderParser::kTypeInt)
		return 0;

	if (!var->mValueToken)
		return 0;

	int v;
	memcpy(&v, &mDataStorage[var->mValueToken - 1], sizeof(int));
	return v;
}

VDShaderConstant VDShaderEffect::GetVector(const VDShaderVariable *var) const {
	VDShaderConstant v = {0};

	if (!var || !var->mValueToken)
		return v;

	if (var->mDesc.mType != nsVDShaderParser::kTypeFloat &&
		var->mDesc.mType != nsVDShaderParser::kTypeFloat2 &&
		var->mDesc.mType != nsVDShaderParser::kTypeFloat3 &&
		var->mDesc.mType != nsVDShaderParser::kTypeFloat4
		)
		return v;

	memcpy(&v, &mDataStorage[var->mValueToken - 1], sizeof(float) * (var->mDesc.mType - nsVDShaderParser::kTypeFloat + 1));
	return v;
}

const char *VDShaderEffect::GetString(const VDShaderVariable *var) const {
	if (!var->mValueToken)
		return "";

	int v;
	memcpy(&v, &mDataStorage[var->mValueToken - 1], sizeof(int));

	return mNameTable.GetName(v);
}

void VDShaderEffect::SetFloat(const VDShaderVariable *var, float v) const {
	if (!var || !var->mValueToken)
		return;

	VDShaderConstant c;
	c.v[0] = v;
	c.v[1] = 0;
	c.v[2] = 0;
	c.v[3] = 0;

	char *dst = (char *)&mDataStorage[var->mValueToken - 1];

	switch(var->mDesc.mType) {
		case nsVDShaderParser::kTypeFloat:
			memcpy(dst, c.v, sizeof(float));
			break;
		default:
			SetVector(var, c);
			break;
	}
}

void VDShaderEffect::SetVector(const VDShaderVariable *var, const VDShaderConstant& c) const {
	if (!var || !var->mValueToken)
		return;

	char *dst = (char *)&mDataStorage[var->mValueToken - 1];

	switch(var->mDesc.mType) {
		case nsVDShaderParser::kTypeBool4:
			dst[3] = (c.v[3] != 0);
		case nsVDShaderParser::kTypeBool3:
			dst[2] = (c.v[2] != 0);
		case nsVDShaderParser::kTypeBool2:
			dst[1] = (c.v[1] != 0);
		case nsVDShaderParser::kTypeBool:
			dst[0] = (c.v[0] != 0);
			break;

		case nsVDShaderParser::kTypeFloat:
			memcpy(dst, c.v, sizeof(float));
			break;
		case nsVDShaderParser::kTypeFloat2:
			memcpy(dst, c.v, sizeof(float)*2);
			break;
		case nsVDShaderParser::kTypeFloat3:
			memcpy(dst, c.v, sizeof(float)*3);
			break;
		case nsVDShaderParser::kTypeFloat4:
			memcpy(dst, c.v, sizeof(float)*4);
			break;
	}
}
