#ifndef EFFECT_H
#define EFFECT_H

#include <vd2/system/refcount.h>
#include <vd2/system/vdstl.h>
#include <map>

namespace nsVDShaderParser {
	enum VDShaderParserType {
		kTypeInvalid,
		kTypeVoid,
		kTypeVertexShader,
		kTypePixelShader,
		kTypeFloat,
		kTypeFloat2,
		kTypeFloat3,
		kTypeFloat4,
		kTypeFloat2x2,
		kTypeFloat2x3,
		kTypeFloat2x4,
		kTypeFloat3x2,
		kTypeFloat3x3,
		kTypeFloat3x4,
		kTypeFloat4x2,
		kTypeFloat4x3,
		kTypeFloat4x4,
		kTypeSampler,
		kTypeTexture,
		kTypeBool,
		kTypeBool2,
		kTypeBool3,
		kTypeBool4,
		kTypeString
	};
}

enum {
	kVDShaderOp_nop,
	kVDShaderOp_abs,
	kVDShaderOp_add,
	kVDShaderOp_cmp,
	kVDShaderOp_crs,
	kVDShaderOp_dp2add,
	kVDShaderOp_dp3,
	kVDShaderOp_dp4,
	kVDShaderOp_exp,
	kVDShaderOp_frc,
	kVDShaderOp_log,
	kVDShaderOp_lrp,
	kVDShaderOp_m3x2,
	kVDShaderOp_m3x3,
	kVDShaderOp_m3x4,
	kVDShaderOp_m4x3,
	kVDShaderOp_m4x4,
	kVDShaderOp_mad,
	kVDShaderOp_max,
	kVDShaderOp_min,
	kVDShaderOp_mov,
	kVDShaderOp_mul,
	kVDShaderOp_nrm,
	kVDShaderOp_pow,
	kVDShaderOp_rcp,
	kVDShaderOp_rsq,
	kVDShaderOp_sincos,
	kVDShaderOp_sub,
	kVDShaderOp_texkill,
	kVDShaderOp_texld,
	kVDShaderOp_texldb,
	kVDShaderOp_texldp
};

enum {
	kVDShaderRegR0 = 0x1000,		// temp registers
	kVDShaderRegC0 = 0x2000,		// fixed constants
	kVDShaderRegV0 = 0x3000,		// input registers
	kVDShaderRegO0 = 0x4000,		// outputs
	kVDShaderRegS0 = 0x5000,		// sampler
	kVDShaderRegIndexMask = 0x0FFF,
	kVDShaderRegTypeMask = 0xF000
};

enum {
	kVDShaderMod_Negate = 0x01
};

enum {
	kVDShaderInsnMod_Saturate = 0x01
};

///////////////////////////////////////////////////////////////////////////////
struct VDShaderConstant {
	float v[4];

	VDShaderConstant& operator=(float f)
	{
		v[0] = f;
		v[1] = f;
		v[2] = f;
		v[3] = f;
		return *this;
	}

	bool operator==(const VDShaderConstant& c) const;
	bool operator!=(const VDShaderConstant& c) const;
};

struct VDShaderConstantHash {
	size_t operator()(const VDShaderConstant& c) const;
};

///////////////////////////////////////////////////////////////////////////////
struct VDShaderInsnSource {
	int mRegister;
	unsigned char mSwizzle;
	unsigned char mModifiers;

	static VDShaderInsnSource reg(int rg) {
		VDShaderInsnSource r = { rg, 0xE4, 0 };
		return r;
	}

	VDShaderInsnSource operator-() const {
		VDShaderInsnSource r(*this);
		r.mModifiers ^= kVDShaderMod_Negate;
		return r;
	}

	VDShaderInsnSource swizzle(uint32 sw) const {
		VDShaderInsnSource r(*this);
		r.mSwizzle = (((mSwizzle >> (((sw >> 12) & 3)*2)) & 3) << 0)
					+ (((mSwizzle >> (((sw >>  8) & 3)*2)) & 3) << 2)
					+ (((mSwizzle >> (((sw >>  4) & 3)*2)) & 3) << 4)
					+ (((mSwizzle >> (((sw >>  0) & 3)*2)) & 3) << 6);
		return r;
	}
};

///////////////////////////////////////////////////////////////////////////////
struct VDShaderInsn {
	int mOp;
	int mDestRegister;
	unsigned char mWriteMask;
	unsigned char mModifiers;
	VDShaderInsnSource mSources[4];

	static VDShaderInsn op(int opcode, int dst, uint8 writemask, const VDShaderInsnSource& src1) {
		VDShaderInsn r = { opcode, dst, writemask, 0, { src1 }};
		return r;
	}

	static VDShaderInsn op(int opcode, int dst, uint8 writemask, const VDShaderInsnSource& src1, const VDShaderInsnSource& src2) {
		VDShaderInsn r = { opcode, dst, writemask, 0, { src1, src2 }};
		return r;
	}

	static VDShaderInsn op(int opcode, int dst, uint8 writemask, const VDShaderInsnSource& src1, const VDShaderInsnSource& src2, const VDShaderInsnSource& src3) {
		VDShaderInsn r = { opcode, dst, writemask, 0, { src1, src2, src3 }};
		return r;
	}
};

///////////////////////////////////////////////////////////////////////////////
enum VDShaderFilterMode {
	kFilterNone,
	kFilterPoint,
	kFilterBilinear
};

struct VDShaderSamplerState {
	uint32	mTextureToken;
	bool	mbAddressUWrap : 1;
	bool	mbAddressUWrapSet : 1;
	bool	mbAddressVWrap : 1;
	bool	mbAddressVWrapSet : 1;
	bool	mbMinFilter : 1;
	bool	mbMinFilterSet : 1;
	bool	mbMagFilter : 1;
	bool	mbMagFilterSet : 1;
	bool	mbMipFilter : 1;
	bool	mbMipFilterSet : 1;
};

struct VDShaderSamplerBinding {
	int		mNameToken;
	int		mStateToken;

	bool operator==(const VDShaderSamplerBinding& src) const {
		return mNameToken == src.mNameToken && mStateToken == src.mStateToken;
	}
};

///////////////////////////////////////////////////////////////////////////////
struct VDShaderAnnotations {
	int mStart;
	int mCount;
};

///////////////////////////////////////////////////////////////////////////////
struct VDShaderVariableDesc {
	int mNameToken;
	int mSemanticToken;
	int mRegister;
	int mType;
	bool mbUniform : 1;
	bool mbUnassigned : 1;
	bool mbExtern : 1;
	bool mbStatic : 1;
	VDShaderAnnotations mAnnotations;
};

struct VDShaderVariable {
	VDShaderVariableDesc mDesc;
	int mValueToken;
	int mSize;
};

struct VDShaderConstantBinding {
	int mNameToken;
	int mRegister;
	int mForcedRegister;
	int mSize;
	bool mbExtern;
};

///////////////////////////////////////////////////////////////////////////////
struct VDShaderFunction {
	VDShaderFunction();
	~VDShaderFunction();

	bool IsConstantRegisterBound(int index) const;

	typedef vdfastvector<VDShaderVariableDesc> Parameters;
	Parameters mParameters;
	int mParameterRegisterCount;
	int mRegisterCount;
	int mSamplerCount;
	int mReturnType;
	int mReturnSemantic;
	int mNameToken;

	typedef vdfastvector<VDShaderConstant> Constants;
	Constants mConstants;

	typedef vdfastvector<VDShaderInsn> Insns;
	Insns mInsns;

	typedef vdfastvector<VDShaderSamplerBinding> SamplerBindings;
	SamplerBindings mSamplerBindings;

	typedef vdfastvector<VDShaderConstantBinding> ConstantBindings;
	ConstantBindings mConstantBindings;
};

///////////////////////////////////////////////////////////////////////////////
class VDShaderNameTable {
public:
	VDShaderNameTable();
	~VDShaderNameTable();

	int AddName(const char *s, size_t len);

	int GetName(const char *s, size_t len) const;

	const char *GetName(int id) const {
		return &mNameHeap[id];
	}

protected:
	typedef vdfastvector<char> NameHeap;
	NameHeap mNameHeap;

	typedef std::multimap<unsigned, int> NameMap;
	NameMap mNameMap;
};

///////////////////////////////////////////////////////////////////////////////
class VDShaderPass {
public:
	VDShaderPass();
	~VDShaderPass();

	int GetNameToken() const { return mNameToken; }
	int GetPixelShaderIndex() const { return mPixelShaderIndex; }
	const VDShaderAnnotations& GetAnnotations() const { return mAnnotations; }

protected:
	friend class VDShaderParser;

	int mNameToken;
	int mPixelShaderIndex;
	VDShaderAnnotations mAnnotations;
};

///////////////////////////////////////////////////////////////////////////////
class VDShaderTechnique {
public:
	VDShaderTechnique();
	~VDShaderTechnique();

	int GetNameToken() const { return mNameToken; }
	int GetPassCount() const { return mPasses.size(); }
	VDShaderPass *GetPassByIndex(int index) const {
		return (unsigned)index < mPasses.size() ? mPasses[index] : NULL;
	}

	const VDShaderAnnotations& GetAnnotations() const { return mAnnotations; }

protected:
	friend class VDShaderParser;

	int mNameToken;

	typedef vdfastvector<VDShaderPass *> Passes;
	Passes mPasses;

	VDShaderAnnotations mAnnotations;
};

///////////////////////////////////////////////////////////////////////////////
class VDShaderEffect : public vdrefcounted<IVDRefCount> {
public:
	VDShaderEffect();
	~VDShaderEffect();

	const VDShaderNameTable& NameTable() const { return mNameTable; }

	int GetFunctionCount() const { return mFunctions.size(); }
	int GetTechniqueCount() const { return mTechniques.size(); }
	int GetVariableCount() const { return mVariables.size(); }
	int GetSamplerStateCount() const { return mSamplerStates.size(); }
	int GetAnnotationCount() const { return mAnnotations.size(); }

	const VDShaderFunction *GetFunctionByIndex(int index) const {
		return (unsigned)index < mFunctions.size() ? mFunctions[index] : NULL;
	}

	const VDShaderTechnique *GetTechniqueByIndex(int index) const {
		return (unsigned)index < mTechniques.size() ? mTechniques[index] : NULL;
	}

	const VDShaderVariable *GetVariableByIndex(int index) const {
		return (unsigned)index < mVariables.size() ? &mVariables[index] : NULL;
	}

	const VDShaderVariable *GetVariableByName(int nameToken) const;

	const VDShaderSamplerState *GetSamplerStateByIndex(int index) const {
		return (unsigned)(index - 1) < mSamplerStates.size() ? &mSamplerStates[index - 1] : NULL;
	}

	const VDShaderVariable *GetAnnotationByIndex(int index) const {
		return (unsigned)(index - 1) < mAnnotations.size() ? &mAnnotations[index - 1] : NULL;
	}

	const VDShaderVariable *GetAnnotationByName(const VDShaderAnnotations& anno, int nameToken) const;

	bool GetBool(const VDShaderVariable *var) const;
	float GetFloat(const VDShaderVariable *var) const;
	int GetInt(const VDShaderVariable *var) const;
	VDShaderConstant GetVector(const VDShaderVariable *var) const;
	const char *GetString(const VDShaderVariable *var) const;

	void SetFloat(const VDShaderVariable *var, float v) const;
	void SetVector(const VDShaderVariable *var, const VDShaderConstant& c) const;

protected:
	friend class VDShaderParser;

	VDShaderNameTable mNameTable;

	typedef vdfastvector<VDShaderFunction *> Functions;
	Functions mFunctions;

	typedef vdfastvector<VDShaderTechnique *> Techniques;
	Techniques mTechniques;

	typedef vdfastvector<VDShaderVariable> Variables;
	Variables mVariables;

	typedef vdfastvector<VDShaderSamplerState> SamplerStates;
	SamplerStates mSamplerStates;

	typedef vdfastvector<VDShaderVariable> Annotations;
	Annotations mAnnotations;

	typedef vdfastvector<char> DataStorage;
	DataStorage mDataStorage;
};

struct VDShaderBindingElement {
	int mIndex;
	int mTypeToken;
	int mSemanticToken;
};

#endif
