//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include <new>
#include <float.h>
#include <vd2/system/error.h>
#include <vd2/system/file.h>
#include <vd2/system/math.h>
#include <vd2/system/registry.h>
#include <vd2/system/strutil.h>
#include <vd2/system/VDString.h>
#include <vd2/system/filesys.h>
#include <vd2/plugin/vdvideofilt.h>
#include <vd2/VDXFrame/VideoFilter.h>
#include "vdfilter.h"
#include "effect.h"
#include "parser.h"
#include "interp.h"
#include "log.h"
#include "jit_scalar.h"
#include "jit_sse2.h"
#include "editor.h"
#include "execengine.h"
#include "paramdialog.h"

int g_VDFilterAPIVersion;

///////////////////////////////////////////////////////////////////////////

struct VDShaderFilterFrameKey {
	VDShaderFilterFrameKey();

	bool operator==(const VDShaderFilterFrameKey& x) const;

	sint64		mCurrentFrame;
	bool		mbClampU;
	bool		mbClampV;
};

VDShaderFilterFrameKey::VDShaderFilterFrameKey()
	: mCurrentFrame(-1)
	, mbClampU(false)
	, mbClampV(false)
{
}

bool VDShaderFilterFrameKey::operator==(const VDShaderFilterFrameKey& x) const {
	return mCurrentFrame == x.mCurrentFrame
			&& mbClampU == x.mbClampU
			&& mbClampV == x.mbClampV;
}

///////////////////////////////////////////////////////////////////////////

class VDShaderFilterTexture;

class VDShaderFilterTextureAllocator {
	VDShaderFilterTextureAllocator(const VDShaderFilterTextureAllocator&);
	VDShaderFilterTextureAllocator& operator=(const VDShaderFilterTextureAllocator&);
public:
	VDShaderFilterTextureAllocator();
	~VDShaderFilterTextureAllocator();

	void Init(uint32 w, uint32 h);
	void Shutdown();

	bool GetBuffer(const VDShaderFilterFrameKey& frameKey, VDShaderFilterTexture **ppTex);
	bool AllocBuffer(const VDShaderFilterFrameKey& frameKey, VDShaderFilterTexture **ppTex);

	void NotifyIdle(VDShaderFilterTexture *buf);

protected:
	uint32 mWidth;
	uint32 mHeight;

	typedef vdfastvector<VDShaderFilterTexture *> Buffers;
	Buffers mActiveBuffers;
	Buffers mFreeBuffers;
};

class VDShaderFilterTexture {
	VDShaderFilterTexture(const VDShaderFilterTexture&);
	VDShaderFilterTexture& operator=(const VDShaderFilterTexture&);
public:
	VDShaderFilterTexture();
	~VDShaderFilterTexture();

	int AddRef();
	int Release();
	bool Init(uint32 w, uint32 h);
	void Attach(VDShaderFilterTextureAllocator *alloc);
	void Detach();

	VDShaderFilterTextureAllocator *mpAllocator;
	uint32		mRefCount;
	uint32		mWidth;
	uint32		mHeight;
	uint32		*mpBits;
	ptrdiff_t	mPitch;
	VDShaderFilterFrameKey	mFrameKey;
};

VDShaderFilterTexture::VDShaderFilterTexture()
	: mpAllocator(NULL)
	, mRefCount(0)
	, mWidth(0)
	, mHeight(0)
	, mpBits(NULL)
	, mPitch(0)
{
}

VDShaderFilterTexture::~VDShaderFilterTexture() {
	if (mpBits)
		free(mpBits);
}

int VDShaderFilterTexture::AddRef() {
	return ++mRefCount;
}

int VDShaderFilterTexture::Release() {
	int v = --mRefCount;

	if (v == 1) {
		if (mpAllocator)
			mpAllocator->NotifyIdle(this);
	} else if (v == 0)
		delete this;

	return v;
}

bool VDShaderFilterTexture::Init(uint32 w, uint32 h) {
	uint32 rowsize = sizeof(uint32) * (w + 2);

	mpBits = (uint32 *)malloc(sizeof(uint32) * (w + 2) * (h + 2));
	if (!mpBits)
		return false;

	mWidth = w;
	mHeight = h;
	mPitch = rowsize;
	return true;
}

void VDShaderFilterTexture::Attach(VDShaderFilterTextureAllocator *alloc) {
	AddRef();
	mpAllocator = alloc;
}

void VDShaderFilterTexture::Detach() {
	mpAllocator = NULL;
	Release();
}

///////////////////////////////////////////////////////////////////////////


VDShaderFilterTextureAllocator::VDShaderFilterTextureAllocator()
	: mWidth(0)
	, mHeight(0)
{
}

VDShaderFilterTextureAllocator::~VDShaderFilterTextureAllocator() {
	Shutdown();
}

void VDShaderFilterTextureAllocator::Init(uint32 w, uint32 h) {
	mWidth = w;
	mHeight = h;
}

void VDShaderFilterTextureAllocator::Shutdown() {
	while(!mActiveBuffers.empty()) {
		VDShaderFilterTexture *tb = mActiveBuffers.back();
		mActiveBuffers.pop_back();

		tb->Detach();
	}

	while(!mFreeBuffers.empty()) {
		VDShaderFilterTexture *tb = mFreeBuffers.back();
		mFreeBuffers.pop_back();

		tb->Detach();
	}
}

bool VDShaderFilterTextureAllocator::GetBuffer(const VDShaderFilterFrameKey& frameKey, VDShaderFilterTexture **ppTex) {
	for(Buffers::const_iterator it(mActiveBuffers.begin()), itEnd(mActiveBuffers.end()); it != itEnd; ++it) {
		VDShaderFilterTexture *buf = *it;

		if (buf->mFrameKey == frameKey) {
			buf->AddRef();
			*ppTex = buf;
			return true;
		}
	}

	for(Buffers::iterator it(mFreeBuffers.begin()), itEnd(mFreeBuffers.end()); it != itEnd; ++it) {
		VDShaderFilterTexture *buf = *it;

		if (buf->mFrameKey == frameKey) {
			mFreeBuffers.erase(it);

			VDASSERT(std::find(mActiveBuffers.begin(), mActiveBuffers.end(), buf) == mActiveBuffers.end());
			mActiveBuffers.push_back(buf);
			buf->AddRef();
			*ppTex = buf;
			return true;
		}
	}

	return false;
}

bool VDShaderFilterTextureAllocator::AllocBuffer(const VDShaderFilterFrameKey& frameKey, VDShaderFilterTexture **ppTex) {
	VDShaderFilterTexture *buf;

	if (mFreeBuffers.empty()) {
		buf = new VDShaderFilterTexture;
		if (!buf)
			return false;

		if (!buf->Init(mWidth, mHeight)) {
			delete buf;
			return false;
		}

		buf->Attach(this);
	} else {
		buf = mFreeBuffers.front();
		mFreeBuffers.erase(mFreeBuffers.begin());
	}

	buf->mFrameKey = frameKey;
	buf->AddRef();

	VDASSERT(std::find(mActiveBuffers.begin(), mActiveBuffers.end(), buf) == mActiveBuffers.end());
	mActiveBuffers.push_back(buf);
	*ppTex = buf;
	return true;
}

void VDShaderFilterTextureAllocator::NotifyIdle(VDShaderFilterTexture *buf) {
	Buffers::iterator it = std::find(mActiveBuffers.begin(), mActiveBuffers.end(), buf);

	if (it == mActiveBuffers.end()) {
		VDASSERT(!"NotifyIdle() called on buffer not in active list.");
	} else {
		*it = mActiveBuffers.back();
		mActiveBuffers.pop_back();
	}

	mFreeBuffers.push_back(buf);
}

///////////////////////////////////////////////////////////////////////////

struct VDShaderFilterSourceChannel {
	VDShaderFilterTextureAllocator mAllocator;
	uint32 mWidth;
	uint32 mHeight;
};

///////////////////////////////////////////////////////////////////////////

struct VDShaderFilterSampler {
	void InitDynamic(int fetchIndex, const VDXFBitmap& bm, bool clampU, bool clampV, bool bilinear);
	void InitStatic(VDShaderFilterTexture *tex, bool clampU, bool clampV, bool bilinear);
	bool IsDynamic() const { return mFetchIndex >= 0; }
	void BindToTexture(VDShaderFilterTexture *tex);
	void Unbind();

	int mFetchIndex;
	VDShaderInterpreterSampler mSampler;
	vdrefptr<VDShaderFilterTexture> mpBoundTexture;
};

void VDShaderFilterSampler::InitDynamic(int fetchIndex, const VDXFBitmap& bm, bool clampU, bool clampV, bool bilinear) {
	uint32 w;
	uint32 h;

	if (g_VDFilterAPIVersion >= 11) {
		w = bm.mpPixmapLayout->w;
		h = bm.mpPixmapLayout->h;
	} else {
		w = bm.w;
		h = bm.h;
	}

	mFetchIndex = fetchIndex;
	mSampler.Init(NULL, (w+2)*sizeof(uint32), w, h, clampU, clampV, bilinear);
}

void VDShaderFilterSampler::InitStatic(VDShaderFilterTexture *tex, bool clampU, bool clampV, bool bilinear) {
	mFetchIndex = -1;

	const uint32 w = tex->mWidth;
	const uint32 h = tex->mHeight;
	mSampler.Init(vdptroffset(tex->mpBits, tex->mPitch) + 1, (w+2)*sizeof(uint32), w, h, clampU, clampV, bilinear);
	mpBoundTexture = tex;
}

void VDShaderFilterSampler::BindToTexture(VDShaderFilterTexture *tex) {
	mSampler.mpSrc = vdptroffset(tex->mpBits, tex->mPitch) + 1;
	mpBoundTexture = tex;
}

void VDShaderFilterSampler::Unbind() {
	mSampler.mpSrc = NULL;
	mpBoundTexture = NULL;
}

///////////////////////////////////////////////////////////////////////////

struct VDShaderFilterPass {
	VDShaderFilterPass();

	void SetSampler(int samplerIndex, int fetchIndex, const VDXFBitmap& bm, bool clampU, bool clampV, bool bilinear);
	void SetSampler(int samplerIndex, VDShaderFilterTexture *tex, bool clampU, bool clampV, bool bilinear);

	IVDShaderExecutionEngine *mpEngine;
	uint32 mFieldMask;

	vdfastvector<VDShaderConstant>	mParamDisplay;
	vdfastvector<VDShaderInterpConstant>	mParamConstants;
	vdfastvector<const VDShaderVariable *>	mParamVarLookup;

	const VDShaderInterpreterSampler *mpSamplers[16];
	VDShaderFilterSampler	mSamplers[16];
};

VDShaderFilterPass::VDShaderFilterPass() {
	for(int i=0; i<16; ++i) {
		mpSamplers[i] = NULL;
		mSamplers[i].mFetchIndex = -1;
	}
}

void VDShaderFilterPass::SetSampler(int samplerIndex, int fetchIndex, const VDXFBitmap& bm, bool clampU, bool clampV, bool bilinear) {
	mSamplers[samplerIndex].InitDynamic(fetchIndex, bm, clampU, clampV, bilinear);
	mpSamplers[samplerIndex] = &mSamplers[samplerIndex].mSampler;
}

void VDShaderFilterPass::SetSampler(int samplerIndex, VDShaderFilterTexture *tex, bool clampU, bool clampV, bool bilinear) {
	mSamplers[samplerIndex].InitStatic(tex, clampU, clampV, bilinear);
	mpSamplers[samplerIndex] = &mSamplers[samplerIndex].mSampler;
}

///////////////////////////////////////////////////////////////////////////

class VDShaderFilter : public VDXVideoFilter, public IVDShaderFilter {
public:
	VDShaderFilter();

	uint32 GetParams();
	bool Init();
	void Start();
	void End();
	void Run();
	bool Configure(VDXHWND hwnd);
	int Serialize(char *buf, int maxbuf);
	int Deserialize(const char *buf, int maxbuf);
	void GetSettingString(char *buf, int maxlen);
	void GetScriptString(char *buf, int maxlen);

	VDXVF_DECLARE_SCRIPT_METHODS();

	enum {
		kMinInputCount = 1,
		kMaxInputCount = 4
	};

	bool Prefetch2(sint64 frame, IVDXVideoPrefetcher *prefetcher);

	void GetParamBindings(vdfastvector<VDFXUIBinding>& paramBindings);
	void SetEffectParameterLive(const char *name, const VDShaderConstant& v);

protected:
	typedef VDShaderFilterPass Pass;
	typedef VDShaderFilterTexture Texture;
	typedef VDShaderFilterSourceChannel SourceChannel;

	void ScriptConfig(IVDXScriptInterpreter *, const VDXScriptValue *, int);

	void GetEffect(VDShaderEffect **ppEffect);
	void ClearUIBindings();
	void UpdateUIBindings(VDShaderEffect& eff);
	int AddFrameFetch(int srcIndex, int srcFrameOffset);
	void UploadTexture(const Texture& dst, const VDXFBitmap& src, bool clampU, bool clampV);
	void UploadTexture(const Texture& dst, const void *src, ptrdiff_t pitch, uint32 w, uint32 h, bool clampU, bool clampV);

	bool		mbConstantsDirty;

	vdrefptr<VDShaderEffect> mpEffect;
	VDShaderInterpreterInterpolator mTexCoord0;

	struct FrameFetch {
		int mSrcIndex;
		int mFrameOffset;
	};

	typedef vdfastvector<FrameFetch> FrameFetches;
	FrameFetches mFrameFetches;

	typedef vdfastvector<SourceChannel *> SourceChannels;
	SourceChannels mSourceChannels;

	typedef vdfastvector<Pass *> Passes;
	Passes mPasses;

	VDShaderFilterConfig	mConfig;

	IVDShaderFilterSource	*mpEditorSource;

	const VDShaderFilterDefinition *mpShaderDef;

	vdfastvector<VDFXUIBinding> mParamUIBindings;
};

VDXVF_BEGIN_SCRIPT_METHODS(VDShaderFilter)
	VDXVF_DEFINE_SCRIPT_METHOD(VDShaderFilter, ScriptConfig, "iiisis")
VDXVF_END_SCRIPT_METHODS()

VDShaderFilter::VDShaderFilter() {
}

uint32 VDShaderFilter::GetParams() {
	if (g_VDFilterAPIVersion >= 11) {
		const VDXPixmapLayout& pxlsrc = *fa->src.mpPixmapLayout;

		if (pxlsrc.format != nsVDXPixmap::kPixFormat_XRGB8888)
			return FILTERPARAM_NOT_SUPPORTED;
	}

	if (mConfig.mbKeepSize) {
		fa->dst.offset = fa->src.offset;
	} else {
		uint32 w, h;

		if (mConfig.mNewWidth <= 0 || mConfig.mNewHeight <= 0) {
			w = 1;
			h = 1;
		} else {
			w = mConfig.mNewWidth;
			h = mConfig.mNewHeight;
		}

		if (g_VDFilterAPIVersion >= 11) {
			VDXPixmapLayout& pxldst = *fa->dst.mpPixmapLayout;

			fa->dst.depth = 0;

			pxldst.w = w;
			pxldst.h = h;
			pxldst.data = 0;
			pxldst.pitch = (pxldst.w * 4 + 15) & ~15;
			pxldst.format = nsVDXPixmap::kPixFormat_XRGB8888;
		} else {
			fa->dst.w = w;
			fa->dst.h = h;
			fa->dst.offset = 0;
			fa->dst.pitch = (fa->dst.w * 4 + 15) & ~15;
			fa->dst.size = fa->dst.pitch * fa->dst.h;
			fa->dst.modulo = fa->dst.pitch - fa->dst.w * 4;
		}
	}

	return FILTERPARAM_SUPPORTS_ALTFORMATS | FILTERPARAM_PURE_TRANSFORM;
}

namespace {
	class VDShaderParserBufOutput : public IVDShaderLogOutput {
	public:
		const char *c_str() const { return &mBuffer[0]; }

		void WriteLogOutput(const char *s) {
			mBuffer.insert(mBuffer.end(), s, s+strlen(s));
		}

		void End() {
			mBuffer.push_back(0);
		}

	protected:
		std::vector<char> mBuffer;
	};
}

bool VDShaderFilter::Init() {
	mpShaderDef = static_cast<const VDShaderFilterDefinition *>(fa->filter->private_data);

	if (mpShaderDef)
		mConfig.mFilename = VDTextWToA(mpShaderDef->mShaderPath);

	return true;
}

void VDShaderFilter::Start() {
	const uint32 numSrcStreams = (g_VDFilterAPIVersion >= 16) ? fa->mSourceStreamCount : 1;
	VDXFBitmap *const srcStream0 = &fa->src;
	const VDXFBitmap *const *srcStreams = (g_VDFilterAPIVersion >= 16) ? fa->mpSourceStreams : &srcStream0;

	for(uint32 i=0; i<numSrcStreams; ++i) {
		SourceChannel *sc = new SourceChannel;
		mSourceChannels.push_back(sc);

		const VDXFBitmap& bm = *srcStreams[i];
		if (g_VDFilterAPIVersion >= 11) {
			sc->mWidth = bm.mpPixmapLayout->w;
			sc->mHeight = bm.mpPixmapLayout->h;
		} else {
			sc->mWidth = bm.w;
			sc->mHeight = bm.h;
		}

		sc->mAllocator.Init(sc->mWidth, sc->mHeight);
	}

	if (!mConfig.mbKeepSize) { 
		if (mConfig.mNewWidth <= 0 || mConfig.mNewHeight <= 0) {
			ff->Except("The output frame size parameters are invalid.");
		}
	}

	try {
		GetEffect(~mpEffect);
	} catch(const MyError& e) {
		ff->Except("%s", e.gets());
	}

	if (!mpEffect->GetTechniqueCount())
		ff->Except("No techniques were declared in the .fx file.");

	const VDShaderTechnique *tech = mpEffect->GetTechniqueByIndex(0);

	int passCount = tech->GetPassCount();
	if (!passCount)
		ff->Except("No passes were declared in the technique \"%s.\"", mpEffect->NameTable().GetName(tech->GetNameToken()));

	const VDShaderBindingElement kElements[]={
		0,
		nsVDShaderParser::kTypeFloat2,
		mpEffect->NameTable().GetName("TEXCOORD0", 9)
	};

	int execMode = mConfig.mExecutionMode;

	if (execMode == VDShaderFilterConfig::kModeDefault)
		execMode = VDShaderFilterConfig::kModeSSE2JIT;

	long cpuFlags = ff->getCPUFlags();

	if (execMode == VDShaderFilterConfig::kModeSSE2JIT && !(cpuFlags & CPUF_SUPPORTS_SSE2))
		execMode = VDShaderFilterConfig::kModeScalarJIT;

	//////////// update effect values
	int numVars = mpEffect->GetVariableCount();
	for(int varIdx = 0; varIdx < numVars; ++varIdx) {
		const VDShaderVariable& var = *mpEffect->GetVariableByIndex(varIdx);

		if (var.mDesc.mbExtern)
			continue;

		switch(var.mDesc.mType) {
			case nsVDShaderParser::kTypeBool:
			case nsVDShaderParser::kTypeBool2:
			case nsVDShaderParser::kTypeBool3:
			case nsVDShaderParser::kTypeBool4:
			case nsVDShaderParser::kTypeFloat:
			case nsVDShaderParser::kTypeFloat2:
			case nsVDShaderParser::kTypeFloat3:
			case nsVDShaderParser::kTypeFloat4:
				break;

			default:
				continue;
		}

		// try to match against a saved parameter
		if (!var.mDesc.mNameToken)
			continue;

		const char *name = mpEffect->NameTable().GetName(var.mDesc.mNameToken);
		for(VDShaderFilterConfig::ParamValues::iterator it(mConfig.mParamValues.begin()), itEnd(mConfig.mParamValues.end()); it != itEnd; ++it) {
			if (it->mName == name) {
				mpEffect->SetVector(&var, it->mValue);
				break;
			}
		}
	}

	//////////// process passes

	for(int passIdx = 0; passIdx < passCount; ++passIdx) {
		Pass *execPass = new(std::nothrow) Pass;
		if (!execPass)
			ff->ExceptOutOfMemory();

		execPass->mpEngine = NULL;
		mPasses.push_back(execPass);

		const VDShaderPass *pass = tech->GetPassByIndex(passIdx);
		int psindex = pass->GetPixelShaderIndex();

		if (psindex < 0)
			ff->Except("No pixel shader declared in pass.");

		const VDShaderAnnotations& passAnno = pass->GetAnnotations();

		const VDShaderVariable *var = mpEffect->GetAnnotationByName(passAnno, mpEffect->NameTable().GetName("vd_fieldmask", 12));

		if (var)
			execPass->mFieldMask = mpEffect->GetInt(var) & 3;
		else
			execPass->mFieldMask = 3;

		const VDShaderFunction *func = mpEffect->GetFunctionByIndex(psindex);
		for(int i=0; i<16; ++i) {
			execPass->mpSamplers[i] = NULL;
		}

		int samplerCount = func->mSamplerBindings.size();
		if (samplerCount > 16)
			ff->Except("Max sampler count exceeded (16 max per function).");

		const VDShaderNameTable& nameTable = mpEffect->NameTable();
		const int nameTokSrc = nameTable.GetName("vd_srctexture", 13);
		const int nameTokSrc2 = nameTable.GetName("vd_prevsrctexture", 17);
		const int nameTokSrc3 = nameTable.GetName("vd_prevsrctexture2", 18);
		const int nameSrcIndex = nameTable.GetName("vd_srcindex", 11);
		const int nameSrcFrameOffset = nameTable.GetName("vd_srcframeoffset", 17);
		const int nameWidth = nameTable.GetName("width", 5);
		const int nameHeight = nameTable.GetName("height", 6);
		const int nameFunction = nameTable.GetName("function", 8);

		for(int i=0; i<samplerCount; ++i) {
			const VDShaderSamplerBinding& binding = func->mSamplerBindings[i];
			bool clampU = true;
			bool clampV = true;
			bool useBilinear = true;
			int srcIndex = 0;
			int srcFrameOffset = 0;

			if (binding.mStateToken) {
				const VDShaderSamplerState *state = mpEffect->GetSamplerStateByIndex(binding.mStateToken);

				if (state) {
					if (state->mbAddressUWrapSet)
						clampU = !state->mbAddressUWrap;

					if (state->mbAddressVWrapSet)
						clampV = !state->mbAddressVWrap;

					if (state->mbMagFilterSet)
						useBilinear = state->mbMagFilter;

					if (state->mTextureToken == nameTokSrc2)
						srcFrameOffset = -1;
					else if (state->mTextureToken == nameTokSrc3)
						srcFrameOffset = -2;

					const VDShaderVariable *texVar = mpEffect->GetVariableByName(state->mTextureToken);

					if (texVar) {
						const int numTexAnnos = texVar->mDesc.mAnnotations.mCount;

						for(int annoIdx=0; annoIdx<numTexAnnos; ++annoIdx) {
							const VDShaderVariable *texAnno = mpEffect->GetAnnotationByIndex(texVar->mDesc.mAnnotations.mStart + annoIdx);

							if (!texAnno)
								continue;

							if (texAnno->mDesc.mNameToken == nameSrcIndex) {
								srcIndex = mpEffect->GetInt(texAnno);

								if (srcIndex < 0 || srcIndex > 15)
									ff->Except("Invalid source index in texture declaration.");
							}

							if (texAnno->mDesc.mNameToken == nameSrcFrameOffset)
								srcFrameOffset = mpEffect->GetInt(texAnno);
						}

						const VDShaderVariable *varFunc = mpEffect->GetAnnotationByName(texVar->mDesc.mAnnotations, nameFunction);
						if (varFunc) {
							const char *funcName = mpEffect->GetString(varFunc);
							int funcNameToken = nameTable.GetName(funcName, strlen(funcName));

							const int numFuncs = mpEffect->GetFunctionCount();
							const VDShaderFunction *texFunc = NULL;

							for(int funcIdx = 0; funcIdx < numFuncs; ++funcIdx) {
								const VDShaderFunction *func = mpEffect->GetFunctionByIndex(funcIdx);

								if (func->mNameToken == funcNameToken) {
									texFunc = func;
									break;
								}
							}

							if (!texFunc)
								ff->Except("Unknown function for texture shader: %s", funcName);

							const VDShaderVariable *varWidth = mpEffect->GetAnnotationByName(texVar->mDesc.mAnnotations, nameWidth);
							const VDShaderVariable *varHeight = mpEffect->GetAnnotationByName(texVar->mDesc.mAnnotations, nameHeight);

							if (!varWidth || !varHeight)
								ff->Except("Both width and height annotations must be specified for a shader created texture.");

							int texw = mpEffect->GetInt(varWidth);
							int texh = mpEffect->GetInt(varHeight);

							if (texw < 1 || texw > 4096 || texh < 1 || texh > 4096 || (texw & (texw-1)) || (texh & (texh-1)))
								ff->Except("Texture width and height must be powers of two between 1-4096.");

							vdrefptr<VDShaderFilterTexture> filterTex(new VDShaderFilterTexture);
							filterTex->Init(texw, texh);

							// compile texture shader on the spot
							vdautoptr<IVDShaderExecutionEngine> ee;
							switch(execMode) {
								case VDShaderFilterConfig::kModeInterp:
								default:
									ee = new VDShaderInterpreter;
									break;

#ifdef _M_IX86
								case VDShaderFilterConfig::kModeScalarJIT:
									ee = new VDShaderScalarJIT;
									break;

								case VDShaderFilterConfig::kModeSSE2JIT:
									ee = new VDShaderSSE2JIT;
									break;
#endif
							}

							ee->LoadSamplers(NULL, 0);

							VDShaderParserBufOutput out;
							const VDShaderBindingElement kElements[]={
								0,
								nsVDShaderParser::kTypeFloat2,
								nameTable.GetName("TEXCOORD0", 9)
							};

							const float dudx = 1.0f / (float)(int)texw;
							const float dvdy = 1.0f / (float)(int)texh;
							const float u0 = 0.5f * dudx;
							const float v0 = 0.5f * dvdy;

							VDShaderInterpreterInterpolator uv0;
							uv0.u[0] = u0;
							uv0.u[1] = v0;
							uv0.u[2] = 0;
							uv0.u[3] = 0;
							uv0.dudx[0] = dudx;
							uv0.dudx[1] = 0;
							uv0.dudx[2] = 0;
							uv0.dudx[3] = 0;

							if (!ee->LoadShader(*texFunc, kElements, 1, NULL, 0, &uv0, 1, &out, false))
								ff->Except("Shader failed to load:\n%s", out.c_str());

							const ptrdiff_t dstPitch = filterTex->mPitch;
							uint32 *dst0 = filterTex->mpBits;
							uint32 *dst = vdptroffset(dst0, dstPitch) + 1;

							for(int y=0; y<texh; ++y) {
								uv0.u[0] = u0;

								ee->Execute(dst, &uv0, texw);

								uv0.u[1] += dvdy;
								vdptrstep(dst, dstPitch);
							}

							// apply borders
							if (clampU) {
								VDMemcpyRect(
									vdptroffset(dst0, dstPitch),
									dstPitch,
									vdptroffset(dst0, dstPitch) + 1,
									dstPitch,
									sizeof(uint32),
									texh);

								VDMemcpyRect(
									vdptroffset(dst0, dstPitch) + texw + 1,
									dstPitch,
									vdptroffset(dst0, dstPitch) + texw,
									dstPitch,
									sizeof(uint32),
									texh);
							} else {
								VDMemcpyRect(
									vdptroffset(dst0, dstPitch),
									dstPitch,
									vdptroffset(dst0, dstPitch) + texw,
									dstPitch,
									sizeof(uint32),
									texh);

								VDMemcpyRect(
									vdptroffset(dst0, dstPitch) + texw + 1,
									dstPitch,
									vdptroffset(dst0, dstPitch) + 1,
									dstPitch,
									sizeof(uint32),
									texh);
							}

							if (clampV) {
								memcpy(dst0, vdptroffset(dst0, dstPitch), (texw + 2)*sizeof(uint32));
								memcpy(vdptroffset(dst0, dstPitch * (texh + 1)), vdptroffset(dst0, dstPitch * texh), (texw + 2)*sizeof(uint32));
							} else {
								memcpy(dst0, vdptroffset(dst0, dstPitch * texh), (texw + 2)*sizeof(uint32));
								memcpy(vdptroffset(dst0, dstPitch * (texh + 1)), vdptroffset(dst0, dstPitch), (texw + 2)*sizeof(uint32));
							}

							execPass->SetSampler(i, filterTex, clampU, clampV, useBilinear);
							continue;
						}
					}
				}
			}

			// search for an existing frame fetch
			const int frameFetchIndex = AddFrameFetch(srcIndex, srcFrameOffset);

			// search for an existing sampler
			execPass->SetSampler(i, frameFetchIndex, *srcStreams[srcIndex], clampU, clampV, useBilinear);
		}

		if (!execPass->mpSamplers[0]) {
			const int frameFetchIndex0 = AddFrameFetch(0, 0);

			execPass->SetSampler(0, frameFetchIndex0, *srcStream0, true, true, true);
		}

		//////////// process constants

		execPass->mParamDisplay.clear();
		execPass->mParamConstants.clear();
		execPass->mParamVarLookup.clear();

		// reserve first two constants for MPC compatibility
		int displaySize = 2;

		// add bindings for the first two constants if they're not bound
		const VDShaderInterpConstant c0 = { 0, 0 };
		const VDShaderInterpConstant c1 = { 1, 16 };

		if (!func->IsConstantRegisterBound(0)) {
			execPass->mParamConstants.push_back(c0);
			execPass->mParamVarLookup.push_back(NULL);
		}

		if (!func->IsConstantRegisterBound(1)) {
			execPass->mParamConstants.push_back(c1);
			execPass->mParamVarLookup.push_back(NULL);
		}

		// process custom parameters
		for(int varIdx = 0; varIdx < numVars; ++varIdx) {
			const VDShaderVariable& var = *mpEffect->GetVariableByIndex(varIdx);

			if (var.mDesc.mbExtern)
				continue;

			switch(var.mDesc.mType) {
				case nsVDShaderParser::kTypeBool:
				case nsVDShaderParser::kTypeBool2:
				case nsVDShaderParser::kTypeBool3:
				case nsVDShaderParser::kTypeBool4:
				case nsVDShaderParser::kTypeFloat:
				case nsVDShaderParser::kTypeFloat2:
				case nsVDShaderParser::kTypeFloat3:
				case nsVDShaderParser::kTypeFloat4:
					break;

				default:
					continue;
			}

			// try to match against a saved parameter
			if (!var.mDesc.mNameToken)
				continue;

			// try to find a binding
			int constantIndex = -1;

			for(VDShaderParserFunction::ConstantBindings::const_iterator it(func->mConstantBindings.begin()), itEnd(func->mConstantBindings.end()); it != itEnd; ++it) {
				const VDShaderConstantBinding& binding = *it;

				if (binding.mNameToken == var.mDesc.mNameToken) {
					constantIndex = binding.mRegister - kVDShaderRegC0;
					break;
				}
			}

			if (constantIndex < 0)
				continue;

			VDShaderInterpConstant ic;

			ic.mConstantIndex = constantIndex;
			ic.mOffset = displaySize * sizeof(VDShaderConstant);
			++displaySize;

			execPass->mParamDisplay.push_back(mpEffect->GetVector(&var));
			execPass->mParamConstants.push_back(ic);
			execPass->mParamVarLookup.push_back(&var);
		}

		execPass->mParamDisplay.resize(displaySize);

		// compute interpolators
		const uint32 dstw = fa->dst.w;
		const uint32 dsth = fa->dst.h;
		const float dudx = 1.0f / (float)(int)dstw;
		const float dvdy = 1.0f / (float)(int)dsth;
		const float u0 = 0.5f * dudx;
		const float v0 = 0.5f * dvdy;

		mTexCoord0.u[0] = u0;
		mTexCoord0.u[1] = v0;
		mTexCoord0.u[2] = 0;
		mTexCoord0.u[3] = 0;
		mTexCoord0.dudx[0] = dudx;
		mTexCoord0.dudx[1] = 0;
		mTexCoord0.dudx[2] = 0;
		mTexCoord0.dudx[3] = 0;

		// compile shader

		switch(execMode) {
			case VDShaderFilterConfig::kModeInterp:
			default:
				execPass->mpEngine = new VDShaderInterpreter;
				break;

#ifdef _M_IX86
			case VDShaderFilterConfig::kModeScalarJIT:
				execPass->mpEngine = new VDShaderScalarJIT;
				break;

			case VDShaderFilterConfig::kModeSSE2JIT:
				execPass->mpEngine = new VDShaderSSE2JIT;
				break;
#endif
		}

		execPass->mpEngine->LoadSamplers(execPass->mpSamplers, 16);

		VDShaderParserBufOutput out;
		if (!execPass->mpEngine->LoadShader(*func, kElements, 1, execPass->mParamConstants.data(), execPass->mParamConstants.size(), &mTexCoord0, 1, &out, false))
			ff->Except("Shader failed to load:\n%s", out.c_str());
	}

	mbConstantsDirty = true;
}

void VDShaderFilter::End() {
	while(!mPasses.empty()) {
		Pass *pass = mPasses.back();
		mPasses.pop_back();

		if (pass->mpEngine)
			delete pass->mpEngine;

		delete pass;
	}

	while(!mSourceChannels.empty()) {
		delete mSourceChannels.back();
		mSourceChannels.pop_back();
	}

	mFrameFetches.clear();

	mpEffect = NULL;
}

void VDShaderFilter::Run() {
	uint32 srcw = fa->src.w;
	uint32 srch = fa->src.h;

	// upload source frames to textures
	VDXFBitmap *const srcStream0 = &fa->src;
	const VDXFBitmap *const *srcStreams = (g_VDFilterAPIVersion >= 16) ? fa->mpSourceStreams : &srcStream0;
	const VDXFBitmap *const *srcFrames = (g_VDFilterAPIVersion >= 14) ? fa->mpSourceFrames : &srcStream0;
	const uint32 numSrcFrames = (g_VDFilterAPIVersion >= 14) ? fa->mSourceFrameCount : 1;

	const uint32 dstw = fa->dst.w;
	const uint32 dsth = fa->dst.h;
	uint32 *dst = (uint32 *)fa->dst.data;
	float dvdy = 1.0f / (float)(int)dsth;

	VDShaderInterpreterInterpolator texcoord0 = mTexCoord0;

	vdptrstep(dst, fa->dst.pitch * (fa->dst.h - 1));

	Pass *execPass = NULL;
	sint32 frameNumber = fa->pfsi ? fa->pfsi->lCurrentFrame : 0;
	const int fieldId = fa->pfsi && (fa->pfsi->lCurrentFrame & 1) ? 2 : 1;

	for(Passes::const_iterator it(mPasses.begin()), itEnd(mPasses.end()); it != itEnd; ++it) {
		Pass *execPassCur = *it;

		if (execPassCur->mFieldMask & fieldId) {
			execPass = execPassCur;
			break;
		}
	}

	if (!execPass)
		return;

	// update sampler pointers
	uint32 boundSamplers = 0;

	for(int pass=0; pass<2; ++pass) {
		for(int i=0; i<16; ++i) {
			VDShaderFilterSampler& sampler = execPass->mSamplers[i];

			if (!sampler.IsDynamic())
				continue;

			if (boundSamplers & (1 << i))
				continue;

			const FrameFetch& frameFetch = mFrameFetches[sampler.mFetchIndex];
			SourceChannel& sc = *mSourceChannels[frameFetch.mSrcIndex];

			VDShaderFilterFrameKey frameKey;
			frameKey.mCurrentFrame = std::max<sint64>(0, frameNumber + frameFetch.mFrameOffset);
			frameKey.mbClampU = sampler.mSampler.mbClampU;
			frameKey.mbClampV = sampler.mSampler.mbClampV;

			vdrefptr<VDShaderFilterTexture> tex;

			if (pass == 0) {
				sc.mAllocator.GetBuffer(frameKey, ~tex);
			} else {
				if (!sc.mAllocator.AllocBuffer(frameKey, ~tex))
					ff->ExceptOutOfMemory();

				const VDXFBitmap& srcFrame = *srcFrames[(uint32)sampler.mFetchIndex < numSrcFrames ? sampler.mFetchIndex : 0];
				UploadTexture(*tex, srcFrame, sampler.mSampler.mbClampU, sampler.mSampler.mbClampV);
			}

			if (tex) {
				sampler.BindToTexture(tex);
				boundSamplers |= (1 << i);
			}

		}
	}

	VDShaderConstant& c0 = execPass->mParamDisplay[0];
	c0.v[0]		= (float)srcw;			// width
	c0.v[1]		= (float)srch;			// height
	c0.v[2]		= 0;					// counter
	c0.v[3]		= 0;					// clock

	VDShaderConstant& c1 = execPass->mParamDisplay[1];
	c1.v[0]		= 1.0f / (float)srcw;	// invwidth
	c1.v[1]		= 1.0f / (float)srch;	// invheight
	c1.v[2]		= 0;					// dummy0
	c1.v[3]		= 0;					// dummy1

	if (fa->pfsi) {
		c0.v[2] = (float)fa->pfsi->lCurrentFrame;
		c0.v[3] = (float)fa->pfsi->lDestFrameMS / 1000.0f;
	}

	if (mbConstantsDirty) {
		int n = execPass->mParamVarLookup.size();

		for(int i=0; i<n; ++i) {
			const VDShaderVariable *var = execPass->mParamVarLookup[i];
			if (!var)
				continue;

			const VDShaderInterpConstant& ic = execPass->mParamConstants[i];
			VDShaderConstant& c = *(VDShaderConstant *)((char *)execPass->mParamDisplay.data() + ic.mOffset);
			c = mpEffect->GetVector(var);
		}

		mbConstantsDirty = false;
	}

	const float u0 = mTexCoord0.u[0];
	if (execPass->mpEngine) {
		execPass->mpEngine->LoadConstants(execPass->mParamDisplay.data());

		for(uint32 y=0; y<dsth; ++y) {
			texcoord0.u[0] = u0;

			execPass->mpEngine->Execute(dst, &texcoord0, dstw);

			texcoord0.u[1] += dvdy;
			vdptrstep(dst, -fa->dst.pitch);
		}
	}

	for(int i=0; i<16; ++i) {
		VDShaderFilterSampler& sampler = execPass->mSamplers[i];

		if (sampler.IsDynamic())
			sampler.Unbind();
	}
}

int VDShaderFilter::Serialize(char *buf, int maxbuf) {
	return 0;
}

int VDShaderFilter::Deserialize(const char *buf, int maxbuf) {
	return 0;
}

void VDShaderFilter::GetSettingString(char *buf, int maxlen) {
	if (mpShaderDef) {
	} else {
		_snprintf(buf, maxlen, " (%s)", mConfig.mFilename.c_str());
	}
}

void VDShaderFilter::GetScriptString(char *buf, int maxlen) {
	VDStringA params;
	for(VDShaderFilterConfig::ParamValues::const_iterator it(mConfig.mParamValues.begin()), itEnd(mConfig.mParamValues.end()); it != itEnd; ++it) {
		const VDShaderFilterParameterConfig& pc = *it;
		const char *name = pc.mName.c_str();

		switch(pc.mType) {
			case kVDFXParamType_Float:
				params.append_sprintf("float %s=%g;", name, pc.mValue.v[0]);
				break;
			case kVDFXParamType_Float2:
				params.append_sprintf("float2 %s=%g,%g;", name, pc.mValue.v[0], pc.mValue.v[1]);
				break;
			case kVDFXParamType_Float3:
				params.append_sprintf("float3 %s=%g,%g,%g;", name, pc.mValue.v[0], pc.mValue.v[1], pc.mValue.v[2]);
				break;
			case kVDFXParamType_Float4:
				params.append_sprintf("float4 %s=%g,%g,%g,%g;", name, pc.mValue.v[0], pc.mValue.v[1], pc.mValue.v[2], pc.mValue.v[3]);
				break;
		}
	}

	VDStringA tmp;
	const char *s = mConfig.mFilename.c_str();
	while(char c = *s++) {
		if ((unsigned char)((c & 0x7f) - 0x20) >= 0x60) {
			tmp += VDStringA().sprintf("\\x%02x", (unsigned char)c);
			continue;
		}

		if (c == '\\' || c == '"')
			tmp.push_back('\\');
		tmp.push_back(c);
	}

	_snprintf(buf, maxlen, "Config(%d, %d, %d, \"%s\", %d, \"%s\")", mConfig.mNewWidth, mConfig.mNewHeight, mConfig.mbKeepSize, tmp.c_str(), mConfig.mExecutionMode, params.c_str());
}

bool VDShaderFilter::Configure(VDXHWND hwnd) {
	if (mpShaderDef) {
		ClearUIBindings();

		try {
			GetEffect(~mConfig.mpEffect);
		} catch(const MyError&) {
			MyError("Unable to parse effect file: %s", mConfig.mFilename.c_str()).post((struct HWND__ *)hwnd, "VDShaderError");
			return false;
		}

		UpdateUIBindings(*mConfig.mpEffect);

		VDShaderFilterShowConfigDialog(hwnd, mConfig, fa->ifp, this);

		ClearUIBindings();
		mConfig.mpEffect = NULL;
		return true;
	}

	VDShaderFilterConfig old(mConfig);

	bool success = VDOpenShaderEditor(hwnd, mConfig, fa->ifp, g_VDFilterAPIVersion >= 11 ? fa->ifp2 : NULL, this);

	mConfig.mpEffect = NULL;
	ClearUIBindings();

	if (!success) {
		mConfig = old;
		return false;
	}

	return true;
}

bool VDShaderFilter::Prefetch2(sint64 frame, IVDXVideoPrefetcher *prefetcher) {
	for(FrameFetches::const_iterator it(mFrameFetches.begin()), itEnd(mFrameFetches.end());
		it != itEnd;
		++it)
	{
		const FrameFetch& fetch = *it;

		prefetcher->PrefetchFrame(fetch.mSrcIndex, frame + fetch.mFrameOffset, 0);
	}

	return true;
}

void VDShaderFilter::GetParamBindings(vdfastvector<VDFXUIBinding>& paramBindings) {
	paramBindings = mParamUIBindings;
}

void VDShaderFilter::SetEffectParameterLive(const char *name, const VDShaderConstant& v) {
	if (!mpEffect)
		return;

	int nameToken = mpEffect->NameTable().GetName(name, strlen(name));
	if (!nameToken)
		return;

	const VDShaderVariable *var = mpEffect->GetVariableByName(nameToken);
	if (!var)
		return;

	mpEffect->SetVector(var, v);
	mbConstantsDirty = true;

	VDShaderFilterParameterConfig *p = NULL;
	for(VDShaderFilterConfig::ParamValues::iterator it(mConfig.mParamValues.begin()), itEnd(mConfig.mParamValues.end()); it != itEnd; ++it) {
		if (it->mName == name) {
			p = &*it;
			break;
		}
	}

	if (!p) {
		mConfig.mParamValues.resize(mConfig.mParamValues.size() + 1);
		p = &mConfig.mParamValues.back();
		p->mName = name;
	}

	switch(var->mDesc.mType) {
		case nsVDShaderParser::kTypeBool:
		case nsVDShaderParser::kTypeFloat:
			p->mType = kVDFXParamType_Float;
			break;
		case nsVDShaderParser::kTypeBool2:
		case nsVDShaderParser::kTypeFloat2:
			p->mType = kVDFXParamType_Float2;
			break;
		case nsVDShaderParser::kTypeBool3:
		case nsVDShaderParser::kTypeFloat3:
			p->mType = kVDFXParamType_Float3;
			break;
		case nsVDShaderParser::kTypeBool4:
		case nsVDShaderParser::kTypeFloat4:
			p->mType = kVDFXParamType_Float4;
			break;
	}

	p->mValue = v;
}

void VDShaderFilter::ScriptConfig(IVDXScriptInterpreter *, const VDXScriptValue *argv, int argc) {
	mConfig.mNewWidth = argv[0].asInt();
	mConfig.mNewHeight = argv[1].asInt();
	mConfig.mbKeepSize = argv[2].asInt() != 0;

	if (!mpShaderDef)
		mConfig.mFilename = *argv[3].asString();

	unsigned mode = argv[4].asInt();

	if (mode >= (unsigned)VDShaderFilterConfig::kModeCount)
		mode = VDShaderFilterConfig::kModeDefault;

	mConfig.mExecutionMode = mode;

	const char *configStr = *argv[5].asString();

	VDStringRefA s(configStr);
	VDStringRefA paramStr;

	mConfig.mParamValues.clear();

	while(s.split(';', paramStr)) {
		VDStringRefA nameStr;
		if (!paramStr.split('=', nameStr))
			continue;

		VDStringRefA typeStr;
		if (!nameStr.split(' ', typeStr))
			continue;

		nameStr = nameStr.trim(" \t\n\r");
		typeStr = typeStr.trim(" \t\n\r");
		paramStr = paramStr.trim(" \t\n\r");

		VDShaderFilterParameterConfig pc;
		pc.mName = nameStr;
		pc.mValue = 0;

		VDStringA valueStr(paramStr);
		if (typeStr == "float") {
			pc.mType = kVDFXParamType_Float;

			sscanf(valueStr.c_str(), "%g", &pc.mValue.v[0]);
		} else if (typeStr == "float2") {
			pc.mType = kVDFXParamType_Float2;
			sscanf(valueStr.c_str(), "%g, %g", &pc.mValue.v[0], &pc.mValue.v[1]);
		} else if (typeStr == "float3") {
			pc.mType = kVDFXParamType_Float3;
			sscanf(valueStr.c_str(), "%g, %g, %g", &pc.mValue.v[0], &pc.mValue.v[1], &pc.mValue.v[2]);
		} else if (typeStr == "float4") {
			pc.mType = kVDFXParamType_Float4;
			sscanf(valueStr.c_str(), "%g, %g, %g, %g", &pc.mValue.v[0], &pc.mValue.v[1], &pc.mValue.v[2], &pc.mValue.v[3]);
		} else
			continue;

		mConfig.mParamValues.push_back(pc);
	}
}

void VDShaderFilter::GetEffect(VDShaderEffect **ppEffect) {
	if (mConfig.mpEffect) {
		*ppEffect = mConfig.mpEffect;
		mConfig.mpEffect->AddRef();
		return;
	}

	VDStringA buf;

	if (mConfig.mpSource)
		mConfig.mpSource->ReadSource(buf);
	else {
		FILE *f = fopen(mConfig.mFilename.c_str(), "rb");
		if (!f)
			throw MyError("Unable to open shader file:\n%s", mConfig.mFilename.c_str());

		fseek(f, 0, SEEK_END);
		int len = ftell(f);
		buf.resize(len);
		fseek(f, 0, SEEK_SET);
		fread(&buf[0], len, 1, f);
		fclose(f);
	}

	VDShaderParser parser;
	VDShaderParserBufOutput out;

	if (!parser.Parse(buf.data(), buf.size(), mConfig.mFilename.c_str(), mConfig.mbOptimizationEnabled,  &out, ppEffect)) {
		out.End();
		throw MyError("Shader failed to compile:\n%s", out.c_str());
	}
}

void VDShaderFilter::ClearUIBindings() {
	mParamUIBindings.clear();
}

void VDShaderFilter::UpdateUIBindings(VDShaderEffect& eff) {
	const VDShaderNameTable& nameTable = eff.NameTable();
	int hTunable		= nameTable.GetName("vd_tunable", 10);
	int hTunableSteps	= nameTable.GetName("vd_tunablesteps", 15);
	int hTunableMin		= nameTable.GetName("vd_tunablemin", 13);
	int hTunableMax		= nameTable.GetName("vd_tunablemax", 13);

	int varCount = eff.GetVariableCount();

	mParamUIBindings.clear();
	for(int varIdx=0; varIdx<varCount; ++varIdx) {
		const VDShaderVariable& var = *eff.GetVariableByIndex(varIdx);
		const VDShaderAnnotations& annos = var.mDesc.mAnnotations;
	
		// check for a tunable parameter
		const VDShaderVariable *pAnnoTunable = eff.GetAnnotationByName(annos, hTunable);
		if (!eff.GetBool(pAnnoTunable))
			continue;

		VDFXUIBinding uibinding;
		uibinding.mpVar = &var;
		uibinding.mpName = eff.NameTable().GetName(var.mDesc.mNameToken);
		uibinding.mParamType = kVDFXParamType_None;
		uibinding.mSteps = 0;
		memset(&uibinding.mDefaultValue, 0, sizeof uibinding.mDefaultValue);

		const VDShaderVariable *hAnnoTunableSteps = eff.GetAnnotationByName(annos, hTunableSteps);
		if (hAnnoTunableSteps) {
			int steps = eff.GetInt(hAnnoTunableSteps);
			if (steps >= 0)
				uibinding.mSteps = steps;
		}

		const VDShaderVariable *hAnnoMinValue = eff.GetAnnotationByName(annos, hTunableMin);
		const VDShaderVariable *hAnnoMaxValue = eff.GetAnnotationByName(annos, hTunableMax);

		switch(var.mDesc.mType) {
			case nsVDShaderParser::kTypeFloat:
				uibinding.mMinValue.v[0] = FLT_MIN;
				uibinding.mMinValue.v[1] = FLT_MIN;
				uibinding.mMinValue.v[2] = FLT_MIN;
				uibinding.mMinValue.v[3] = FLT_MIN;
				uibinding.mMaxValue.v[0] = FLT_MAX;
				uibinding.mMaxValue.v[1] = FLT_MAX;
				uibinding.mMaxValue.v[2] = FLT_MAX;
				uibinding.mMaxValue.v[3] = FLT_MAX;

				switch(var.mDesc.mType) {
					case nsVDShaderParser::kTypeFloat:
						{
							uibinding.mDefaultValue.v[0] = eff.GetFloat(&var);
							uibinding.mParamType = kVDFXParamType_Float;

							if (hAnnoMinValue)
								uibinding.mMinValue.v[0] = eff.GetFloat(hAnnoMinValue);

							if (hAnnoMaxValue)
								uibinding.mMaxValue.v[0] = eff.GetFloat(hAnnoMaxValue);
						}
						break;
					case nsVDShaderParser::kTypeFloat2:
					case nsVDShaderParser::kTypeFloat3:
					case nsVDShaderParser::kTypeFloat4:
						{
							uibinding.mDefaultValue = eff.GetVector(&var);

							if (hAnnoMinValue)
								uibinding.mMinValue = eff.GetVector(hAnnoMinValue);

							if (hAnnoMaxValue)
								uibinding.mMaxValue = eff.GetVector(hAnnoMaxValue);

							switch(var.mDesc.mType) {
								case nsVDShaderParser::kTypeFloat:
									uibinding.mParamType = kVDFXParamType_Float;
									break;
								case nsVDShaderParser::kTypeFloat2:
									uibinding.mParamType = kVDFXParamType_Float2;
									break;
								case nsVDShaderParser::kTypeFloat3:
									uibinding.mParamType = kVDFXParamType_Float3;
									break;
								case nsVDShaderParser::kTypeFloat4:
									uibinding.mParamType = kVDFXParamType_Float4;
									break;
							}
						}
						break;
				}
				break;
		}

		if (uibinding.mParamType)
			mParamUIBindings.push_back(uibinding);
	}
}

int VDShaderFilter::AddFrameFetch(int srcIndex, int srcFrameOffset) {
	int frameFetchIndex = -1;

	for(FrameFetches::const_iterator it(mFrameFetches.begin()), itEnd(mFrameFetches.end());
		it != itEnd;
		++it)
	{
		const FrameFetch& fetch = *it;

		if (fetch.mSrcIndex == srcIndex && fetch.mFrameOffset == srcFrameOffset) {
			frameFetchIndex = it - mFrameFetches.begin();
			break;
		}
	}

	if (frameFetchIndex < 0) {
		frameFetchIndex = (int)mFrameFetches.size();

		FrameFetch& fetch = mFrameFetches.push_back();
		fetch.mSrcIndex = srcIndex;
		fetch.mFrameOffset = srcFrameOffset;
	}

	return frameFetchIndex;
}

void VDShaderFilter::UploadTexture(const Texture& dst, const VDXFBitmap& src, bool clampU, bool clampV) {
	if (g_VDFilterAPIVersion >= 11) {
		const VDXPixmap& px = *src.mpPixmap;

		UploadTexture(dst, px.data, px.pitch, px.w, px.h, clampU, clampV);
	} else {
		UploadTexture(dst, (const char *)src.data + src.pitch * (src.h - 1), -(ptrdiff_t)src.pitch, src.w, src.h, clampU, clampV);
	}
}

void VDShaderFilter::UploadTexture(const Texture& texture, const void *src0, ptrdiff_t srcPitch, uint32 w, uint32 h, bool clampU, bool clampV) {
	const ptrdiff_t dstPitch = texture.mPitch;

	const uint32 *src = (const uint32 *)src0;
	uint32 *dst = vdptroffset(texture.mpBits, dstPitch) + 1;
	for(uint32 y=0; y<h; ++y) {
		memcpy(dst, src, w*sizeof(uint32));

		if (clampU) {
			dst[-1] = dst[0];
			dst[w] = dst[w-1];
		} else {
			dst[-1] = dst[w-1];
			dst[w] = dst[0];
		}

		vdptrstep(src, srcPitch);
		vdptrstep(dst, dstPitch);
	}

	if (clampV) {
		memcpy(texture.mpBits, vdptroffset(texture.mpBits, dstPitch), (w + 2)*sizeof(uint32));
		memcpy(vdptroffset(texture.mpBits, dstPitch*(h + 1)), vdptroffset(texture.mpBits, dstPitch*h), (w + 2)*sizeof(uint32));
	} else {
		memcpy(texture.mpBits, vdptroffset(texture.mpBits, dstPitch*h), (w + 2)*sizeof(uint32));
		memcpy(vdptroffset(texture.mpBits, dstPitch*(h + 1)), vdptroffset(texture.mpBits, dstPitch), (w + 2)*sizeof(uint32));
	}
}

///////////////////////////////////////////////////////////////////////////
VDXVideoFilterDefinition<VDShaderFilter> g_customFilterDef("Avery Lee", "vdshader", "Applies a custom pixel shader to incoming video. (Version 1.2)");

void VDRegisterShaderEditor();
void VDUnregisterShaderEditor();

namespace {
	std::vector<VDShaderFilterDefinition *> g_userFilters;
}

void ScanForFilters(const wchar_t *path, VDXFilterModule *fm, const VDXFilterFunctions *ff) {
	VDDirectoryIterator it(VDMakePath(path, L"*.fx").c_str());

	while(it.Next()) {
		const VDStringW shaderPath(it.GetFullPath());

		try {
			vdautoptr<VDShaderFilterDefinition> ufd(new VDShaderFilterDefinition);
			ufd->mName = VDTextWToA(it.GetName());
			ufd->mFilterDef = g_customFilterDef;

			VDTextInputFile tif(shaderPath.c_str());
			while(const char *s = tif.GetNextLine()) {
				if (strncmp(s, "//", 2))
					break;

				s += 2;
				s = strskipspace(s);

				const char *tagstart = s;

				if (!isalpha((unsigned char)*s++))
					continue;

				while(isalnum((unsigned char)*s) || *s == '_')
					++s;

				VDStringSpanA tag(tagstart, s);

				if (*s++ != ':')
					continue;

				s = strskipspace(s);

				VDStringSpanA data(s);

				data = data.trim(" \n\r\t\v");

				if (tag.comparei("Author") == 0) {
					ufd->mAuthor = data;
				} else if (tag.comparei("Name") == 0) {
					ufd->mName = data;
					ufd->mName = data + " (VDFX)";
				} else if (tag.comparei("Description") == 0) {
					ufd->mDescription = data;
				} else if (tag.comparei("Minimum sources") == 0) {
					int n = atoi(VDStringA(data).c_str());

					if (n) {
						--n;

						ufd->mFilterDef.mSourceCountLowMinus1 = n;
						if (ufd->mFilterDef.mSourceCountHighMinus1 < n)
							ufd->mFilterDef.mSourceCountHighMinus1 = n;
					}
				} else if (tag.comparei("Maximum sources") == 0) {
					int n = atoi(VDStringA(data).c_str());

					if (n) {
						--n;

						ufd->mFilterDef.mSourceCountHighMinus1 = n;
						if (ufd->mFilterDef.mSourceCountLowMinus1 > n)
							ufd->mFilterDef.mSourceCountLowMinus1 = n;
					}
				}
			}

			// Check if we have more than one minimum input and we have a pre-V16 host.
			// If so, we can't run this filter and should skip it.
			if (g_VDFilterAPIVersion < 16 && ufd->mFilterDef.mSourceCountLowMinus1 > 0)
				continue;

			ufd->mDescription += "\n(VDShader-based filter)";
			
			ufd->mShaderPath = shaderPath.c_str();

			ufd->mFilterDef.name	= ufd->mName.c_str();
			ufd->mFilterDef.maker	= ufd->mAuthor.c_str();
			ufd->mFilterDef.desc	= ufd->mDescription.c_str();
			ufd->mFilterDef.private_data = ufd;

			ufd->mpActiveFilterDef = ff->addFilter(fm, &ufd->mFilterDef, sizeof ufd->mFilterDef);

			g_userFilters.push_back(ufd.release());
		} catch(const MyError&) {
		}
	}
}

extern "C" __declspec(dllexport) int __cdecl VirtualdubFilterModuleInit2(struct VDXFilterModule *fm, const VDXFilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
	VDRegistryAppKey::setDefaultKey("Software\\Freeware\\VDShader\\");
	VDRegisterShaderEditor();

	ff->addFilter(fm, &g_customFilterDef, sizeof(VDXFilterDefinition));

	const VDStringW localPath(VDGetLocalModulePath());
	const VDStringW shadersPath(VDMakePath(localPath.c_str(), L"FXFilters"));

	ScanForFilters(shadersPath.c_str(), fm, ff);

	g_VDFilterAPIVersion = vdfd_ver;
	VDXVideoFilter::SetAPIVersion(vdfd_ver);

	vdfd_ver		= 16;
	vdfd_compat		= 9;	// we need this version for copy constructor support

	return 0;
}

extern "C" __declspec(dllexport) void __cdecl VirtualdubFilterModuleDeinit(struct VDXFilterModule *fm, const VDXFilterFunctions *ff) {
	while(!g_userFilters.empty()) {
		vdautoptr<VDShaderFilterDefinition> ufd(g_userFilters.back());
		g_userFilters.pop_back();

		ff->removeFilter(ufd->mpActiveFilterDef);
	}

	ff->removeFilter(&g_customFilterDef);
	VDUnregisterShaderEditor();
}
