//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include "parser.h"
#include "intrinsics.h"
#include "vecarith.h"

using namespace nsVDShaderParser;
using namespace nsVDShaderArith;

bool VDShaderParserConvertVarToType(IVDShaderErrorOutput& lexer, VDShaderParserFunction *func, VDShaderParserVariable& var, int dstType);

bool VDShaderCompileIntrinsic_dot(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& vB = args[1];
	VDShaderParserVariable& vA = args[0];

	int rA = 0;
	
	if (vA.mbIsConst != vB.mbIsConst) {
		func.PromoteToConstantRegisters(vA);
		func.PromoteToConstantRegisters(vB);
	}

	if (vA.mbIsConst) {
		switch(vA.mType) {
			case kTypeFloat2:
				vA.m[0] = dp2(vA.m[0], vB.m[0]);
				break;
			case kTypeFloat3:
				vA.m[0] = dp3(vA.m[0], vB.m[0]);
				break;
			case kTypeFloat4:
				vA.m[0] = dp4(vA.m[0], vB.m[0]);
				break;
		}
		vA.m[0].v[1] = 0.0f;
		vA.m[0].v[2] = 0.0f;
		vA.m[0].v[3] = 0.0f;
	} else {
		rA = func.AllocateRegisters(1);

		switch(vA.mType) {
			case kTypeFloat2:
				func.dp2add(rA, 0x1, vA.mRegister, vB.mRegister, func.AddConstantZero(), 0);
				break;
			case kTypeFloat3:
				func.dp3(rA, 0x1, vA.mRegister, vB.mRegister);
				break;
			case kTypeFloat4:
				func.dp4(rA, 0x1, vA.mRegister, vB.mRegister);
				break;
		}
	}

	vA.mType = kTypeFloat;
	vA.mRegister = rA;
	return true;
}

bool VDShaderCompileIntrinsic_length(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& v = args[0];

	if (v.mbIsConst) {
		switch(v.mType) {
			case kTypeFloat2:
				v.m[0].v[0] = sqrtf(dp2(v.m[0], v.m[0]).v[0]);
				break;
			case kTypeFloat3:
				v.m[0].v[0] = sqrtf(dp3(v.m[0], v.m[0]).v[0]);
				break;
			case kTypeFloat4:
				v.m[0].v[0] = sqrtf(dp4(v.m[0], v.m[0]).v[0]);
				break;
		}
		v.m[0].v[1] = 0.0f;
		v.m[0].v[2] = 0.0f;
		v.m[0].v[3] = 0.0f;
	} else {
		int rA = func.AllocateRegisters(3);
		int rB = rA + 1;
		int rC = rB + 1;

		switch(v.mType) {
			case kTypeFloat2:
				func.dp2add(rA, 0x1, v.mRegister, v.mRegister, func.AddConstantZero(), 0);
				break;
			case kTypeFloat3:
				func.dp3(rA, 0x1, v.mRegister, v.mRegister);
				break;
			case kTypeFloat4:
				func.dp4(rA, 0x1, v.mRegister, v.mRegister);
				break;
		}

		func.rsq(rB, 0x1, rA);
		func.rcp(rC, 0x1, rB);
		v.mRegister = rC;
	}

	v.mType = kTypeFloat;
	return true;
}

bool VDShaderCompileIntrinsic_normalize(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& v = args[0];

	if (v.mbIsConst) {
		float sum = 0.0f;

		switch(v.mType) {
			case kTypeFloat4:
				sum = dp2(v.m[0], v.m[0]).v[0];
				break;
			case kTypeFloat3:
				sum = dp3(v.m[0], v.m[0]).v[0];
				break;
			case kTypeFloat2:
				sum = dp4(v.m[0], v.m[0]).v[0];
				break;
		}

		if (fabsf(sum) < 1e-30f)
			return lexer.SetError("Attempted normalize() on zero-magnitude vector");

		float invscale = 1.0f / sqrtf(sum);

		v.m[0].v[0] *= invscale;
		v.m[0].v[1] *= invscale;
		v.m[0].v[2] *= invscale;
		v.m[0].v[3] *= invscale;
	} else {
		int rA = func.AllocateRegisters(3);

		func.nrm(rA, kColWriteMaskForType[v.mType], v.mRegister);
		v.mRegister = rA;
	}

	return true;
}

bool VDShaderCompileIntrinsic_mul_v2_m2x2(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& v = args[0];
	VDShaderParserVariable& m = args[1];

	if (!v.mbIsConst || !m.mbIsConst) {
		func.PromoteToConstantRegisters(v);
		func.PromoteToConstantRegisters(m);
	}

	if (v.mbIsConst) {
		float x = v.m[0].v[0];
		float y = v.m[0].v[1];

		v.m[0].v[0] = m.m[0].v[0]*x + m.m[1].v[0]*y;
		v.m[0].v[1] = m.m[0].v[1]*x + m.m[1].v[1]*y;
	} else {
		int rA = func.AllocateRegisters(2);

		func.mul_swizzle(rA+0, 0x3, v.mRegister, 0x00, m.mRegister + 0, 0xE4);
		func.mad_swizzle(rA+1, 0x3, v.mRegister, 0x55, m.mRegister + 1, 0xE4, rA+0, 0xE4);
		v.mRegister = rA + 1;
	}

	return true;
}

bool VDShaderCompileIntrinsic_mul_m2x2_v2(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& m = args[0];
	VDShaderParserVariable& v = args[1];

	if (!v.mbIsConst || !m.mbIsConst) {
		func.PromoteToConstantRegisters(v);
		func.PromoteToConstantRegisters(m);
	}

	if (v.mbIsConst) {
		float x = v.m[0].v[0];
		float y = v.m[0].v[1];

		float rx = m.m[0].v[0]*x + m.m[0].v[1]*y;
		float ry = m.m[1].v[0]*x + m.m[1].v[1]*y;

		m.m[0].v[0] = rx;
		m.m[0].v[1] = ry;
		m.mType = kTypeFloat2;
	} else {
		int rA = func.AllocateRegisters(1);
		int c0 = func.AddConstantZero();

		func.dp2add(rA, 0x1, m.mRegister + 0, v.mRegister, c0, 0);
		func.dp2add(rA, 0x2, m.mRegister + 1, v.mRegister, c0, 0);
		m.mType = kTypeFloat4;
		m.mRegister = rA;
	}

	return true;
}

bool VDShaderCompileIntrinsic_mul_v3_m3x3(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& v = args[0];
	VDShaderParserVariable& m = args[1];

	if (!v.mbIsConst || !m.mbIsConst) {
		func.PromoteToConstantRegisters(v);
		func.PromoteToConstantRegisters(m);
	}

	if (v.mbIsConst) {
		float x = v.m[0].v[0];
		float y = v.m[0].v[1];
		float z = v.m[0].v[2];

		v.m[0].v[0] = m.m[0].v[0]*x + m.m[1].v[0]*y + m.m[2].v[0]*z;
		v.m[0].v[1] = m.m[0].v[1]*x + m.m[1].v[1]*y + m.m[2].v[1]*z;
		v.m[0].v[2] = m.m[0].v[2]*x + m.m[1].v[2]*y + m.m[2].v[2]*z;
	} else {
		int rA = func.AllocateRegisters(3);

		func.mul_swizzle(rA+0, 0x7, v.mRegister, 0x00, m.mRegister + 0, 0xE4);
		func.mad_swizzle(rA+1, 0x7, v.mRegister, 0x55, m.mRegister + 1, 0xE4, rA+0, 0xE4);
		func.mad_swizzle(rA+2, 0x7, v.mRegister, 0xAA, m.mRegister + 2, 0xE4, rA+1, 0xE4);
		v.mRegister = rA + 2;
	}

	return true;
}

bool VDShaderCompileIntrinsic_mul_v4_m4x4(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& v = args[0];
	VDShaderParserVariable& m = args[1];

	if (!v.mbIsConst || !m.mbIsConst) {
		func.PromoteToConstantRegisters(v);
		func.PromoteToConstantRegisters(m);
	}

	if (v.mbIsConst) {
		float x = v.m[0].v[0];
		float y = v.m[0].v[1];
		float z = v.m[0].v[2];
		float w = v.m[0].v[3];

		v.m[0].v[0] = m.m[0].v[0]*x + m.m[1].v[0]*y + m.m[2].v[0]*z + m.m[3].v[0]*w;
		v.m[0].v[1] = m.m[0].v[1]*x + m.m[1].v[1]*y + m.m[2].v[1]*z + m.m[3].v[1]*w;
		v.m[0].v[2] = m.m[0].v[2]*x + m.m[1].v[2]*y + m.m[2].v[2]*z + m.m[3].v[2]*w;
		v.m[0].v[3] = m.m[0].v[3]*x + m.m[1].v[3]*y + m.m[2].v[3]*z + m.m[3].v[3]*w;
	} else {
		int rA = func.AllocateRegisters(4);

		func.mul_swizzle(rA+0, 0xF, v.mRegister, 0x00, m.mRegister + 0, 0xE4);
		func.mad_swizzle(rA+1, 0xF, v.mRegister, 0x55, m.mRegister + 1, 0xE4, rA+0, 0xE4);
		func.mad_swizzle(rA+2, 0xF, v.mRegister, 0xAA, m.mRegister + 2, 0xE4, rA+1, 0xE4);
		func.mad_swizzle(rA+3, 0xF, v.mRegister, 0xFF, m.mRegister + 3, 0xE4, rA+2, 0xE4);
		v.mRegister = rA + 3;
	}

	return true;
}

bool VDShaderCompileIntrinsic_mul_m4x4_v4(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& m = args[0];
	VDShaderParserVariable& v = args[1];

	if (!v.mbIsConst || !m.mbIsConst) {
		func.PromoteToConstantRegisters(v);
		func.PromoteToConstantRegisters(m);
	}

	if (v.mbIsConst) {
		float x = v.m[0].v[0];
		float y = v.m[0].v[1];
		float z = v.m[0].v[2];
		float w = v.m[0].v[3];

		float rx = m.m[0].v[0]*x + m.m[0].v[1]*y + m.m[0].v[2]*z + m.m[0].v[3]*w;
		float ry = m.m[1].v[0]*x + m.m[1].v[1]*y + m.m[1].v[2]*z + m.m[1].v[3]*w;
		float rz = m.m[2].v[0]*x + m.m[2].v[1]*y + m.m[2].v[2]*z + m.m[2].v[3]*w;
		float rw = m.m[3].v[0]*x + m.m[3].v[1]*y + m.m[3].v[2]*z + m.m[3].v[3]*w;

		m.m[0].v[0] = rx;
		m.m[0].v[1] = ry;
		m.m[0].v[2] = rz;
		m.m[0].v[3] = rw;
		m.mType = kTypeFloat4;
	} else {
		int rA = func.AllocateRegisters(1);

		func.dp4(rA, 0x1, m.mRegister + 0, v.mRegister);
		func.dp4(rA, 0x2, m.mRegister + 1, v.mRegister);
		func.dp4(rA, 0x4, m.mRegister + 2, v.mRegister);
		func.dp4(rA, 0x8, m.mRegister + 3, v.mRegister);

		m.mType = kTypeFloat4;
		m.mRegister = rA;
	}

	return true;
}

bool VDShaderCompileIntrinsic_tex2D(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& samp = args[0];
	VDShaderParserVariable& uv = args[1];

	func.PromoteToConstantRegisters(samp);
	func.PromoteToConstantRegisters(uv);

	int rA = func.AllocateRegisters(1);

	func.texld(rA, 0xF, uv.mRegister, 0xE4, samp.mRegister);

	samp.mbIsConst = false;
	samp.mType = kTypeFloat4;
	samp.mRegister = rA;
	return true;
}

bool VDShaderCompileIntrinsic_min(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& s = args[0];
	VDShaderParserVariable& t = args[1];

	if (s.mbIsConst && t.mbIsConst) {
		s.m[0] = min(s.m[0], t.m[0]);
	} else {
		int rA = func.AllocateRegisters(1);

		func.PromoteToConstantRegisters(s);
		func.PromoteToConstantRegisters(t);

		unsigned char finalMask = kColWriteMaskForType[s.mType];
		func.min(rA, finalMask, s.mRegister, t.mRegister);

		s.mRegister = rA;
	}

	return true;
}

bool VDShaderCompileIntrinsic_max(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& s = args[0];
	VDShaderParserVariable& t = args[1];

	if (s.mbIsConst && t.mbIsConst) {
		s.m[0] = max(s.m[0], t.m[0]);
	} else {
		int rA = func.AllocateRegisters(1);

		func.PromoteToConstantRegisters(s);
		func.PromoteToConstantRegisters(t);

		unsigned char finalMask = kColWriteMaskForType[s.mType];
		func.max(rA, finalMask, s.mRegister, t.mRegister);

		s.mRegister = rA;
	}

	return true;
}

namespace {
	// These are for the implementation of the SINCOS macro according to the
	// DirectX DDK; we use them in the sin/cos intrinsics.
	static const VDShaderConstant kSinCosConst2 = { 3.1415926535f, 0.5f, 6.28319f, 0.159155f };
}

bool VDShaderCompileIntrinsic_sin(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& s = args[0];

	if (s.mbIsConst) {
		s.m[0].v[0] = sinf(s.m[0].v[0]);
	} else {
		int r0 = func.AllocateRegisters(5);
		int r1 = r0+1;
		int r2 = r0+2;
		int r3 = r0+3;
		int r4 = r0+4;
		int c2 = func.AddConstant(kSinCosConst2);

		// range reduce
		//
		// mad r0.x, v0.x, c2.w, c2.y 
		// frc r0.x, r0.x
		// mad r0.x, r0.x, c2.z, -c2.x

		func.mad_swizzle(r0, 0x01, s.mRegister, 0x00, c2, 0xFF, c2, 0x55);
		func.frc(r1, 0x01, r0, 0x00);
		func.mad_swizzle_mods(r2, 0x01, r1, 0x00, 0, c2, 0xAA, 0, c2, 0xE4, kVDShaderMod_Negate);

		func.sincos(r3, 0x02, r2, 0);

		func.mov_swizzle(r4, 0x01, r3, 0x55);

		s.mRegister = r4;
	}

	return true;
}

bool VDShaderCompileIntrinsic_cos(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& s = args[0];

	if (s.mbIsConst) {
		s.m[0].v[0] = cosf(s.m[0].v[0]);
	} else {
		int r0 = func.AllocateRegisters(5);
		int r1 = r0+1;
		int r2 = r0+2;
		int r3 = r0+3;
		int r4 = r0+4;
		int c2 = func.AddConstant(kSinCosConst2);

		// range reduce
		//
		// mad r0.x, v0.x, c2.w, c2.y 
		// frc r0.x, r0.x
		// mad r0.x, r0.x, c2.z, -c2.x

		func.mad_swizzle(r0, 0x01, s.mRegister, 0x00, c2, 0xFF, c2, 0x55);
		func.frc(r1, 0x01, r0, 0x00);
		func.mad_swizzle_mods(r2, 0x01, r1, 0x00, 0, c2, 0xAA, 0, c2, 0xE4, kVDShaderMod_Negate);

		func.sincos(r3, 0x01, r2, 0);

		func.mov_swizzle(r4, 0x01, r3, 0x00);

		s.mRegister = r4;
	}

	return true;
}

bool VDShaderCompileIntrinsic_tan(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& s = args[0];

	if (s.mbIsConst) {
		s.m[0].v[0] = tanf(s.m[0].v[0]);
	} else {
		int r0 = func.AllocateRegisters(6);
		int r1 = r0+1;
		int r2 = r0+2;
		int r3 = r0+3;
		int r4 = r0+4;
		int r5 = r0+5;
		int c2 = func.AddConstant(kSinCosConst2);

		// range reduce
		//
		// mad r0.x, v0.x, c2.w, c2.y 
		// frc r0.x, r0.x
		// mad r0.x, r0.x, c2.z, -c2.x

		func.mad_swizzle(r0, 0x01, s.mRegister, 0x00, c2, 0xFF, c2, 0x55);
		func.frc(r1, 0x01, r0, 0x00);
		func.mad_swizzle_mods(r2, 0x01, r1, 0x00, 0, c2, 0xAA, 0, c2, 0xE4, kVDShaderMod_Negate);

		func.sincos(r3, 0x03, r2, 0);

		func.rcp(r4, 0x01, r3);
		func.mul_swizzle(r5, 0x01, r3, 0x55, r4, 0);

		s.mRegister = r5;
	}

	return true;
}

bool VDShaderCompileIntrinsic_asin(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& s = args[0];
	int cols = kColCountForType[s.mType];

	if (s.mbIsConst) {
		for(int i=0; i<cols; ++i) {
			float x = s.m[0].v[i];

			if (!(fabsf(x) <= 1.0f))
				return lexer.SetErrorF("Cannot compute asin(%g)", x);

			s.m[0].v[i] = asinf(x);
		}
	} else {
		// asin() approximation from:
		//
		// Milton Abramowitz and Irene A. Stegun, eds., Handbook of Math-ematical Functions With Formulas,
		//	Graphs, and Mathematical Tables, NBS Applied Mathematics Series 55, National Bureau of Standards,
		//	Washington, DC (1964). 
		//
		// For 0 <= x <= 1:
		//
		//	arcsin x = (pi/2) - sqrt(1-x)(a0 + a1*x + a2*x^2 + a3*x^3) + e(x)
		//
		//	|e(x)| < 5e-5
		//
		//	a0 = 1.5707288
		//	a1 = -0.2121144
		//	a2 = 0.0742610
		//	a3 = -0.0187293
		//
		// Therefore:
		//
		//	abs	r0.x, v0.x
		//	mad r1.x, r0.x, c0.x, c0.y
		//	mad r2.x, r0.x, r1.x, c0.z
		//	mad r3.x, r0.x, r2.x, c0.w
		//	add r4.x, -r0.x, c1.x
		//	rsq r5.x, r4.x
		//	rcp r6.x, r5.x
		//	mad r7.x, -r3.x, r6.x, c1.y
		//	cmp o0.x, v0.x, r7.x, -r7.x

		static const VDShaderConstant kAsincosC0 = { -0.0187293f, 0.0742610f, -0.2121144f, 1.5707288f };
		static const VDShaderConstant kAsincosC1 = { 1, 1.570796327f, 3.1415926535f };

		int c0 = func.AddConstant(kAsincosC0);
		int c1 = func.AddConstant(kAsincosC1);
		int r = func.AllocateRegisters(9);
		int writeMask = kColWriteMaskForType[s.mType];

		func.abs(r+0, writeMask, s.mRegister);
		func.mad_swizzle(r+1, writeMask, r+0, 0xE4, c0, 0x00, c0, 0x55);
		func.mad_swizzle(r+2, writeMask, r+0, 0xE4, r+1, 0xE4, c0, 0xaa);
		func.mad_swizzle(r+3, writeMask, r+0, 0xE4, r+2, 0xE4, c0, 0xff);
		func.add_swizzle_mods(r+4, writeMask, r+0, 0xE4, kVDShaderMod_Negate, c1, 0, 0);
		func.rsq(r+5, 0x01, r+4);
		func.rcp(r+6, 0x01, r+5);
		func.mad_swizzle_mods(r+7, writeMask, r+3, 0xE4, kVDShaderMod_Negate, r+6, 0xE4, 0, c1, 0x55, 0);
		func.cmp_swizzle_mods(r+8, writeMask, s.mRegister, 0xE4, 0, r+7, 0xE4, 0, r+7, 0xE4, kVDShaderMod_Negate);

		s.mRegister = r+8;
	}

	return true;
}

bool VDShaderCompileIntrinsic_acos(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& s = args[0];
	int cols = kColCountForType[s.mType];

	if (s.mbIsConst) {
		for(int i=0; i<cols; ++i) {
			float x =  s.m[0].v[i];

			if (!(fabsf(x) <= 1.0f))
				return lexer.SetErrorF("Cannot compute acos(%g)", x);

			s.m[0].v[i] = acosf(x);
		}
	} else {
		// asin() approximation from:
		//
		// Milton Abramowitz and Irene A. Stegun, eds., Handbook of Math-ematical Functions With Formulas,
		//	Graphs, and Mathematical Tables, NBS Applied Mathematics Series 55, National Bureau of Standards,
		//	Washington, DC (1964). 
		//
		// For 0 <= x <= 1:
		//
		//	arcsin x = (pi/2) - sqrt(1-x)(a0 + a1*x + a2*x^2 + a3*x^3) + e(x)
		//
		//	|e(x)| < 5e-5
		//
		//	a0 = 1.5707288
		//	a1 = -0.2121144
		//	a2 = 0.0742610
		//	a3 = -0.0187293
		//
		//	arccos x = (pi/2) - arcsin
		//
		// Therefore:
		//
		//	abs	r0.x, v0.x
		//	mad r1.x, r0.x, c0.x, c0.y
		//	mad r2.x, r0.x, r1.x, c0.z
		//	mad r3.x, r0.x, r2.x, c0.w
		//	add r4.x, -r0.x, c1.x
		//	rsq r5.x, r4.x
		//	rcp r6.x, r5.x
		//	mul r7.x, r3.x, r6.x
		//	add r8.x, -r7.x, c1.y
		//	cmp o0.x, v0.x, r7.x, r8.x

		static const VDShaderConstant kAsincosC0 = { -0.0187293f, 0.0742610f, -0.2121144f, 1.5707288f };
		static const VDShaderConstant kAsincosC1 = { 1, 1.570796327f, 3.1415926535f };

		int c0 = func.AddConstant(kAsincosC0);
		int c1 = func.AddConstant(kAsincosC1);
		int r = func.AllocateRegisters(10);
		int writeMask = kColWriteMaskForType[s.mType];

		func.abs(r+0, writeMask, s.mRegister);
		func.mad_swizzle(r+1, writeMask, r+0, 0xE4, c0, 0x00, c0, 0x55);
		func.mad_swizzle(r+2, writeMask, r+0, 0xE4, r+1, 0xE4, c0, 0xaa);
		func.mad_swizzle(r+3, writeMask, r+0, 0xE4, r+2, 0xE4, c0, 0xff);
		func.add_swizzle_mods(r+4, writeMask, r+0, 0xE4, kVDShaderMod_Negate, c1, 0, 0);

		for(int i=0; i<cols; ++i) {
			func.rsq(r+5, 1 << i, r+4, i);
			func.rcp(r+6, 1 << i, r+5, i);
		}
		func.mul(r+7, writeMask, r+3, r+6);
		func.add_swizzle_mods(r+8, writeMask, r+7, 0xE4, kVDShaderMod_Negate, c1, 0xAA, 0);
		func.cmp_swizzle_mods(r+9, writeMask, s.mRegister, 0xE4, 0, r+7, 0xE4, 0, r+8, 0xE4, 0);

		s.mRegister = r+9;
	}

	return true;
}

bool VDShaderCompileIntrinsic_atan(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& s = args[0];

	if (s.mbIsConst) {
		float x = s.m[0].v[0];

		s.m[0].v[0] = atanf(x);
	} else {
		// atan() approximation from:
		//
		// Milton Abramowitz and Irene A. Stegun, eds., Handbook of Math-ematical Functions With Formulas,
		//	Graphs, and Mathematical Tables, NBS Applied Mathematics Series 55, National Bureau of Standards,
		//	Washington, DC (1964). 
		//
		// For -1 <= x <= 1:
		//
		//	arctan x = a1*x + a3*x^3 + a5*x^5 + a7*x^7 + a9*x^9 + e(x)
		//
		//	|e(x)| < 1e-5
		//
		//	a1 = 0.9998660
		//	a3 = -0.3302995
		//	a5 = 0.1801410
		//	a7 = -0.0851330
		//	a9 = 0.0208351
		//
		// Therefore:
		//
		//	abs r0.x, v0.x
		//	rcp r1.x, r0.x
		//	min r2.x, r0.x, r1.x			;y = min(|x|, 1/|x|)
		//	mul r3.x, r2.x, r2.x			;y^2
		//	mad r4.x, r3.x, c0.x, c0.y		;a9*y^2 + a7
		//	mad r5.x, r4.x, r3.x, c0.z		;a9*y^4 + a7*y^2 + a5
		//	mad r6.x, r5.x, r3.x, c0.w		;a9*y^6 + a7*y^4 + a5*y^2 + a3
		//	mad r7.x, r6.x, r3.x, c1.x		;a9*y^8 + a7*y^6 + a5*y^4 + a3*y^2 + a1
		//	mul r8.x, r7.x, r2.x			;u = a9*y^9 + a7*y^7 + a5*y^5 + a3*y^3 + a1*y
		//	add r9.x, -r8.x, c1.y			;pi/2 - a9*y^9 + a7*y^7 + a5*y^5 + a3*y^3 + a1*y
		//	add r10.x, r0.x, c1.z			;|x| - 1
		//	cmp r11.x, r10.x, r9.x, r8.x	;v = |x| >= 1 ? pi/2 - u : u
		//	cmp r12.x, v0.x, r11.x, -r11.x	;x >= 0 ? v : -v

		static const VDShaderConstant kAtanC0 = {
			0.0208351f,		// a9
			-0.0851330f,	// a7
			0.1801410f,		// a5
			-0.3302995f,	// a3
		};
		static const VDShaderConstant kAtanC1 = {
			0.9998660f,		// a1
			1.570796327f,
			-1
		};

		int c0 = func.AddConstant(kAtanC0);
		int c1 = func.AddConstant(kAtanC1);
		int r = func.AllocateRegisters(13);

		//	abs r0.x, v0.x
		func.abs(r+0, 0x01, s.mRegister);

		//	rcp r1.x, r0.x
		func.rcp(r+1, 0x01, r+0);

		//	min r2.x, r0.x, r1.x			;y = min(|x|, 1/|x|)
		func.min(r+2, 0x01, r+0, r+1);

		//	mul r3.x, r2.x, r2.x			;y^2
		func.mul(r+3, 0x01, r+2, r+2);

		//	mad r4.x, r3.x, c0.x, c0.y		;a9*y^2 + a7
		func.mad_swizzle(r+4, 0x01, r+3, 0, c0, 0, c0, 0x55);

		//	mad r5.x, r4.x, r3.x, c0.z		;a9*y^4 + a7*y^2 + a5
		func.mad_swizzle(r+5, 0x01, r+4, 0, r+3, 0, c0, 0xaa);

		//	mad r6.x, r5.x, r3.x, c0.w		;a9*y^6 + a7*y^4 + a5*y^2 + a3
		func.mad_swizzle(r+6, 0x01, r+5, 0, r+3, 0, c0, 0xff);

		//	mad r7.x, r6.x, r3.x, c1.x		;a9*y^8 + a7*y^6 + a5*y^4 + a3*y^2 + a1
		func.mad_swizzle(r+7, 0x01, r+6, 0, r+3, 0, c1, 0x00);

		//	mul r8.x, r7.x, r2.x			;u = a9*y^9 + a7*y^7 + a5*y^5 + a3*y^3 + a1*y
		func.mul(r+8, 0x01, r+7, r+2);

		//	add r9.x, -r8.x, c1.y			;pi/2 - a9*y^9 + a7*y^7 + a5*y^5 + a3*y^3 + a1*y
		func.add_swizzle_mods(r+9, 0x01, r+8, 0, kVDShaderMod_Negate, c1, 0x55, 0);

		//	add r10.x, r0.x, c1.z			;|x| - 1
		func.add_swizzle_mods(r+10, 0x01, r+0, 0, 0, c1, 0xaa, 0);

		//	cmp r11.x, r10.x, r9.x, r8.x	;v = |x| >= 1 ? pi/2 - u : u
		func.cmp(r+11, 0x01, r+10, 0, r+9, r+8);

		//	cmp r12.x, v0.x, r11.x, -r11.x	;x >= 0 ? v : -v
		func.cmp_swizzle_mods(r+12, 0x01, s.mRegister, 0, 0, r+11, 0, 0, r+11, 0, kVDShaderMod_Negate);

		s.mRegister = r+12;
	}

	return true;
}

bool VDShaderCompileIntrinsic_frac(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& s = args[0];

	if (s.mbIsConst) {
		s.m[0].v[0] = s.m[0].v[0] - floorf(s.m[0].v[0]);
	} else {
		int r0 = func.AllocateRegisters(1);
		func.frc(r0, 0x01, s.mRegister, 0x00);

		s.mRegister = r0;
	}

	return true;
}

bool VDShaderCompileIntrinsic_abs(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& v = args[0];

	if (v.mbIsConst) {
		v.m[0] = abs(v.m[0]);
	} else {
		int rA = func.AllocateRegisters(1);

		func.abs(rA, kColWriteMaskForType[v.mType], v.mRegister);

		v.mRegister = rA;
	}

	return true;
}

bool VDShaderCompileIntrinsic_cross(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& x = args[0];
	VDShaderParserVariable& y = args[1];

	if (x.mbIsConst) {
		x.m[0] = cross(x.m[0], y.m[0]);
	} else {
		int rA = func.AllocateRegisters(1);

		func.PromoteToConstantRegisters(x);
		func.PromoteToConstantRegisters(y);

		func.crs(rA, 0x07, x.mRegister, y.mRegister);

		x.mRegister = rA;
	}

	return true;
}

bool VDShaderCompileIntrinsic_lerp(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& x = args[0];
	VDShaderParserVariable& y = args[1];
	VDShaderParserVariable& z = args[2];

	if (x.mbIsConst && y.mbIsConst && z.mbIsConst) {
		x.m[0] = lerp(x.m[0], y.m[0], z.m[0]);
	} else {
		int rA = func.AllocateRegisters(1);

		func.PromoteToConstantRegisters(x);
		func.PromoteToConstantRegisters(y);
		func.PromoteToConstantRegisters(z);

		func.lrp(rA, kColWriteMaskForType[x.mType], z.mRegister, y.mRegister, x.mRegister);

		x.mRegister = rA;
	}

	return true;
}

bool VDShaderCompileIntrinsic_rsqrt(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& x = args[0];

	if (x.mbIsConst) {
		if (x.m[0].v[0] < 1e-30f)
			return lexer.SetError("Attempted rsqrt() with zero or negative number");

		x.m[0].v[0] = 1.0f / sqrtf(x.m[0].v[0]);
	} else {
		int rA = func.AllocateRegisters(1);

		func.rsq(rA, 1, x.mRegister);

		x.mRegister = rA;
	}

	return true;
}

bool VDShaderCompileIntrinsic_sqrt(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& x = args[0];

	if (x.mbIsConst) {
		if (x.m[0].v[0] < 0)
			return lexer.SetError("Attempted sqrt() with negative number");

		x.m[0].v[0] = sqrtf(x.m[0].v[0]);
	} else {
		int rA = func.AllocateRegisters(2);
		int rB = rA + 1;

		func.rsq(rA, 1, x.mRegister);
		func.mul(rB, 1, rA, x.mRegister);

		x.mRegister = rB;
	}

	return true;
}

bool VDShaderCompileIntrinsic_log2(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& x = args[0];
	int cols = kColCountForType[x.mType];
	int rows = kRowCountForType[x.mType];

	if (x.mbIsConst) {
		for(int i=0; i<rows; ++i) {
			for(int j=0; j<cols; ++j) {
				if (x.m[i].v[j] < 0)
					return lexer.SetError("Indefinite logarithm");
				else if (x.m[i].v[j] == 0)
					return lexer.SetError("Attempted logarithm of zero");

				x.m[i].v[j] = logf(x.m[i].v[j]) * 1.4426950408889634073599246810019f;
			}
		}
	} else {
		int rA = func.AllocateRegisters(rows);

		for(int i=0; i<rows; ++i) {
			for(int j=0; j<cols; ++j) {
				func.log_swizzle(rA+i, 1 << j, x.mRegister + i, j);
			}
		}

		x.mRegister = rA;
	}

	return true;
}

bool VDShaderCompileIntrinsic_exp2(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& x = args[0];
	int cols = kColCountForType[x.mType];
	int rows = kRowCountForType[x.mType];

	if (x.mbIsConst) {
		for(int i=0; i<rows; ++i) {
			for(int j=0; j<cols; ++j) {
				x.m[i].v[j] = expf(x.m[i].v[j]) * 0.69314718055994530941723212145818f;
			}
		}
	} else {
		int rA = func.AllocateRegisters(rows);

		for(int i=0; i<rows; ++i) {
			for(int j=0; j<cols; ++j) {
				func.exp_swizzle(rA+i, 1 << j, x.mRegister + i, j);
			}
		}

		x.mRegister = rA;
	}

	return true;
}

bool VDShaderCompileIntrinsic_pow(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& x = args[0];
	VDShaderParserVariable& y = args[1];

	if (x.mbIsConst && y.mbIsConst) {
		if (x.m[0].v[0] < 0)
			return lexer.SetError("pow() of negative value");

		switch(kColCountForType[x.mType]) {
			case 3:
				x.m[0].v[3] = powf(x.m[0].v[3], y.m[0].v[3]);
			case 2:
				x.m[0].v[2] = powf(x.m[0].v[2], y.m[0].v[2]);
			case 1:
				x.m[0].v[1] = powf(x.m[0].v[1], y.m[0].v[1]);
			case 0:
				x.m[0].v[0] = powf(x.m[0].v[0], y.m[0].v[0]);
				break;
		}
	} else {
		int rA = func.AllocateRegisters(3);
		int rB = rA + 1;
		int rC = rA + 2;

		func.PromoteToConstantRegisters(x);
		func.PromoteToConstantRegisters(y);

		int n = kColCountForType[x.mType];
		int wm = kColWriteMaskForType[x.mType];
		
		for(int i=0; i<n; ++i)
			func.log_swizzle(rA, 1 << i, x.mRegister, i);

		func.mul(rB, wm, rA, y.mRegister);

		for(int i=0; i<n; ++i)
			func.exp_swizzle(rC, 1 << i, rB, i);

		x.mRegister = rC;
	}

	return true;
}

bool VDShaderCompileIntrinsic_reflect(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& x = args[0];
	VDShaderParserVariable& y = args[1];

	if (x.mbIsConst && y.mbIsConst) {
		x.m[0] = x.m[0] - 2.0f * y.m[0] * dp3(x.m[0], y.m[0]);
	} else {
		int rA = func.AllocateRegisters(3);
		int rB = rA + 1;
		int rC = rA + 2;

		func.PromoteToConstantRegisters(x);
		func.PromoteToConstantRegisters(y);

		func.dp3(rA, 1, x.mRegister, y.mRegister);
		func.add(rB, 1, rA, rA);
		func.mad_swizzle_mods(rC, 7, rB, 0, kVDShaderMod_Negate, y.mRegister, 0xE4, 0, x.mRegister, 0xE4, 0);

		x.mRegister = rC;
	}

	return true;
}

bool VDShaderCompileIntrinsic_saturate(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, VDShaderParserVariable *args, int numArgs) {
	VDShaderParserVariable& x = args[0];

	if (x.mbIsConst) {
		x.m[0] = saturate(x.m[0]);
	} else {
		int rA = func.AllocateRegisters(1);

		func.PromoteToConstantRegisters(x);

		func.mov_sat(rA, kColWriteMaskForType[x.mType], x.mRegister);

		x.mRegister = rA;
	}

	return true;
}

namespace {
	template<VDShaderParserType t1> struct VDShaderParserArgTypes0 { static const VDShaderParserType kTypeArray[]; };
	template<VDShaderParserType t1> const VDShaderParserType VDShaderParserArgTypes0<t1>::kTypeArray[] = { t1, kTypeInvalid };
	template<VDShaderParserType t1, VDShaderParserType t2> struct VDShaderParserArgTypes1 { static const VDShaderParserType kTypeArray[]; };
	template<VDShaderParserType t1, VDShaderParserType t2> const VDShaderParserType VDShaderParserArgTypes1<t1, t2>::kTypeArray[] = { t1, t2, kTypeInvalid };
	template<VDShaderParserType t1, VDShaderParserType t2, VDShaderParserType t3> struct VDShaderParserArgTypes2 { static const VDShaderParserType kTypeArray[]; };
	template<VDShaderParserType t1, VDShaderParserType t2, VDShaderParserType t3> const VDShaderParserType VDShaderParserArgTypes2<t1, t2, t3>::kTypeArray[] = { t1, t2, t3, kTypeInvalid };
	template<VDShaderParserType t1, VDShaderParserType t2, VDShaderParserType t3, VDShaderParserType t4> struct VDShaderParserArgTypes3 { static const VDShaderParserType kTypeArray[]; };
	template<VDShaderParserType t1, VDShaderParserType t2, VDShaderParserType t3, VDShaderParserType t4> const VDShaderParserType VDShaderParserArgTypes3<t1, t2, t3, t4>::kTypeArray[] = { t1, t2, t3, t4, kTypeInvalid };
}

#define ARGS_0(rv) VDShaderParserArgTypes0<kType##rv>::kTypeArray
#define ARGS_1(rv, t1) VDShaderParserArgTypes1<kType##rv, kType##t1>::kTypeArray
#define ARGS_2(rv, t1, t2) VDShaderParserArgTypes2<kType##rv, kType##t1, kType##t2>::kTypeArray
#define ARGS_3(rv, t1, t2, t3) VDShaderParserArgTypes3<kType##rv, kType##t1, kType##t2, kType##t3>::kTypeArray

const VDShaderIntrinsicEntry kVDShaderIntrinsics[]={
	{ "dot",		VDShaderCompileIntrinsic_dot, ARGS_2(Float, Float2, Float2) },
	{ NULL,			VDShaderCompileIntrinsic_dot, ARGS_2(Float, Float3, Float3) },
	{ NULL,			VDShaderCompileIntrinsic_dot, ARGS_2(Float, Float4, Float4) },
	{ "length",		VDShaderCompileIntrinsic_length, ARGS_1(Float, Float2) },
	{ NULL,			VDShaderCompileIntrinsic_length, ARGS_1(Float, Float3) },
	{ NULL,			VDShaderCompileIntrinsic_length, ARGS_1(Float, Float4) },
	{ "normalize",	VDShaderCompileIntrinsic_normalize, ARGS_1(Float2, Float2) },
	{ NULL,			VDShaderCompileIntrinsic_normalize, ARGS_1(Float3, Float3) },
	{ NULL,			VDShaderCompileIntrinsic_normalize, ARGS_1(Float4, Float4) },
	{ "mul",		VDShaderCompileIntrinsic_mul_v2_m2x2, ARGS_2(Float2, Float2, Float2x2) },
	{ NULL,			VDShaderCompileIntrinsic_mul_m2x2_v2, ARGS_2(Float2, Float2x2, Float2) },
	{ NULL,			VDShaderCompileIntrinsic_mul_v3_m3x3, ARGS_2(Float3, Float3, Float3x3) },
	{ NULL,			VDShaderCompileIntrinsic_mul_v4_m4x4, ARGS_2(Float4, Float4, Float4x4) },
	{ NULL,			VDShaderCompileIntrinsic_mul_m4x4_v4, ARGS_2(Float4, Float4x4, Float4) },
	{ "tex2D",		VDShaderCompileIntrinsic_tex2D, ARGS_2(Float4, Sampler, Float2) },
	{ "min",		VDShaderCompileIntrinsic_min, ARGS_2(Float, Float, Float) },
	{ NULL,			VDShaderCompileIntrinsic_min, ARGS_2(Float2, Float2, Float2) },
	{ NULL,			VDShaderCompileIntrinsic_min, ARGS_2(Float3, Float3, Float3) },
	{ NULL,			VDShaderCompileIntrinsic_min, ARGS_2(Float4, Float4, Float4) },
	{ "max",		VDShaderCompileIntrinsic_max, ARGS_2(Float, Float, Float) },
	{ NULL,			VDShaderCompileIntrinsic_max, ARGS_2(Float2, Float2, Float2) },
	{ NULL,			VDShaderCompileIntrinsic_max, ARGS_2(Float3, Float3, Float3) },
	{ NULL,			VDShaderCompileIntrinsic_max, ARGS_2(Float4, Float4, Float4) },
	{ "sin",		VDShaderCompileIntrinsic_sin, ARGS_1(Float, Float) },
	{ "cos",		VDShaderCompileIntrinsic_cos, ARGS_1(Float, Float) },
	{ "tan",		VDShaderCompileIntrinsic_tan, ARGS_1(Float, Float) },
	{ "frac",		VDShaderCompileIntrinsic_frac, ARGS_1(Float, Float) },
	{ "abs",		VDShaderCompileIntrinsic_abs, ARGS_1(Float, Float) },
	{ NULL,			VDShaderCompileIntrinsic_abs, ARGS_1(Float2, Float2) },
	{ NULL,			VDShaderCompileIntrinsic_abs, ARGS_1(Float3, Float3) },
	{ NULL,			VDShaderCompileIntrinsic_abs, ARGS_1(Float4, Float4) },
	{ "cross",		VDShaderCompileIntrinsic_cross, ARGS_2(Float3, Float3, Float3) },
	{ "lerp",		VDShaderCompileIntrinsic_lerp, ARGS_3(Float, Float, Float, Float) },
	{ NULL,			VDShaderCompileIntrinsic_lerp, ARGS_3(Float2, Float2, Float2, Float2) },
	{ NULL,			VDShaderCompileIntrinsic_lerp, ARGS_3(Float3, Float3, Float3, Float3) },
	{ NULL,			VDShaderCompileIntrinsic_lerp, ARGS_3(Float4, Float4, Float4, Float4) },
	{ "rsqrt",		VDShaderCompileIntrinsic_rsqrt, ARGS_1(Float, Float) },
	{ "sqrt",		VDShaderCompileIntrinsic_sqrt, ARGS_1(Float, Float) },
	{ "log2",		VDShaderCompileIntrinsic_log2, ARGS_1(Float, Float) },
	{ NULL,			VDShaderCompileIntrinsic_log2, ARGS_1(Float2, Float2) },
	{ NULL,			VDShaderCompileIntrinsic_log2, ARGS_1(Float3, Float3) },
	{ NULL,			VDShaderCompileIntrinsic_log2, ARGS_1(Float4, Float4) },
	{ "exp2",		VDShaderCompileIntrinsic_exp2, ARGS_1(Float, Float) },
	{ NULL,			VDShaderCompileIntrinsic_exp2, ARGS_1(Float2, Float2) },
	{ NULL,			VDShaderCompileIntrinsic_exp2, ARGS_1(Float3, Float3) },
	{ NULL,			VDShaderCompileIntrinsic_exp2, ARGS_1(Float4, Float4) },
	{ "pow",		VDShaderCompileIntrinsic_pow, ARGS_2(Float, Float, Float) },
	{ NULL,			VDShaderCompileIntrinsic_pow, ARGS_2(Float2, Float2, Float2) },
	{ NULL,			VDShaderCompileIntrinsic_pow, ARGS_2(Float3, Float3, Float3) },
	{ NULL,			VDShaderCompileIntrinsic_pow, ARGS_2(Float4, Float4, Float4) },
	{ "asin",		VDShaderCompileIntrinsic_asin, ARGS_1(Float, Float) },
	{ "acos",		VDShaderCompileIntrinsic_acos, ARGS_1(Float, Float) },
	{ NULL,			VDShaderCompileIntrinsic_acos, ARGS_1(Float2, Float2) },
	{ NULL,			VDShaderCompileIntrinsic_acos, ARGS_1(Float3, Float3) },
	{ NULL,			VDShaderCompileIntrinsic_acos, ARGS_1(Float4, Float4) },
	{ "atan",		VDShaderCompileIntrinsic_atan, ARGS_1(Float, Float) },
	{ "reflect",	VDShaderCompileIntrinsic_reflect, ARGS_2(Float3, Float3, Float3) },
	{ "saturate",	VDShaderCompileIntrinsic_saturate, ARGS_1(Float, Float) },
	{ NULL,			VDShaderCompileIntrinsic_saturate, ARGS_1(Float2, Float2) },
	{ NULL,			VDShaderCompileIntrinsic_saturate, ARGS_1(Float3, Float3) },
	{ NULL,			VDShaderCompileIntrinsic_saturate, ARGS_1(Float4, Float4) },
	{ NULL, NULL, NULL }
};

const VDShaderIntrinsicEntry *VDShaderParserGetIntrinsicByName(const char *s) {
	const VDShaderIntrinsicEntry *p = kVDShaderIntrinsics;

	while(p->mpCompiler) {
		if (p->mpName && !strcmp(s, p->mpName))
			return p;

		++p;
	}

	return NULL;
}

namespace nsVDShaderParser {
	enum VDShaderParserClass {
		kClassNone,
		kClassScalar,
		kClassVector,
		kClassMatrix
	};
};

VDShaderParserClass VDShaderParserGetTypeClass(int type) {
	switch(type) {
		case kTypeFloat:
			return kClassScalar;

		case kTypeFloat2:
		case kTypeFloat3:
		case kTypeFloat4:
			return kClassVector;

		case kTypeFloat2x2:
		case kTypeFloat2x3:
		case kTypeFloat2x4:
		case kTypeFloat3x2:
		case kTypeFloat3x3:
		case kTypeFloat3x4:
		case kTypeFloat4x2:
		case kTypeFloat4x3:
		case kTypeFloat4x4:
			return kClassMatrix;

		default:
			return kClassNone;
	}
}

bool VDShaderParserCanConvertType(int fromType, int toType) {
	if (fromType == toType)
		return true;

	int fromClass = VDShaderParserGetTypeClass(fromType);
	int toClass = VDShaderParserGetTypeClass(toType);

	switch(fromClass) {
		case kClassScalar:
			switch(toClass) {
				case kClassScalar:
				case kClassVector:
				case kClassMatrix:
					return true;
			};
			break;

		case kClassVector:
			switch(toClass) {
				case kClassScalar:
					return true;
				case kClassVector:
					return kColCountForType[fromType] >= kColCountForType[toType];
				case kClassMatrix:
					return false;
			}
			break;

		case kClassMatrix:
			switch(toClass) {
				case kClassScalar:
					return true;
				case kClassVector:
					return false;
				case kClassMatrix:
					return	kColCountForType[fromType] >= kColCountForType[toType] &&
							kRowCountForType[fromType] >= kRowCountForType[toType];
			}
			break;
	}

	return false;
}

bool VDShaderParserCompileIntrinsic(IVDShaderErrorOutput& lexer, VDShaderParserFunction& func, const VDShaderIntrinsicEntry *pBaseEntry, VDShaderParserVariable *args, int numArgs) {
	typedef vdfastvector<const VDShaderIntrinsicEntry *> Candidates;
	Candidates candidates;
	Candidates candidates2;

	// find all functions that can match the correct number of arguments
	const VDShaderIntrinsicEntry *pEntry = pBaseEntry;
	do {
		const VDShaderParserType *pTypeList = pEntry->mTypes + 1;
		int funcNumArgs = 0;

		while(pTypeList[funcNumArgs])
			++funcNumArgs;

		if (funcNumArgs == numArgs)
			candidates.push_back(pEntry);

		++pEntry;
	} while(pEntry->mpCompiler && !pEntry->mpName);

	if (candidates.empty())
		return lexer.SetErrorF("No form of intrinsic '%s' takes %d arguments", pBaseEntry->mpName, numArgs);

	for(int arg = 0; arg < numArgs; ++arg) {
		const int passedArgType = args[arg].mType;

		// check for exact matches
		candidates2.clear();
		for(Candidates::const_iterator it(candidates.begin()), itEnd(candidates.end()); it != itEnd; ++it) {
			const VDShaderIntrinsicEntry& ent = **it;
			const int funcArgType = ent.mTypes[arg + 1];

			if (passedArgType == funcArgType)
				candidates2.push_back(&ent);
		}

		if (!candidates2.empty()) {
			candidates.swap(candidates2);
			continue;
		}

		// check for matches by implicitly cast
		for(Candidates::const_iterator it(candidates.begin()), itEnd(candidates.end()); it != itEnd; ++it) {
			const VDShaderIntrinsicEntry& ent = **it;
			const int funcArgType = ent.mTypes[arg + 1];

			if (VDShaderParserCanConvertType(passedArgType, funcArgType))
				candidates2.push_back(&ent);
		}

		if (!candidates2.empty()) {
			candidates.swap(candidates2);
			continue;
		}

		return lexer.SetErrorF("No form of intrinsic '%s' matched the given argument list", pBaseEntry->mpName);
	}

	if (candidates.size() > 1)
		return lexer.SetErrorF("Ambiguous match for intrinsic '%s': %d overloads match", pBaseEntry->mpName, (int)candidates.size());

	vdfastvector<VDShaderParserVariable> convertedArgs(args, args + numArgs);
	const VDShaderIntrinsicEntry& ent = *candidates.front();

	for(int arg = 0; arg < numArgs; ++arg) {
		VDShaderParserConvertVarToType(lexer, &func, convertedArgs[arg], ent.mTypes[arg + 1]);
	}

	if (convertedArgs.empty() && ent.mTypes[0] != kTypeVoid)
		convertedArgs.push_back();

	if (!ent.mpCompiler(lexer, func, convertedArgs.data(), numArgs))
		return false;

	if (ent.mTypes[0] != kTypeVoid)
		args[0] = convertedArgs[0];

	args[0].mNameToken = 0;
	args[0].mSemanticToken = 0;
	args[0].mbUnassigned = false;

	return true;
}
