//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include <vd2/system/thunk.h>
#include "jit_sse2.h"
#include "jit.h"
#include "x86disasm.h"
#include "log.h"
#include <xmmintrin.h>
#include <float.h>

extern const unsigned char g_VDDisasmX86[];

VDShaderSSE2JIT::VDShaderSSE2JIT()
	: mpShader(NULL)
{
	VDInitThunkAllocator();
}

VDShaderSSE2JIT::~VDShaderSSE2JIT() {
	if (mpShader) {
		VDFreeThunkMemory(mpShader, mShaderLen);
		mpShader = NULL;
	}

	VDShutdownThunkAllocator();
}

void VDShaderSSE2JIT::LoadSamplers(const VDShaderInterpreterSampler *const *samplers, int numSamplers) {
	mpSamplers = samplers;
	mSamplerCount = numSamplers;
}

namespace {
	struct SlotManager {
		const x86::xmmword_ref_t GetVariableSlot(int reg, int esp_offset = 0) {
			using namespace x86;

			switch(reg & kVDShaderRegTypeMask) {
				case kVDShaderRegR0:
					return xmmword_ptr [esp + (reg - kVDShaderRegR0)*16 + mTempOffset + esp_offset];

				case kVDShaderRegC0:
					return xmmword_ptr [mpConstants + (reg - kVDShaderRegC0)*4];

				case kVDShaderRegV0:
					return xmmword_ptr [esp + (reg - kVDShaderRegV0)*16 + mInputOffset + esp_offset];

				case kVDShaderRegS0:
					return xmmword_ptr [esp + (reg - kVDShaderRegS0)*32 + mSamplerOffset + esp_offset];

				case kVDShaderRegO0:
					return xmmword_ptr [esp + (reg - kVDShaderRegO0)*16 + mOutputOffset + esp_offset];

				default:
					VDASSERT(false);
					{
						static const __declspec(align(16)) float zero[4] = { 0.0f };
						return xmmword_ptr [&zero];
					}
			}
		}

		const float *mpConstants;
		int mNumConstants;
		ptrdiff_t mTempOffset;
		ptrdiff_t mInputOffset;
		ptrdiff_t mSamplerOffset;
		ptrdiff_t mOutputOffset;
	};

	class RegisterTracker {
	public:
		RegisterTracker(SlotManager *slots);

		void Flush(x86::BasicBlock *bb, int esp_offset);

		x86::varx Load(x86::BasicBlock *bb, int reg, bool markDirty, int esp_offset);
		x86::varx Allocate(x86::BasicBlock *bb, int reg, int esp_offset);
		x86::varx AllocateAnon(x86::BasicBlock *bb, int esp_offset);
		void Flush(x86::BasicBlock *bb, int reg, int esp_offset);
		void Destroy(int reg);
		void Unlock(const x86::varx& var);
		void UnlockAndRetag(const x86::varx& var, int reg);

	protected:
		int Evict(x86::BasicBlock *bb, int esp_offset);

		int mReg[8];
		int mAge[8];
		bool mbDirty[8];
		bool mbLocked[8];
		int mClock;

		SlotManager *mpSlots;
	};

	RegisterTracker::RegisterTracker(SlotManager *slots)
		: mpSlots(slots)
	{
		memset(mReg, 0, sizeof mReg);
		memset(mAge, 0, sizeof mAge);
		memset(mbDirty, 0, sizeof mbDirty);
		memset(mbLocked, 0, sizeof mbLocked);
		mClock = 0;
	}

	void RegisterTracker::Flush(x86::BasicBlock *bb, int esp_offset) {
		for(int i=0; i<8; ++i) {
			if (mReg[i] > 0)
				Flush(bb, mReg[i], esp_offset);
		}

		memset(mReg, 0, sizeof mReg);
		memset(mAge, 0, sizeof mAge);
		memset(mbDirty, 0, sizeof mbDirty);
		memset(mbLocked, 0, sizeof mbLocked);
		mClock = 0;
	}

	x86::varx RegisterTracker::Load(x86::BasicBlock *bb, int reg, bool markDirty, int esp_offset) {
		for(int i=0; i<8; ++i) {
			if (mReg[i] == reg) {
				mAge[i] = ++mClock;
				return x86::xmm(i);
			}
		}

		int j = Evict(bb, esp_offset);
		bb->movaps(x86::xmm(j), mpSlots->GetVariableSlot(reg, esp_offset));
		
		mReg[j] = reg;
		mbDirty[j] = markDirty;

		return x86::xmm(j);
	}

	x86::varx RegisterTracker::Allocate(x86::BasicBlock *bb, int reg, int esp_offset) {
		for(int i=0; i<8; ++i) {
			if (mReg[i] == reg) {
				mbLocked[i] = true;
				return x86::xmm(i);
			}
		}

		int j = Evict(bb, esp_offset);
		mReg[j] = reg;
		mbDirty[j] = true;

		return x86::xmm(j);
	}

	x86::varx RegisterTracker::AllocateAnon(x86::BasicBlock *bb, int esp_offset) {
		int j = Evict(bb, esp_offset);
		mReg[j] = -1;
		mbDirty[j] = false;

		return x86::xmm(j);
	}

	void RegisterTracker::Destroy(int reg) {
		for(int i=0; i<8; ++i) {
			if (mReg[i] == reg) {
				mReg[i] = 0;
				mAge[i] = 0;
				mbDirty[i] = false;
				mbLocked[i] = false;
			}
		}
	}

	void RegisterTracker::Flush(x86::BasicBlock *bb, int reg, int esp_offset) {
		for(int i=0; i<8; ++i) {
			if (mReg[i] == reg) {
				if (mbDirty[i]) {
					mbDirty[i] = false;
					bb->movaps(mpSlots->GetVariableSlot(reg, esp_offset), x86::xmm(i));
				}
				return;
			}
		}
	}

	int RegisterTracker::Evict(x86::BasicBlock *bb, int esp_offset) {
		int minage = 0x7fffffff;
		int cand = -1;

		for(int i=0; i<8; ++i) {
			if (!mReg[i]) {
				cand = i;
				break;
			}

			if (mbLocked[i])
				continue;

			if (mAge[i] < minage) {
				minage = mAge[i];
				cand = i;
			}
		}

		VDASSERT(cand >= 0);

		if (mbDirty[cand])
			bb->movaps(mpSlots->GetVariableSlot(mReg[cand], esp_offset), x86::xmm(cand));

		mbDirty[cand] = false;
		mReg[cand] = 0;
		mAge[cand] = ++mClock;
		mbLocked[cand] = true;

		return cand;
	}

	void RegisterTracker::Unlock(const x86::varx& var) {
		int i = var.val & 7;

		mbLocked[i] = false;

		if (mReg[i] == -1) {
			mReg[i] = 0;
			mAge[i] = 0;
			mbDirty[i] = false;
		}
	}

	void RegisterTracker::UnlockAndRetag(const x86::varx& var, int reg) {
		int j = var.val & 7;

		for(int i=0; i<8; ++i) {
			if (i == j || mReg[i] != reg)
				continue;

			mbLocked[i] = false;
			mAge[i] = 0;
			mbDirty[i] = false;
			mReg[i] = 0;
		}

		mbLocked[j] = false;
		mbDirty[j] = true;
		mReg[j] = reg;
	}
}

bool VDShaderSSE2JIT::LoadShader(const VDShaderFunction& func, const VDShaderBindingElement *pElements, int numElements, const VDShaderInterpConstant *pConstantInputs, int numConstantInputs, const VDShaderInterpreterInterpolator *interps, uint32 interpCount, IVDShaderLogOutput *out, bool showDisassembly) {
	if (mpShader) {
		VDFreeThunkMemory(mpShader, mShaderLen);
		mpShader = NULL;
	}

	// compile shader to scalar form
	mScalarForm.Clear();

	VDShaderScalarTransformer xform(mScalarForm);
	if (!xform.LoadShader(func, pElements, numElements, mpSamplers, mSamplerCount, pConstantInputs, numConstantInputs, interps, interpCount, 4, out, showDisassembly))
		return false;

	using namespace x86;
	DynamicContext ctx;

	BasicBlock *bbent = ctx.GetEntryBlock();
	BasicBlock *bbexit = ctx.GetExitBlock();

	int numConstants = mScalarForm.mConstants.size();
	mConstants.resize(numConstants * 4);

	const float *srcConstants = mScalarForm.mConstants.data();
	float *pConstants = mConstants.data();
	for(int i=0; i<numConstants; ++i) {
		pConstants[i*4+0] = srcConstants[i];
		pConstants[i*4+1] = srcConstants[i];
		pConstants[i*4+2] = srcConstants[i];
		pConstants[i*4+3] = srcConstants[i];
	}

	const int numTemps = mScalarForm.mTempCount;
	const int numInputs = mScalarForm.mInputCount;
	const int numSamplers = mSamplerCount;
	const uint32 tempBase = 96;
	const uint32 inputBase = tempBase + numTemps * 16;
	const uint32 samplerBase = inputBase + numInputs * 32;
	const uint32 outputBase = samplerBase + numSamplers * 32;
	const uint32 localSize = outputBase + 64;

	bbent->lea(ebp, dword_ptr [esp + 12]);
	bbent->and(esp, -16);
	bbent->sub(esp, localSize);

	// load count
	bbent->mov(esi, dword_ptr [ebp+16]);

	// copy interpolators
	bbent->mov(ebx, dword_ptr [ebp+12]);

	static const __declspec(align(16)) float kInterpScale[4]={ 0, 1, 2, 3 };

	bbent->movdqa(xmm3, xmmword_ptr [kInterpScale]);

	for(int i=0; i<numInputs; ++i) {
		bbent->movss(xmm1, dword_ptr [ebx + 4*i + 16]);
		bbent->pshufd(xmm1, xmm1, 0);
		bbent->movaps(xmm2, xmm1);
		bbent->addps(xmm1, xmm1);
		bbent->mulps(xmm2, xmm3);
		bbent->addps(xmm1, xmm1);
		bbent->movaps(xmmword_ptr [esp + inputBase + 16 + 32*i], xmm1);
		bbent->movss(xmm0, dword_ptr [ebx + 4*i]);
		bbent->pshufd(xmm0, xmm0, 0);
		bbent->addps(xmm0, xmm2);
		bbent->movaps(xmmword_ptr [esp + inputBase + 32*i], xmm0);
	}

	for(int i=0; i<numSamplers; ++i) {
		const VDShaderInterpreterSampler *samp = mpSamplers[i];

		if (!samp)
			continue;

		bbent->mov(eax, (sint32)samp->mpSrc);
		bbent->movd(xmm0, eax);
		bbent->pshufd(xmm0, xmm0, 0);
		bbent->movdqa(xmmword_ptr [esp + samplerBase + i*32], xmm0);

		bbent->mov(eax, (sint32)samp->mPitch);
		bbent->movd(xmm0, eax);
		bbent->pshufd(xmm0, xmm0, 0);
		bbent->movdqa(xmmword_ptr [esp + samplerBase + i*32 + 16], xmm0);
	}

	// load destination
	bbent->mov(edi, dword_ptr [ebp+8]);

	SlotManager local = { pConstants, numConstants, tempBase, inputBase, samplerBase, outputBase };
	RegisterTracker regTrack(&local);

	BasicBlock *bb = ctx.CreateBasicBlock();
	BasicBlock *bbloop = bb;
	bbent->jmp(bbloop);

	bool outputsSet[4] = { false };

	VDShaderScalarForm::Insns::const_iterator it(mScalarForm.mInsns.begin()), itEnd(mScalarForm.mInsns.end());
	for(; it!=itEnd; ++it) {
		const VDShaderInterpreterTransformedInsn& insn = *it;
		if (!insn.op)
			break;

		if (insn.op == kVDShaderInterpOp_beginloop) {
			regTrack.Flush(bb, 0);
			bbloop = ctx.CreateBasicBlock();
			bb->jmp(bbloop);
			bb = bbloop;
			continue;
		}

		if ((insn.dst & kVDShaderRegTypeMask) == kVDShaderRegO0) {
			VDASSERT(insn.op == kVDShaderInterpOp_mov);
			VDASSERT(insn.src[0]);

			outputsSet[insn.dst & 3] = true;
		}

		VDASSERT(!insn.dst || (insn.dst & kVDShaderRegTypeMask) == kVDShaderRegR0 || (insn.dst & kVDShaderRegTypeMask) == kVDShaderRegO0);
		switch(insn.op) {
			case kVDShaderInterpOp_texld:
				{
					const VDShaderInterpreterSampler& samp = *mpSamplers[insn.src[2] - kVDShaderRegS0];

					if (samp.mbBilinear) {
						static const __declspec(align(16)) uint32 k15[4] = { 0xF000F, 0xF000F, 0xF000F, 0xF000F };
						static const __declspec(align(16)) uint32 k32[4] = { 32, 32, 32, 32 };

						regTrack.Flush(bb, insn.src[1], 0);
						regTrack.Flush(bb, insn.src[0], 0);
						regTrack.Destroy(insn.dst);

						const varx iu = regTrack.AllocateAnon(bb, 0);
						const varx u = regTrack.Load(bb, insn.src[0], false, 0);
						bb->cvttps2dq(iu, u);
						bb->psubd(iu, xmmword_ptr [k32]);
						regTrack.Unlock(u);

						const varx iv = regTrack.AllocateAnon(bb, 0);
						const varx v = regTrack.Load(bb, insn.src[1], false, 0);
						bb->cvttps2dq(iv, v);
						bb->psubd(iv, xmmword_ptr [k32]);
						regTrack.Unlock(v);

						// compute u/v integer and fractional coordinates 
						const varx uf = regTrack.AllocateAnon(bb, 0);
						const varx vf = regTrack.AllocateAnon(bb, 0);
						bb->movdqa(uf, iu);
						bb->psrad(iu, 6);
						bb->psrad(uf, 2);
						bb->pshuflw(uf, uf, 0xA0);
						bb->pshufhw(uf, uf, 0xA0);
						bb->movdqa(xmmword_ptr [esp], iu);
						bb->pand(uf, xmmword_ptr [k15]);

						bb->movdqa(vf, iv);
						bb->psrad(iv, 6);
						bb->psrad(vf, 2);
						bb->pshuflw(vf, vf, 0xA0);
						bb->pshufhw(vf, vf, 0xA0);
						bb->movdqa(xmmword_ptr [esp+16], iv);
						bb->pand(vf, xmmword_ptr [k15]);

						// unpack and vlerp pixels 0-3
						const varx z = regTrack.AllocateAnon(bb, 0);
						const varx t = regTrack.AllocateAnon(bb, 0);
						bb->pxor(z, z);

						bb->imul(eax, dword_ptr [esp+16], samp.mPitch);
						bb->add(eax, dword_ptr [&samp.mpSrc]);
						bb->mov(ecx, dword_ptr [esp]);
						bb->pshufd(t, vf, 0);
						bb->movq(iu, qword_ptr [eax + ecx*4]);
						bb->punpcklbw(iu, z);
						bb->movq(iv, qword_ptr [eax + ecx*4 + samp.mPitch]);
						bb->punpcklbw(iv, z);
						bb->psubw(iv, iu);
						bb->psllw(iu, 4);
						bb->pmullw(iv, t);
						bb->paddw(iu, iv);
						bb->movlps(qword_ptr [esp+32], iu);
						bb->movhps(qword_ptr [esp+48], iu);

						bb->imul(eax, dword_ptr [esp+20], samp.mPitch);
						bb->add(eax, dword_ptr [&samp.mpSrc]);
						bb->mov(ecx, dword_ptr [esp+4]);
						bb->pshufd(t, vf, 0x55);
						bb->movq(iu, qword_ptr [eax + ecx*4]);
						bb->punpcklbw(iu, z);
						bb->movq(iv, qword_ptr [eax + ecx*4 + samp.mPitch]);
						bb->punpcklbw(iv, z);
						bb->psubw(iv, iu);
						bb->psllw(iu, 4);
						bb->pmullw(iv, t);
						bb->paddw(iu, iv);
						bb->movlps(qword_ptr [esp+40], iu);
						bb->movhps(qword_ptr [esp+56], iu);

						bb->imul(eax, dword_ptr [esp+24], samp.mPitch);
						bb->add(eax, dword_ptr [&samp.mpSrc]);
						bb->mov(ecx, dword_ptr [esp+8]);
						bb->pshufd(t, vf, 0xAA);
						bb->movq(iu, qword_ptr [eax + ecx*4]);
						bb->punpcklbw(iu, z);
						bb->movq(iv, qword_ptr [eax + ecx*4 + samp.mPitch]);
						bb->punpcklbw(iv, z);
						bb->psubw(iv, iu);
						bb->psllw(iu, 4);
						bb->pmullw(iv, t);
						bb->paddw(iu, iv);
						bb->movlps(qword_ptr [esp+64], iu);
						bb->movhps(qword_ptr [esp+80], iu);

						bb->imul(eax, dword_ptr [esp+28], samp.mPitch);
						bb->add(eax, dword_ptr [&samp.mpSrc]);
						bb->mov(ecx, dword_ptr [esp+12]);
						bb->pshufd(t, vf, 0xFF);
						bb->movq(iu, qword_ptr [eax + ecx*4]);
						bb->punpcklbw(iu, z);
						bb->movq(iv, qword_ptr [eax + ecx*4 + samp.mPitch]);
						bb->punpcklbw(iv, z);
						bb->psubw(iv, iu);
						bb->psllw(iu, 4);
						bb->pmullw(iv, t);
						bb->paddw(iu, iv);
						bb->movlps(qword_ptr [esp+72], iu);
						bb->movhps(qword_ptr [esp+88], iu);

						// hlerp pixel pairs and pack
						bb->movdqa(iu, xmmword_ptr [esp+32]);
						bb->pshufd(t, uf, 0x50);
						bb->movdqa(iv, xmmword_ptr [esp+48]);
						bb->psubw(iv, iu);
						bb->psllw(iu, 4);
						bb->pmullw(iv, t);
						bb->paddw(iu, iv);

						bb->movdqa(vf, xmmword_ptr [esp+64]);
						bb->pshufd(t, uf, 0xFA);
						bb->movdqa(z, xmmword_ptr [esp+80]);
						bb->psubw(z, vf);
						bb->psllw(vf, 4);
						bb->pmullw(z, t);
						bb->paddw(vf, z);

						bb->psrlw(iu, 7);
						bb->pxor(z, z);
						bb->psrlw(vf, 7);
						bb->pavgw(iu, z);
						bb->pavgw(vf, z);

						bb->packuswb(iu, vf);

						regTrack.Unlock(t);
						regTrack.Unlock(z);
						regTrack.Unlock(vf);
						regTrack.Unlock(uf);
						regTrack.Unlock(iv);
						regTrack.UnlockAndRetag(iu, insn.dst);
					} else {
						ptrdiff_t dstOffset = tempBase + (insn.dst - kVDShaderRegR0)*16;

						const varx tu = regTrack.AllocateAnon(bb, 0);
						const varx u = regTrack.Load(bb, insn.src[0], false, 0);

						bb->cvttps2dq(tu, u);
						regTrack.Unlock(u);


						const varx tv = regTrack.AllocateAnon(bb, 0);
						const varx v = regTrack.Load(bb, insn.src[1], false, 0);
						bb->cvttps2dq(tv, v);
						regTrack.Unlock(v);

						bb->pslld(tu, 2);
						bb->movaps(dword_ptr [esp], tv);
//						bb->paddd(tu, local.GetVariableSlot(insn.src[2]));
						bb->movaps(dword_ptr [esp+16], tu);

						regTrack.Unlock(tu);
						regTrack.Unlock(tv);
						regTrack.Destroy(insn.dst);

						bb->imul(eax, dword_ptr [esp], samp.mPitch);
						bb->add(eax, dword_ptr [&samp.mpSrc]);
						bb->mov(ecx, dword_ptr [esp+16]);
						bb->mov(eax, dword_ptr [eax+ecx]);
						bb->mov(dword_ptr [esp + dstOffset], eax);

						bb->imul(eax, dword_ptr [esp+4], samp.mPitch);
						bb->add(eax, dword_ptr [&samp.mpSrc]);
						bb->mov(ecx, dword_ptr [esp+20]);
						bb->mov(eax, dword_ptr [eax+ecx]);
						bb->mov(dword_ptr [esp + dstOffset + 4], eax);

						bb->imul(eax, dword_ptr [esp+8], samp.mPitch);
						bb->add(eax, dword_ptr [&samp.mpSrc]);
						bb->mov(ecx, dword_ptr [esp+24]);
						bb->mov(eax, dword_ptr [eax+ecx]);
						bb->mov(dword_ptr [esp + dstOffset + 8], eax);

						bb->imul(eax, dword_ptr [esp+12], samp.mPitch);
						bb->add(eax, dword_ptr [&samp.mpSrc]);
						bb->mov(ecx, dword_ptr [esp+28]);
						bb->mov(eax, dword_ptr [eax+ecx]);
						bb->mov(dword_ptr [esp + dstOffset + 12], eax);
					}
				}
				continue;
			case kVDShaderInterpOp_unpackr:
			case kVDShaderInterpOp_unpackg:
			case kVDShaderInterpOp_unpackb:
			case kVDShaderInterpOp_unpacka:
				{
					int idx = (insn.op - kVDShaderInterpOp_unpackb);

					const varx srcval = regTrack.Load(bb, insn.src[0], false, 0);
					const varx dst = regTrack.AllocateAnon(bb, 0);
					bb->movdqa(dst, srcval);
					regTrack.Unlock(srcval);

					if (idx)
						bb->psrld(dst, idx*8);

					if (idx != 3) {
						static const __declspec(align(16)) uint32 kMask255[4]={255, 255, 255, 255};
						bb->pand(dst, xmmword_ptr [kMask255]);
					}

					bb->cvtdq2ps(dst, dst);

					static const __declspec(align(16)) float kScaleInv255[4]={1.0f/255.0f, 1.0f/255.0f, 1.0f/255.0f, 1.0f/255.0f};
					bb->mulps(dst, xmmword_ptr [kScaleInv255]);

					regTrack.UnlockAndRetag(dst, insn.dst);
				}
				continue;
		}

		varx src1;
		varx src2;
		varx src3;
		varx dst;

		if (insn.src[0])
			src1 = regTrack.Load(bb, insn.src[0], false, 0);

		if (insn.src[1])
			src2 = regTrack.Load(bb, insn.src[1], false, 0);

		if (insn.src[2])
			src3 = regTrack.Load(bb, insn.src[2], false, 0);

		if (insn.dst)
			dst = regTrack.AllocateAnon(bb, 0);

		switch(insn.op) {
			case kVDShaderInterpOp_mov:
				bb->movaps(dst, src1);
				break;
			case kVDShaderInterpOp_add:
				bb->movaps(dst, src1);
				bb->addps(dst, src2);
				break;
			case kVDShaderInterpOp_sub:
				bb->movaps(dst, src1);
				bb->subps(dst, src2);
				break;
			case kVDShaderInterpOp_mul:
				bb->movaps(dst, src1);
				bb->mulps(dst, src2);
				break;
			case kVDShaderInterpOp_rsq:
				{
					static const __declspec(align(16)) float kThree[4]={ 3.0f, 3.0f, 3.0f, 3.0f };
					static const __declspec(align(16)) float kHalf[4]={ 0.5f, 0.5f, 0.5f, 0.5f };
					static const __declspec(align(16)) uint32 kDropSign[4]={ 0x7FFFFFFF, 0x7FFFFFFF, 0x7FFFFFFF, 0x7FFFFFFF };
					static const __declspec(align(16)) uint32 kClamp[4]={ 0x00800000, 0x00800000, 0x00800000, 0x00800000 };

					varx t = regTrack.AllocateAnon(bb, 0);
					varx u = regTrack.AllocateAnon(bb, 0);
					varx v = regTrack.AllocateAnon(bb, 0);

					bb->movaps(t, src1);
					bb->andps(t, xmmword_ptr [kDropSign]);
					bb->rsqrtps(dst, t);
					bb->movaps(v, xmmword_ptr [kClamp]);
					bb->maxps(t, v);
					bb->mulps(t, dst);
					bb->movaps(u, xmmword_ptr [kThree]);
					bb->mulps(t, dst);
					bb->subps(u, t);
					bb->maxps(u, v);
					bb->mulps(dst, xmmword_ptr [kHalf]);
					bb->mulps(dst, u);

					regTrack.Unlock(v);
					regTrack.Unlock(u);
					regTrack.Unlock(t);
				}
				break;
			case kVDShaderInterpOp_rcp:
				{
					static const __declspec(align(16)) float kTwo[4]={ 2.0f, 2.0f, 2.0f, 2.0f };

					varx t = regTrack.AllocateAnon(bb, 0);
					varx u = regTrack.AllocateAnon(bb, 0);

					bb->movaps(t, src1);
					bb->rcpps(dst, src1);
					bb->mulps(t, dst);
					bb->movaps(u, xmmword_ptr [kTwo]);
					bb->minps(t, u);
					bb->subps(u, t);
					bb->mulps(dst, u);

					regTrack.Unlock(u);
					regTrack.Unlock(t);
				}
				break;
			case kVDShaderInterpOp_frc:
				{
					static const __declspec(align(16)) uint32 kDropSign[4]={ 0x7FFFFFFF, 0x7FFFFFFF, 0x7FFFFFFF, 0x7FFFFFFF };
					static const __declspec(align(16)) uint32 kThreshold[4]={ 0x4b000000, 0x4b000000, 0x4b000000, 0x4b000000 };
					static const __declspec(align(16)) uint32 kCorrect[4]={ 0x3f800000, 0x3f800000, 0x3f800000, 0x3f800000 };

					varx t = regTrack.AllocateAnon(bb, 0);
					varx u = regTrack.AllocateAnon(bb, 0);
					varx v = regTrack.AllocateAnon(bb, 0);

					bb->movaps(t, src1);
					bb->cvtps2dq(v, src1);
					bb->movaps(u, src1);
					bb->pand(t, xmmword_ptr [kDropSign]);
					bb->pcmpgtd(t, xmmword_ptr [kThreshold]);
					bb->cvtdq2ps(v, v);
					bb->subps(u, v);
					bb->pandn(t, u);
					bb->movaps(dst, t);
					bb->pxor(v, v);
					bb->cmpltps(dst, v);
					bb->andps(dst, xmmword_ptr [kCorrect]);
					bb->addps(dst, t);

					regTrack.Unlock(v);
					regTrack.Unlock(u);
					regTrack.Unlock(t);
				}
				break;
			case kVDShaderInterpOp_neg:
				{
					static const __declspec(align(16)) uint32 kNegate[4]={ 0x80000000, 0x80000000, 0x80000000, 0x80000000 };
					bb->movaps(dst, src1);
					bb->xorps(dst, xmmword_ptr [kNegate]);
				}
				break;
			case kVDShaderInterpOp_min:
				bb->movaps(dst, src1);
				bb->minps(dst, src2);
				break;
			case kVDShaderInterpOp_max:
				bb->movaps(dst, src1);
				bb->maxps(dst, src2);
				break;
			case kVDShaderInterpOp_abs:
				{
					static const __declspec(align(16)) uint32 kDropSign[4]={ 0x7fffffff, 0x7fffffff, 0x7fffffff, 0x7fffffff };
					bb->movaps(dst, src1);
					bb->andps(dst, xmmword_ptr [kDropSign]);
				}
				break;
			case kVDShaderInterpOp_log:
				{
					static const __declspec(align(16)) uint32 kExpMask[4]={ 0x7f800000, 0x7f800000, 0x7f800000, 0x7f800000 };
					static const __declspec(align(16)) uint32 kMantMask[4]={ 0x007fffff, 0x007fffff, 0x007fffff, 0x007fffff };

					varx t = regTrack.AllocateAnon(bb, 0);
					varx u = regTrack.AllocateAnon(bb, 0);
					varx v = regTrack.AllocateAnon(bb, 0);
					varx w = regTrack.AllocateAnon(bb, 0);

					bb->movaps(u, xmmword_ptr [kMantMask]);
					bb->movaps(t, xmmword_ptr [kExpMask]);
					bb->andps(u, src1);
					bb->andps(t, src1);
					bb->cvtdq2ps(u, u);
					bb->psrld(t, 23);
					bb->cvtdq2ps(t, t);
					
#define CONSTANT(name, value) static const __declspec(align(16)) float name[4] = { (value), (value), (value), (value) }
					CONSTANT(kc0, 2.7927758f);
					CONSTANT(kc2, 0.96862515f);
					CONSTANT(kc3, -0.003397648f);
#undef CONSTANT
					// compute change of variable: y = x/(2+x)
					static const __declspec(align(16)) float kTwoScaled[4]={ 16777216.0f, 16777216.0f, 16777216.0f, 16777216.0f };
					static const __declspec(align(16)) float kTwo[4]={ 2.0f, 2.0f, 2.0f, 2.0f };
					bb->movaps(dst, xmmword_ptr [kTwoScaled]);
					bb->movaps(v, u);
					bb->addps(u, dst);					// u = x+2
					bb->rcpps(w, u);					// w ~= 1/(2+x)
					bb->movaps(dst, xmmword_ptr [kTwo]);
					bb->mulps(u, w);
					bb->subps(dst, u);
					bb->mulps(w, dst);					// w = 1/(2+x)
					bb->mulps(w, v);					// w = y = x/(2+x)

					// compute Pade approximation y*(c0 + c1*y*y)/(c2 + c3*y*y)		(c1=1)
					bb->movaps(dst, w);					// dst = y
					bb->mulps(w, w);					// w = y*y
					bb->movaps(u, w);					// u = y*y
					bb->addps(w, xmmword_ptr [kc0]);	// w = c0 + c1*y*y
					bb->mulps(dst, w);					// dst = y*(c0 + c1*y*y)
					bb->mulps(u, xmmword_ptr [kc3]);	// u = c3*y*y
					bb->addps(u, xmmword_ptr [kc2]);	// u = c2 + c3*y*y
					bb->movaps(w, xmmword_ptr [kTwo]);
					bb->rcpps(v, u);					// v ~= 1/(c2 + c3*y*y)
					bb->mulps(u, v);
					bb->subps(w, u);
					bb->mulps(v, w);					// v = 1/(c2 + c3*y*y)
					bb->mulps(dst, v);					// dst = y*(c0 + c1*y*y)/(c2 + c3*y*y)

					// merge exponent
					static const __declspec(align(16)) float kCorrect[4]={ 127.0f, 127.0f, 127.0f, 127.0f };
					bb->subps(t, xmmword_ptr [kCorrect]);
					bb->addps(dst, t);

					regTrack.Unlock(w);
					regTrack.Unlock(v);
					regTrack.Unlock(u);
					regTrack.Unlock(t);
				}
				break;

			case kVDShaderInterpOp_exp:
				{
#define CONSTANT(name, value) static const __declspec(align(16)) float name[4] = { (value), (value), (value), (value) }

					CONSTANT(kc0, 1.0000000);
					CONSTANT(kc1, 0.6931529);
					CONSTANT(kc2, 0.2401517);
					CONSTANT(kc3, 0.0558332);
					CONSTANT(kc4, 0.0089823);
					CONSTANT(kc5, 0.0018797);

#undef CONSTANT

					static const __declspec(align(16)) uint32 kibias[4] = { 127, 127, 127, 127 };
					static const __declspec(align(16)) float kmin[4] = { -127, -127, -127, -127 };
					static const __declspec(align(16)) float kmax[4] = { 128, 128, 128, 128 };
					static const __declspec(align(16)) uint32 kAndMask[4] = { 0x007FFFFF, 0x007FFFFF, 0x007FFFFF, 0x007FFFFF };
					static const __declspec(align(16)) uint32 kOrMask[4] = { 0x3F800000, 0x3F800000, 0x3F800000, 0x3F800000 };

					varx t = regTrack.AllocateAnon(bb, 0);

					bb->movaps(t, xmmword_ptr [kmin]);
					bb->maxps(t, src1);
					bb->minps(t, xmmword_ptr [kmax]);
					bb->cvttps2dq(t, t);
					bb->cvtdq2ps(dst, t);
					bb->subps(src1, dst);
					bb->paddd(t, xmmword_ptr [kibias]);
					bb->movaps(dst, xmmword_ptr [kc5]);
					bb->pslld(t, 23);
					bb->mulps(dst, src1);
					bb->addps(dst, xmmword_ptr [kc4]);
					bb->mulps(dst, src1);
					bb->addps(dst, xmmword_ptr [kc3]);
					bb->mulps(dst, src1);
					bb->addps(dst, xmmword_ptr [kc2]);
					bb->mulps(dst, src1);
					bb->addps(dst, xmmword_ptr [kc1]);
					bb->mulps(dst, src1);
					bb->addps(dst, xmmword_ptr [kc0]);
					bb->mulps(dst, t);

					regTrack.Unlock(t);
				}
				break;
			case kVDShaderInterpOp_cmp:
				{
					varx t = regTrack.AllocateAnon(bb, 0);

					bb->movaps(dst, src1);
					bb->xorps(t, t);
					bb->cmpltps(dst, t);
					bb->movaps(t, src3);
					bb->andps(t, dst);
					bb->andnps(dst, src2);
					bb->orps(dst, t);

					regTrack.Unlock(t);
				}
				break;
		}

		if (insn.src[0])
			regTrack.Unlock(src1);

		if (insn.src[1])
			regTrack.Unlock(src2);

		if (insn.src[2])
			regTrack.Unlock(src3);

		if (insn.dst)
			regTrack.UnlockAndRetag(dst, insn.dst);
	}

	regTrack.Flush(bb, 0);

	for(int i=0; i<4; ++i) {
		varx dst = xmm(i);

		if (outputsSet[i])
			bb->movaps(dst, local.GetVariableSlot(kVDShaderRegO0 + i));
		else
			bb->xorps(dst, dst);
	}

	static const __declspec(align(16)) float kScale[4] = { 255.0f, 255.0f, 255.0f, 255.0f };
	bb->movaps(xmm4, xmmword_ptr [kScale]);
	bb->mulps(xmm0, xmm4);
	bb->mulps(xmm1, xmm4);
	bb->mulps(xmm2, xmm4);
	bb->mulps(xmm3, xmm4);
	bb->cvtps2dq(xmm0, xmm0);
	bb->cvtps2dq(xmm1, xmm1);
	bb->cvtps2dq(xmm2, xmm2);
	bb->cvtps2dq(xmm3, xmm3);
	bb->packssdw(xmm0, xmm0);
	bb->packssdw(xmm1, xmm1);
	bb->packssdw(xmm2, xmm2);
	bb->packssdw(xmm3, xmm3);
	bb->packuswb(xmm0, xmm0);
	bb->packuswb(xmm1, xmm1);
	bb->packuswb(xmm2, xmm2);
	bb->packuswb(xmm3, xmm3);
	bb->punpcklbw(xmm2, xmm1);
	bb->punpcklbw(xmm0, xmm3);
	bb->sub(esi, 4);
	bb->punpcklwd(xmm2, xmm0);

	BasicBlock *bbleftovers = ctx.CreateBasicBlock();
	bb->js(bbleftovers);

	BasicBlock *bbloopend = ctx.CreateBasicBlock();
	bb->jmp(bbloopend);

	bb = bbloopend;

	bb->movups(xmmword_ptr [edi], xmm2);

	bb->lea(edi, dword_ptr [edi+16]);
	bb->jnz(bbloop);

	BasicBlock *bboutro = ctx.CreateBasicBlock();
	bb->jmp(bboutro);

	bbleftovers->add(esi, 4);

	BasicBlock *bbleftloop = ctx.CreateBasicBlock();
	bbleftovers->jmp(bbleftloop);

	bbleftloop->movd(dword_ptr [edi], xmm2);
	bbleftloop->add(edi, 4);
	bbleftloop->pshufd(xmm2, xmm2, 0xF9);
	bbleftloop->sub(esi, 1);
	bbleftloop->jnz(bbleftloop);
	bbleftloop->jmp(bboutro);

	bboutro->lea(esp, dword_ptr [ebp-12]);
	bboutro->jmp(bbexit);

	ctx.Generate();

	uint32 codelen = ctx.GetCodeLen();
	mShaderLen = codelen;
	mpShader = (Shader)VDAllocateThunkMemory(codelen);
	if (mpShader) {
		ctx.Copy(mpShader);
		VDFlushThunkMemory(mpShader, codelen);

		if (out && showDisassembly) {
			out->WriteLogOutput("//\n");
			out->WriteLogOutput("// Generated by VirtualDub vector x86 SSE2 Compiler\n");
			out->WriteLogOutput("//\n");

			VDDisassemblyContext ctx;
			ctx.Init(g_VDDisasmX86);

			const unsigned char *sbase = (const unsigned char *)mpShader;
			const unsigned char *s = sbase;
			const unsigned char *send = sbase + codelen;

			ctx.SetPhysToVirtOffset(-(ptrdiff_t)sbase);

			while(s != send) {
				int count;
				const char *text = ctx.Disassemble(s, send - s, count);

				VDShaderWriteLogOutputF(*out, "%4x: %s\n", s - sbase, text);
				s += count;
			}
		}
	}

	return true;
}

void VDShaderSSE2JIT::LoadConstants(const void *data) {
	for(VDShaderScalarForm::ConstantUploads::const_iterator it(mScalarForm.mConstantUploads.begin()), itEnd(mScalarForm.mConstantUploads.end()); it != itEnd; ++it) {
		const VDShaderScalarForm::ConstantUploadEntry& cent = *it;
		float *dst = &mConstants[cent.mDstIdx * 4];
		const float v = *(const float *)((const char *)data + cent.mSrcOffset);

		dst[0] = v;
		dst[1] = v;
		dst[2] = v;
		dst[3] = v;
	}
}
