//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "parser.h"
#include "disasm.h"
#include "interp.h"
#include "intrinsics.h"
#include "vecarith.h"
#include <vd2/system/vdalloc.h>

using namespace nsVDShaderTokens;
using namespace nsVDShaderParser;
using namespace nsVDShaderArith;

namespace {
	int GetBinaryResultType(int srcType, int dstType) {
		static const int kTypeLookup[4][4]={
			kTypeFloat, kTypeFloat2, kTypeFloat3, kTypeFloat4,
			kTypeInvalid, kTypeFloat2x2, kTypeFloat2x3, kTypeFloat2x4,
			kTypeInvalid, kTypeFloat3x2, kTypeFloat3x3, kTypeFloat3x4,
			kTypeInvalid, kTypeFloat4x2, kTypeFloat4x3, kTypeFloat4x4,
		};

		int srcRows = kRowCountForType[srcType];
		int srcCols = kColCountForType[srcType];
		int dstRows = kRowCountForType[dstType];
		int dstCols = kColCountForType[dstType];

		if ((!srcRows && !srcCols) || (!dstRows && !dstCols))
			return 0;

		int maxRows = srcRows > dstRows ? srcRows : dstRows;
		int maxCols = srcCols > dstCols ? srcCols : dstCols;

		return kTypeLookup[maxRows-1][maxCols-1];
	}
}

const char *VDShaderGetTypeName(int type) {
	switch(type) {
	case kTypeVoid:		return "void";
	case kTypeVertexShader:	return "vertexshader";
	case kTypePixelShader:	return "pixelshader";
	case kTypeFloat:	return "float";
	case kTypeFloat2:	return "float2";
	case kTypeFloat3:	return "float3";
	case kTypeFloat4:	return "float4";
	case kTypeFloat2x2:	return "float2x2";
	case kTypeFloat2x3:	return "float2x3";
	case kTypeFloat2x4:	return "float2x4";
	case kTypeFloat3x2:	return "float3x2";
	case kTypeFloat3x3:	return "float3x3";
	case kTypeFloat3x4:	return "float3x4";
	case kTypeFloat4x2:	return "float4x2";
	case kTypeFloat4x3:	return "float4x3";
	case kTypeFloat4x4:	return "float4x4";
	case kTypeSampler:	return "sampler";
	case kTypeTexture:	return "texture";
	case kTypeBool:		return "bool";
	case kTypeBool2:	return "bool2";
	case kTypeBool3:	return "bool3";
	case kTypeBool4:	return "bool4";
	case kTypeString:	return "string";
	default:			return "<invalid type>";
	}
}

///////////////////////////////////////////////////////////////////////////////
VDShaderParserFunction::VDShaderParserFunction() {
}

VDShaderParserFunction::~VDShaderParserFunction() {
}

int VDShaderParserFunction::AllocateParamRegisters(int rows) {
	int base = mParameterRegisterCount;

	mParameterRegisterCount += rows;

	return base + kVDShaderRegV0;
}

int VDShaderParserFunction::AllocateRegisters(int rows) {
	int base = mRegisterCount;

	mRegisterCount += rows;

	return base + kVDShaderRegR0;
}

int VDShaderParserFunction::AddConstantZero() {
	VDShaderConstant zero = {0};
	return AddConstant(zero);
}

int VDShaderParserFunction::AddConstant(const VDShaderConstant& c) {
	size_t h = VDShaderConstantHash()(c);

	std::pair<ConstantMap::iterator, ConstantMap::iterator> r(mConstantMap.equal_range(h));
	for(; r.first != r.second; ++r.first) {
		const VDShaderConstant& c2 = mConstants[r.first->second];

		if (c == c2)
			return kVDShaderRegC0 + r.first->second;
	}

	int idx = (int)mConstants.size();
	mConstantMap.insert(ConstantMap::value_type(h, idx));

	mConstants.push_back(c);
	return kVDShaderRegC0 + idx;
}

int VDShaderParserFunction::AddMappedConstant(const VDShaderConstantBinding& binding, const VDShaderConstant *defaultValue) {
	for(ConstantBindings::const_iterator it(mConstantBindings.begin()), itEnd(mConstantBindings.end()); it != itEnd; ++it) {
		const VDShaderConstantBinding& candBinding = *it;

		if (candBinding.mNameToken == binding.mNameToken)
			return candBinding.mRegister;
	}

	VDShaderConstantBinding newBinding(binding);
	newBinding.mRegister = kVDShaderRegC0 + mConstants.size();
	mConstantBindings.push_back(newBinding);

	mConstants.insert(mConstants.end(), defaultValue, defaultValue + newBinding.mSize);

	return newBinding.mRegister;
}

int VDShaderParserFunction::PromoteToConstantRegisters(VDShaderParserVariable& var) {
	if (var.mbIsConst) {
		VDASSERT(var.mType != kTypeSampler);

		int rows = kRowCountForType[var.mType];

		int c = kVDShaderRegC0 + (int)mConstants.size();

		mConstants.insert(mConstants.end(), &var.m[0], &var.m[rows]);

		var.mRegister = c;
		var.mbIsConst = false;
	} else if (var.mType == kTypeSampler) {
		SamplerBindings::const_iterator it(std::find(mSamplerBindings.begin(), mSamplerBindings.end(), var.mSamplerBinding));

		if (it != mSamplerBindings.end()) {
			var.mRegister = kVDShaderRegS0 + (it - mSamplerBindings.begin());
		} else {
			var.mRegister = kVDShaderRegS0 + mSamplerCount++;
			mSamplerBindings.push_back(var.mSamplerBinding);
		}
	} else if (var.mbUnassigned) {
		VDASSERT(!var.mRegister || (var.mRegister & kVDShaderRegTypeMask) == kVDShaderRegC0);

		VDShaderConstantBinding binding;

		binding.mNameToken = var.mNameToken;
		binding.mRegister = 0;
		binding.mForcedRegister = var.mRegister;
		binding.mSize = kRowCountForType[var.mType];
		binding.mbExtern = var.mbExtern;

		var.mRegister = AddMappedConstant(binding, var.m);
		var.mbUnassigned = false;
	}

	return var.mRegister;
}

bool VDShaderParserFunction::BakeHardcodedRegisters(IVDShaderErrorOutput& errors) {
	if (mConstantBindings.empty())
		return true;

	vdfastvector<const VDShaderConstantBinding *> newRegisterToBindingLookup;
	vdfastvector<bool> newRegistersAssigned;
	vdfastvector<int> registerMap(mConstants.size(), -1);
	vdfastvector<VDShaderConstant> newConstants;
	int newRegisterCount = 0;
	int newRegisterSearchIdx = 0;

	// bake hardcoded bindings first, then un-hardcoded
	for(int pass = 0; pass < 2; ++pass) {
		int bindingIdx = 0;
		for(ConstantBindings::iterator it(mConstantBindings.begin()), itEnd(mConstantBindings.end()); it != itEnd; ++it, ++bindingIdx) {
			VDShaderConstantBinding& binding = *it;
			int newRegIdxStart = 0;

			if (binding.mForcedRegister) {
				if (pass)
					continue;

				VDASSERT((binding.mForcedRegister & kVDShaderRegTypeMask) == kVDShaderRegC0);
				newRegIdxStart = binding.mForcedRegister - kVDShaderRegC0;
			} else {
				if (!pass)
					continue;

				int searchLimit = newRegisterCount + 1 - binding.mSize;
				int length = 0;
				while(newRegisterSearchIdx < newRegisterCount) {
					if (newRegistersAssigned[newRegisterSearchIdx++])
						length = 0;
					else {
						if (++length >= binding.mSize)
							break;
					}
				}

				newRegIdxStart = newRegisterSearchIdx - length;
			}

			int oldRegIdxStart = binding.mRegister & kVDShaderRegIndexMask;
			int oldRegIdxEnd = oldRegIdxStart + binding.mSize;
			int newRegIdxEnd = newRegIdxStart + binding.mSize;

			if (newRegisterCount < newRegIdxEnd) {
				newRegisterCount = newRegIdxEnd;
				newRegisterToBindingLookup.resize(newRegisterCount, NULL);
				newRegistersAssigned.resize(newRegisterCount, false);

				const VDShaderConstant zero = {0};
				newConstants.resize(newRegisterCount, zero);
			}

			for(int i = 0; i < binding.mSize; ++i) {
				int newRegisterIdx = newRegIdxStart + i;
				if (newRegisterToBindingLookup[newRegisterIdx])
					return errors.SetErrorF("Conflicting variables explicitly bound to register c%d", newRegisterIdx);

				newRegisterToBindingLookup[newRegisterIdx] = &binding;
				newRegistersAssigned[newRegisterIdx] = true;

				registerMap[oldRegIdxStart + i] = newRegisterIdx;
				newConstants[newRegIdxStart + i] = mConstants[oldRegIdxStart + i];
			}

			binding.mRegister = newRegIdxStart + kVDShaderRegC0;
		}
	}

	// rename constant registers
	newRegisterSearchIdx = 0;
	for(Insns::iterator it(mInsns.begin()), itEnd(mInsns.end()); it!=itEnd; ++it) {
		VDShaderInsn& insn = *it;

		for(int i=0; i<4; ++i) {
			VDShaderInsnSource& src = insn.mSources[i];

			if ((src.mRegister & kVDShaderRegTypeMask) == kVDShaderRegC0) {
				int oldIdx = src.mRegister & kVDShaderRegIndexMask;
				int newIdx = registerMap[oldIdx];

				if (newIdx < 0) {
					while(newRegisterSearchIdx < newRegisterCount) {
						if (!newRegistersAssigned[newRegisterSearchIdx])
							break;

						++newRegisterSearchIdx;
					}

					if (newRegisterSearchIdx >= newRegisterCount) {
						newConstants.push_back();
						newRegistersAssigned.push_back(false);
						++newRegisterCount;
					}

					newIdx = newRegisterSearchIdx;

					registerMap[oldIdx] = newIdx;
					newConstants[newIdx] = mConstants[oldIdx];
					newRegistersAssigned[newIdx] = true;
				}

				src.mRegister = kVDShaderRegC0 + newIdx;
			}
		}
	}

	mConstants.swap(newConstants);

	return true;
}

void VDShaderParserFunction::mov(int dst, unsigned char mask, int src1) {
	VDASSERT(src1);
	insn(kVDShaderOp_mov, mask, dst, src1, 0xE4, 0, 0xE4);
}

void VDShaderParserFunction::mov_swizzle(int dst, unsigned char mask, int src1, unsigned char swizzle) {
	VDASSERT(src1);
	insn(kVDShaderOp_mov, mask, dst, src1, swizzle, 0, 0xE4);
}

void VDShaderParserFunction::mov_mods(int dst, unsigned char mask, int src1, unsigned char swizzle, unsigned char mods) {
	VDASSERT(src1);
	insn_mod(kVDShaderOp_mov, mask, dst, src1, swizzle, mods);
}

void VDShaderParserFunction::mov_sat(int dst, unsigned char mask, int src1) {
	VDASSERT(src1);
	insn(kVDShaderOp_mov, mask, dst, src1, 0xE4, 0, 0xE4);
	mInsns.back().mModifiers = kVDShaderInsnMod_Saturate;
}

void VDShaderParserFunction::abs(int dst, unsigned char mask, int src1) {
	insn(kVDShaderOp_abs, mask, dst, src1, 0xE4, 0, 0xE4);
}

void VDShaderParserFunction::add(int dst, unsigned char mask, int src1, int src2) {
	insn(kVDShaderOp_add, mask, dst, src1, 0xE4, src2, 0xE4);
}

void VDShaderParserFunction::add_swizzle_mods(int dst, unsigned char mask, int src1, unsigned char src1swizzle, unsigned char src1mods, int src2, unsigned char src2swizzle, unsigned char src2mods) {
	insn_mod(kVDShaderOp_add, mask, dst, src1, src1swizzle, src1mods, src2, src2swizzle, src2mods);
}

void VDShaderParserFunction::cmp(int dst, unsigned char mask, int src1, unsigned char src1swizzle, int src2, int src3) {
	insn(kVDShaderOp_cmp, mask, dst, src1, src1swizzle, src2, 0xE4, src3, 0xE4);
}

void VDShaderParserFunction::cmp_swizzle_mods(int dst, unsigned char mask, int src1, unsigned char src1swizzle, unsigned char src1mods, int src2, unsigned char src2swizzle, unsigned char src2mods, int src3, unsigned char src3swizzle, unsigned char src3mods) {
	insn_mod(kVDShaderOp_cmp, mask, dst, src1, src1swizzle, src1mods, src2, src2swizzle, src2mods, src3, src3swizzle, src3mods);
}

void VDShaderParserFunction::crs(int dst, unsigned char mask, int src1, int src2) {
	insn(kVDShaderOp_crs, mask, dst, src1, 0xE4, src2, 0xE4);
}

void VDShaderParserFunction::sub(int dst, unsigned char mask, int src1, int src2) {
	insn(kVDShaderOp_sub, mask, dst, src1, 0xE4, src2, 0xE4);
}

void VDShaderParserFunction::mul(int dst, unsigned char mask, int src1, int src2) {
	insn(kVDShaderOp_mul, mask, dst, src1, 0xE4, src2, 0xE4);
}

void VDShaderParserFunction::mul_swizzle(int dst, unsigned char mask, int src1, unsigned char src1swizzle, int src2, unsigned char src2swizzle) {
	insn(kVDShaderOp_mul, mask, dst, src1, src1swizzle, src2, src2swizzle);
}

void VDShaderParserFunction::mad_swizzle(int dst, unsigned char mask, int src1, unsigned char src1swizzle, int src2, unsigned char src2swizzle, int src3, unsigned char src3swizzle) {
	insn(kVDShaderOp_mad, mask, dst, src1, src1swizzle, src2, src2swizzle, src3, src3swizzle);
}

void VDShaderParserFunction::mad_swizzle_mods(int dst, unsigned char mask, int src1, unsigned char src1swizzle, unsigned char src1mods, int src2, unsigned char src2swizzle, unsigned char src2mods, int src3, unsigned char src3swizzle, unsigned char src3mods) {
	insn_mod(kVDShaderOp_mad, mask, dst, src1, src1swizzle, src1mods, src2, src2swizzle, src2mods, src3, src3swizzle, src3mods);
}

void VDShaderParserFunction::dp2add(int dst, unsigned char mask, int src1, int src2, int src3, unsigned char src3swizzle) {
	insn(kVDShaderOp_dp2add, mask, dst, src1, 0xE4, src2, 0xE4, src3, src3swizzle);
}

void VDShaderParserFunction::dp3(int dst, unsigned char mask, int src1, int src2) {
	insn(kVDShaderOp_dp3, mask, dst, src1, 0xE4, src2, 0xE4);
}

void VDShaderParserFunction::dp4(int dst, unsigned char mask, int src1, int src2) {
	insn(kVDShaderOp_dp4, mask, dst, src1, 0xE4, src2, 0xE4);
}

void VDShaderParserFunction::rsq(int dst, unsigned char mask, int src1) {
	insn(kVDShaderOp_rsq, mask, dst, src1, 0xE4, 0, 0xE4);
}

void VDShaderParserFunction::rsq(int dst, unsigned char mask, int src1, unsigned char src1component) {
	insn(kVDShaderOp_rsq, mask, dst, src1, src1component, 0, 0xE4);
}

void VDShaderParserFunction::rcp(int dst, unsigned char mask, int src1) {
	insn(kVDShaderOp_rcp, mask, dst, src1, 0xE4, 0, 0xE4);
}

void VDShaderParserFunction::rcp(int dst, unsigned char mask, int src1, unsigned char src1component) {
	insn(kVDShaderOp_rcp, mask, dst, src1, src1component, 0, 0xE4);
}

void VDShaderParserFunction::frc(int dst, unsigned char mask, int src1, unsigned char src1component) {
	insn(kVDShaderOp_frc, mask, dst, src1, src1component, 0, 0xE4);
}

void VDShaderParserFunction::lrp(int dst, unsigned char mask, int src1, int src2, int src3) {
	insn(kVDShaderOp_lrp, mask, dst, src1, 0xE4, src2, 0xE4, src3, 0xE4);
}

void VDShaderParserFunction::texld(int dst, unsigned char mask, int src1, unsigned char src1component, int src2) {
	insn(kVDShaderOp_texld, mask, dst, src1, src1component, src2, 0xE4);
}

void VDShaderParserFunction::min(int dst, unsigned char mask, int src1, int src2) {
	insn(kVDShaderOp_min, mask, dst, src1, 0xE4, src2, 0xE4);
}

void VDShaderParserFunction::max(int dst, unsigned char mask, int src1, int src2) {
	insn(kVDShaderOp_max, mask, dst, src1, 0xE4, src2, 0xE4);
}

void VDShaderParserFunction::nrm(int dst, unsigned char mask, int src1) {
	insn(kVDShaderOp_nrm, mask, dst, src1, 0xE4, 0, 0xE4);
}

void VDShaderParserFunction::log(int dst, unsigned char mask, int src1) {
	insn(kVDShaderOp_log, mask, dst, src1, 0xE4, 0, 0xE4);
}

void VDShaderParserFunction::log_swizzle(int dst, unsigned char mask, int src1, unsigned char src1component) {
	insn(kVDShaderOp_log, mask, dst, src1, src1component, 0, 0xE4);
}

void VDShaderParserFunction::exp(int dst, unsigned char mask, int src1) {
	insn(kVDShaderOp_exp, mask, dst, src1, 0xE4, 0, 0xE4);
}

void VDShaderParserFunction::exp_swizzle(int dst, unsigned char mask, int src1, unsigned char src1component) {
	insn(kVDShaderOp_exp, mask, dst, src1, src1component, 0, 0xE4);
}

void VDShaderParserFunction::sincos(int dst, unsigned char mask, int src0, unsigned char src0component) {
	static const VDShaderConstant kSinCosConst0 = { -1.5500992e-006f, -2.1701389e-005f,  0.0026041667f, 0.00026041668f };
	static const VDShaderConstant kSinCosConst1 = { -0.020833334f, -0.12500000f, 1.0f, 0.50000000f };

	int c0 = AddConstant(kSinCosConst0);
	int c1 = AddConstant(kSinCosConst1);

	insn(kVDShaderOp_sincos, mask, dst, src0, (src0component & 3) * 0x55, c0, 0xE4, c1, 0xE4);
}

void VDShaderParserFunction::insn_mod(int op, unsigned char mask, int dst, int src1, unsigned char src1swizzle, unsigned char src1mod) {
	VDShaderInsn insn;

	insn.mOp = op;
	insn.mWriteMask = mask;
	insn.mModifiers = 0;
	insn.mDestRegister = dst;
	insn.mSources[0].mRegister = src1;
	insn.mSources[0].mSwizzle = src1swizzle;
	insn.mSources[0].mModifiers = src1mod;
	insn.mSources[1].mRegister = 0;
	insn.mSources[1].mSwizzle = 0;
	insn.mSources[1].mModifiers = 0;
	insn.mSources[2].mRegister = 0;
	insn.mSources[2].mSwizzle = 0;
	insn.mSources[2].mModifiers = 0;
	insn.mSources[3].mRegister = 0;
	insn.mSources[3].mSwizzle = 0;
	insn.mSources[3].mModifiers = 0;

	VDASSERT((insn.mDestRegister & kVDShaderRegTypeMask) != kVDShaderRegR0 || (insn.mDestRegister - kVDShaderRegR0) < mRegisterCount);
	VDASSERT((insn.mSources[0].mRegister & kVDShaderRegTypeMask) != kVDShaderRegC0 || (insn.mSources[0].mRegister - kVDShaderRegC0) < (int)mConstants.size());
	VDASSERT((insn.mSources[1].mRegister & kVDShaderRegTypeMask) != kVDShaderRegC0 || (insn.mSources[1].mRegister - kVDShaderRegC0) < (int)mConstants.size());
	VDASSERT(insn.mSources[0].mRegister || !insn.mSources[1].mRegister);

	mInsns.push_back(insn);
}

void VDShaderParserFunction::insn_mod(int op, unsigned char mask, int dst, int src1, unsigned char src1swizzle, unsigned char src1mod, int src2, unsigned char src2swizzle, unsigned char src2mod) {
	VDShaderInsn insn;

	insn.mOp = op;
	insn.mWriteMask = mask;
	insn.mModifiers = 0;
	insn.mDestRegister = dst;
	insn.mSources[0].mRegister = src1;
	insn.mSources[0].mSwizzle = src1swizzle;
	insn.mSources[0].mModifiers = src1mod;
	insn.mSources[1].mRegister = src2;
	insn.mSources[1].mSwizzle = src2swizzle;
	insn.mSources[1].mModifiers = src2mod;
	insn.mSources[2].mRegister = 0;
	insn.mSources[2].mSwizzle = 0;
	insn.mSources[2].mModifiers = 0;
	insn.mSources[3].mRegister = 0;
	insn.mSources[3].mSwizzle = 0;
	insn.mSources[3].mModifiers = 0;

	VDASSERT((insn.mDestRegister & kVDShaderRegTypeMask) != kVDShaderRegR0 || (insn.mDestRegister - kVDShaderRegR0) < mRegisterCount);
	VDASSERT((insn.mSources[0].mRegister & kVDShaderRegTypeMask) != kVDShaderRegC0 || (insn.mSources[0].mRegister - kVDShaderRegC0) < (int)mConstants.size());
	VDASSERT((insn.mSources[1].mRegister & kVDShaderRegTypeMask) != kVDShaderRegC0 || (insn.mSources[1].mRegister - kVDShaderRegC0) < (int)mConstants.size());
	VDASSERT(insn.mSources[0].mRegister || !insn.mSources[1].mRegister);

	mInsns.push_back(insn);
}

void VDShaderParserFunction::insn_mod(int op, unsigned char mask, int dst, int src1, unsigned char src1swizzle, unsigned char src1mod, int src2, unsigned char src2swizzle, unsigned char src2mod, int src3, unsigned char src3swizzle, unsigned char src3mod) {
	VDShaderInsn insn;

	insn.mOp = op;
	insn.mWriteMask = mask;
	insn.mModifiers = 0;
	insn.mDestRegister = dst;
	insn.mSources[0].mRegister = src1;
	insn.mSources[0].mSwizzle = src1swizzle;
	insn.mSources[0].mModifiers = src1mod;
	insn.mSources[1].mRegister = src2;
	insn.mSources[1].mSwizzle = src2swizzle;
	insn.mSources[1].mModifiers = src2mod;
	insn.mSources[2].mRegister = src3;
	insn.mSources[2].mSwizzle = src3swizzle;
	insn.mSources[2].mModifiers = src3mod;
	insn.mSources[3].mRegister = 0;
	insn.mSources[3].mSwizzle = 0;
	insn.mSources[3].mModifiers = 0;

	VDASSERT((insn.mDestRegister & kVDShaderRegTypeMask) != kVDShaderRegR0 || (insn.mDestRegister - kVDShaderRegR0) < mRegisterCount);
	VDASSERT((insn.mSources[0].mRegister & kVDShaderRegTypeMask) != kVDShaderRegC0 || (insn.mSources[0].mRegister - kVDShaderRegC0) < (int)mConstants.size());
	VDASSERT((insn.mSources[1].mRegister & kVDShaderRegTypeMask) != kVDShaderRegC0 || (insn.mSources[1].mRegister - kVDShaderRegC0) < (int)mConstants.size());
	VDASSERT((insn.mSources[2].mRegister & kVDShaderRegTypeMask) != kVDShaderRegC0 || (insn.mSources[2].mRegister - kVDShaderRegC0) < (int)mConstants.size());
	VDASSERT(insn.mSources[0].mRegister || (!insn.mSources[1].mRegister && !insn.mSources[2].mRegister));

	mInsns.push_back(insn);
}

void VDShaderParserFunction::insn(int op, unsigned char mask, int dst, int src1, unsigned char src1swizzle, int src2, unsigned char src2swizzle) {
	VDShaderInsn insn;

	insn.mOp = op;
	insn.mWriteMask = mask;
	insn.mModifiers = 0;
	insn.mDestRegister = dst;
	insn.mSources[0].mRegister = src1;
	insn.mSources[0].mSwizzle = src1swizzle;
	insn.mSources[0].mModifiers = 0;
	insn.mSources[1].mRegister = src2;
	insn.mSources[1].mSwizzle = src2swizzle;
	insn.mSources[1].mModifiers = 0;
	insn.mSources[2].mRegister = 0;
	insn.mSources[2].mSwizzle = 0;
	insn.mSources[2].mModifiers = 0;
	insn.mSources[3].mRegister = 0;
	insn.mSources[3].mSwizzle = 0;
	insn.mSources[3].mModifiers = 0;

	VDASSERT((insn.mDestRegister & kVDShaderRegTypeMask) != kVDShaderRegR0 || (insn.mDestRegister - kVDShaderRegR0) < mRegisterCount);
	VDASSERT((insn.mSources[0].mRegister & kVDShaderRegTypeMask) != kVDShaderRegC0 || (insn.mSources[0].mRegister - kVDShaderRegC0) < (int)mConstants.size());
	VDASSERT((insn.mSources[1].mRegister & kVDShaderRegTypeMask) != kVDShaderRegC0 || (insn.mSources[1].mRegister - kVDShaderRegC0) < (int)mConstants.size());
	VDASSERT(insn.mSources[0].mRegister || !insn.mSources[1].mRegister);

	mInsns.push_back(insn);
}

void VDShaderParserFunction::insn(int op, unsigned char mask, int dst, int src1, unsigned char src1swizzle, int src2, unsigned char src2swizzle, int src3, unsigned char src3swizzle) {
	VDShaderInsn insn;

	insn.mOp = op;
	insn.mWriteMask = mask;
	insn.mModifiers = 0;
	insn.mDestRegister = dst;
	insn.mSources[0].mRegister = src1;
	insn.mSources[0].mSwizzle = src1swizzle;
	insn.mSources[0].mModifiers = 0;
	insn.mSources[1].mRegister = src2;
	insn.mSources[1].mSwizzle = src2swizzle;
	insn.mSources[1].mModifiers = 0;
	insn.mSources[2].mRegister = src3;
	insn.mSources[2].mSwizzle = src3swizzle;
	insn.mSources[2].mModifiers = 0;
	insn.mSources[3].mRegister = 0;
	insn.mSources[3].mSwizzle = 0;
	insn.mSources[3].mModifiers = 0;

	VDASSERT((insn.mDestRegister & kVDShaderRegTypeMask) != kVDShaderRegR0 || (insn.mDestRegister - kVDShaderRegR0) < mRegisterCount);
	VDASSERT((insn.mSources[0].mRegister & kVDShaderRegTypeMask) != kVDShaderRegC0 || (insn.mSources[0].mRegister - kVDShaderRegC0) < (int)mConstants.size());
	VDASSERT((insn.mSources[1].mRegister & kVDShaderRegTypeMask) != kVDShaderRegC0 || (insn.mSources[1].mRegister - kVDShaderRegC0) < (int)mConstants.size());
	VDASSERT(insn.mSources[0].mRegister || !insn.mSources[1].mRegister);

	mInsns.push_back(insn);
}

namespace {
	struct RegisterComponentSource {
		int mReg;
		int mComp;
		unsigned char mMods;
	};

	struct RegisterInfo {
		RegisterComponentSource mComponents[4];
		unsigned char mReferenceMask;
		unsigned char mWriteMask;

		RegisterInfo()
			: mReferenceMask(0)
			, mWriteMask(0)
		{
			for(int i=0; i<4; ++i) {
				mComponents[i].mReg = 0;
				mComponents[i].mComp = 0;
				mComponents[i].mMods = 0;
			}
		}
	};
}

bool VDShaderParserFunction::Inline(IVDShaderErrorOutput& lexer, const VDShaderParserFunction& calledFunc, const VDShaderParserVariable *arguments, VDShaderVariableDesc& returnValue) {
	int childInputBase = AllocateRegisters(calledFunc.mParameterRegisterCount);
	int childTempBase = AllocateRegisters(calledFunc.mRegisterCount);
	int outputBase = AllocateRegisters(kRowCountForType[calledFunc.mReturnType]);

	std::vector<int> samplerRemap(calledFunc.mSamplerCount, 0);

	int argCount = calledFunc.mParameters.size();
	int argReg = childInputBase;
	for(int arg=0; arg<argCount; ++arg) {
		const VDShaderParserVariable& argv = arguments[arg];
		int argType = argv.mType;

		if (argType == kTypeSampler) {
			int reg = calledFunc.mParameters[arg].mRegister;

			VDASSERT((reg & kVDShaderRegTypeMask) == kVDShaderRegS0);
			int i = reg - kVDShaderRegS0;

			VDASSERT(argv.mSamplerBinding.mNameToken);

			SamplerBindings::const_iterator it(std::find(mSamplerBindings.begin(), mSamplerBindings.end(), argv.mSamplerBinding));

			if (it != mSamplerBindings.end()) {
				samplerRemap[i] = kVDShaderRegS0 + (it - mSamplerBindings.begin());
			} else {
				samplerRemap[i] = kVDShaderRegS0 + mSamplerCount++;
				mSamplerBindings.push_back(argv.mSamplerBinding);
			}
		} else {
			int rows = kRowCountForType[argType];
			int mask = (1 << kColCountForType[argType]) - 1;

			if (argv.mbIsConst) {
				for(int row=0; row<rows; ++row)
					mov(argReg++, mask, AddConstant(argv.m[row]));
			} else {
				int srcReg = argv.mRegister;
				for(int row=0; row<rows; ++row)
					mov(argReg++, mask, srcReg);
			}
		}
	}

	// copy over all remaining samplers
	for(int i=0; i<calledFunc.mSamplerCount; ++i) {
		if (samplerRemap[i])
			continue;

		VDASSERT(mSamplerCount == mSamplerBindings.size());

		const VDShaderSamplerBinding& binding = calledFunc.mSamplerBindings[i];

		VDASSERT(binding.mNameToken);

		SamplerBindings::const_iterator it(std::find(mSamplerBindings.begin(), mSamplerBindings.end(), binding));

		if (it != mSamplerBindings.end()) {
			samplerRemap[i] = kVDShaderRegS0 + (it - mSamplerBindings.begin());
		} else {
			samplerRemap[i] = kVDShaderRegS0 + mSamplerCount++;
			mSamplerBindings.push_back(binding);
		}
	}

	// copy constants
	int constantCount = calledFunc.mConstants.size();
	int boundConstantCount = calledFunc.mConstantBindings.size();
	vdfastvector<int> constantRemap(constantCount, -1);

	for(int i=0; i<boundConstantCount; ++i) {
		const VDShaderConstantBinding& binding = calledFunc.mConstantBindings[i];

		int regBase = AddMappedConstant(binding, &calledFunc.mConstants[binding.mRegister & kVDShaderRegIndexMask]);
		for(int j=0; j<binding.mSize; ++j)
			constantRemap[binding.mRegister - kVDShaderRegC0 + j] = regBase + j;
	}

	for(int i=0; i<constantCount; ++i) {
		if (constantRemap[i] < 0)
			constantRemap[i] = AddConstant(calledFunc.mConstants[i]);
	}

	// inline function body
	for(Insns::const_iterator it(calledFunc.mInsns.begin()), itEnd(calledFunc.mInsns.end()); it!=itEnd; ++it) {
		VDShaderInsn insn = *it;

		// rename destination register
		switch(insn.mDestRegister & kVDShaderRegTypeMask) {
			case kVDShaderRegR0:
				insn.mDestRegister += childTempBase - kVDShaderRegR0;
				break;
			case kVDShaderRegO0:
				insn.mDestRegister = outputBase + (insn.mDestRegister & kVDShaderRegIndexMask);
				break;
		}

		// rename source 
		for(int i=0; i<4; ++i) {
			VDShaderInsnSource& src = insn.mSources[i];
			switch(src.mRegister & kVDShaderRegTypeMask) {
			case kVDShaderRegR0:
				src.mRegister += childTempBase - kVDShaderRegR0;
				VDASSERT(src.mRegister - kVDShaderRegR0 < mRegisterCount);
				break;
			case kVDShaderRegV0:
				src.mRegister += childInputBase - kVDShaderRegV0;
				break;
			case kVDShaderRegC0:
				src.mRegister = constantRemap[src.mRegister & kVDShaderRegIndexMask];
				break;
			case kVDShaderRegS0:
				src.mRegister = samplerRemap[src.mRegister & kVDShaderRegIndexMask];
				break;
			}
		}

		mInsns.push_back(insn);
	}

	returnValue.mNameToken = 0;
	returnValue.mSemanticToken = 0;
	returnValue.mRegister = outputBase;
	returnValue.mType = calledFunc.mReturnType;

	return true;
}

bool VDShaderParserFunction::Verify(IVDShaderErrorOutput& lexer) const {
	vdfastvector<uint8> tempCompsWritten(mRegisterCount, 0);

	for(Insns::const_iterator it(mInsns.begin()), itEnd(mInsns.end()); it!=itEnd; ++it) {
		const VDShaderInsn& insn = *it;

		unsigned srcmask = kInsnComponentMasks[insn.mOp];
		switch(insn.mOp) {
			case kVDShaderOp_dp2add:
			case kVDShaderOp_dp3:
			case kVDShaderOp_dp4:
			case kVDShaderOp_texld:
			case kVDShaderOp_frc:
			case kVDShaderOp_rcp:
			case kVDShaderOp_rsq:
			case kVDShaderOp_log:
			case kVDShaderOp_exp:
			case kVDShaderOp_sincos:
				break;
			default:
				{
					unsigned insnmask = insn.mWriteMask * 0x11111111;

					srcmask &= insnmask;
				}
				break;
		}

		for(int i=0; i<4; ++i) {
			const VDShaderInsnSource& src = insn.mSources[i];

			if (!src.mRegister)
				break;

			if ((src.mRegister & kVDShaderRegTypeMask) == kVDShaderRegR0) {
				// check for valid register
				int rindex = src.mRegister - kVDShaderRegR0;
				if (rindex >= mRegisterCount)
					return lexer.SetError("Shader validation failed: Invalid temp register as source");

				// check if components are written
				const uint8 rawCompsRead = (srcmask >> (4 * i)) & 15;
				uint8 compsRead = 0;

				for(int i=0; i<4; ++i) {
					if (!(rawCompsRead & (1 << i)))
						continue;

					int swizComp = (src.mSwizzle >> (i+i)) & 3;

					compsRead |= 1 << swizComp;
				}

				if (~tempCompsWritten[rindex] & compsRead)
					return lexer.SetError("Shader validation failed: Uninitialized register components read.");
			}
		}

		if ((insn.mDestRegister & kVDShaderRegTypeMask) == kVDShaderRegR0) {
			int dstIdx = insn.mDestRegister & kVDShaderRegIndexMask;

			tempCompsWritten[dstIdx] |= insn.mWriteMask;
		}
	}

	return true;
}

bool VDShaderParserFunction::Optimize(IVDShaderErrorOutput& lexer, const VDShaderNameTable& nameTable) {
	vdfastvector<RegisterInfo> regInfos(mRegisterCount);
	vdfastvector<int> tempRemap;
	vdfastvector<int> constantRemap;
	vdfastvector<int> boundConstantLookup;
	vdfastvector<VDShaderConstant> newConstants;
	vdfastvector<VDShaderConstantBinding> newConstantBindings;

	vdfastvector<VDShaderInsn> newinsns;

	bool eliminatedAnything;
	do {
		eliminatedAnything = false;

		// perform live register analysis
		for(vdfastvector<RegisterInfo>::iterator it(regInfos.begin()), itEnd(regInfos.end()); it!=itEnd; ++it) {
			RegisterInfo& regInfo = *it;

			regInfo.mReferenceMask = 0;
			regInfo.mWriteMask = 0;
			regInfo.mComponents[0].mComp = 0;
			regInfo.mComponents[0].mReg = 0;
			regInfo.mComponents[0].mMods = 0;
			regInfo.mComponents[1].mComp = 0;
			regInfo.mComponents[1].mReg = 0;
			regInfo.mComponents[1].mMods = 0;
			regInfo.mComponents[2].mComp = 0;
			regInfo.mComponents[2].mReg = 0;
			regInfo.mComponents[2].mMods = 0;
			regInfo.mComponents[3].mComp = 0;
			regInfo.mComponents[3].mReg = 0;
			regInfo.mComponents[3].mMods = 0;
		}

		// copy propagation / register packing
		constantRemap.clear();
		constantRemap.resize(mConstants.size(), -1);
		newConstants.clear();

		boundConstantLookup.clear();
		boundConstantLookup.resize(mConstants.size(), -1);

		for(ConstantBindings::const_iterator itB(mConstantBindings.begin()), itBEnd(mConstantBindings.end()); itB != itBEnd; ++itB) {
			const VDShaderConstantBinding& b = *itB;
			
			int offset = itB - mConstantBindings.begin();
			for(int i=0; i<b.mSize; ++i)
				boundConstantLookup[(b.mRegister - kVDShaderRegC0) + i] = offset;
		}

		newConstantBindings.clear();

		tempRemap.clear();
		tempRemap.resize(mRegisterCount, -1);
		int newRegisterCount = 0;
		int newBoundConstantRegisterCount = 0;

		for(Insns::iterator it(mInsns.begin()), itEnd(mInsns.end()); it!=itEnd; ++it) {
			VDShaderInsn& insn = *it;

			unsigned srcmask = kInsnComponentMasks[insn.mOp];

			switch(insn.mOp) {
				case kVDShaderOp_dp2add:
				case kVDShaderOp_dp3:
				case kVDShaderOp_dp4:
				case kVDShaderOp_texld:
				case kVDShaderOp_frc:
				case kVDShaderOp_rcp:
				case kVDShaderOp_rsq:
				case kVDShaderOp_log:
				case kVDShaderOp_exp:
				case kVDShaderOp_sincos:
					break;
				default:
					{
						unsigned insnmask = insn.mWriteMask * 0x11111111;

						srcmask &= insnmask;
					}
					break;
			}

			for(int i=0; i<4; ++i) {
				VDShaderInsnSource& src = insn.mSources[i];

				if (!src.mRegister)
					break;

				if ((src.mRegister & kVDShaderRegTypeMask) == kVDShaderRegC0) {
					int ci = src.mRegister & kVDShaderRegIndexMask;
					int nci = constantRemap[ci];

					if (nci < 0) {
						int bci = boundConstantLookup[ci];
						if (bci >= 0) {
							const VDShaderConstantBinding& binding = mConstantBindings[bci];
							VDShaderConstantBinding newBinding(binding);

							nci = newConstants.size();
							newBinding.mRegister = kVDShaderRegC0 + nci;

							newConstantBindings.push_back(newBinding);
							newConstants.insert(newConstants.end(), mConstants.begin() + ci, mConstants.begin() + ci + binding.mSize);

							for(int j=0; j<binding.mSize; ++j) {
								VDASSERT(constantRemap[ci + j] < 0);

								constantRemap[ci + j] = nci + j;
							}
						} else {
							nci = newConstants.size();
							constantRemap[ci] = nci;
							newConstants.push_back(mConstants[ci]);
						}
					}

					src.mRegister = kVDShaderRegC0 + nci;
				} else if ((src.mRegister & kVDShaderRegTypeMask) == kVDShaderRegR0) {
					// check for valid register
					int rindex = src.mRegister - kVDShaderRegR0;
					if (rindex >= mRegisterCount)
						return lexer.SetError("Internal compiler error: Invalid temp register as source");

					// remap temp register
					rindex = tempRemap[rindex];
					if (rindex < 0) {
						// We may get a mov instruction that references some uninitialized components
						// for a write merge.
						if (insn.mOp == kVDShaderOp_mov) {
							insn.mOp = kVDShaderOp_nop;
							goto nop_insn;
						} else {
							lexer.SetError("Internal compiler error: Uninitialized temp register used");
							VDShaderDisassembleFunction(*lexer.GetLogOutput(), *this, &nameTable);
							return false;
						}
					}
					src.mRegister = kVDShaderRegR0 + rindex;

					// attempt to remap via copy propagation
					do {
						const RegisterInfo& srcRegInfo = regInfos[src.mRegister & kVDShaderRegIndexMask];
						int upstreamReg = 0;
						unsigned char swizzle = 0;
						unsigned char modifiers = 0;

						for(int j=0; j<4; ++j) {
							swizzle >>= 2;

							if (srcmask & (1 << (i*4 + j))) {
								int refComp = (src.mSwizzle >> (j+j)) & 3;
								int refReg = srcRegInfo.mComponents[refComp].mReg;
								unsigned char refMods = srcRegInfo.mComponents[refComp].mMods;

								if (!refReg) {
									upstreamReg = 0;
									break;
								}

								if (!upstreamReg)
									upstreamReg = refReg;
								else if (refReg != upstreamReg || refMods != modifiers) {
									upstreamReg = 0;
									break;
								}

								swizzle += srcRegInfo.mComponents[refComp].mComp << 6;
								modifiers = srcRegInfo.mComponents[refComp].mMods;
							} else {
								swizzle += j << 6;
							}
						}

						if (!upstreamReg)
							break;

						src.mRegister = upstreamReg;
						src.mSwizzle = swizzle;
						src.mModifiers ^= modifiers;
					} while((src.mRegister & kVDShaderRegTypeMask) == kVDShaderRegR0);

					if ((src.mRegister & kVDShaderRegTypeMask) == kVDShaderRegR0) {
						RegisterInfo& regInfo = regInfos[src.mRegister & kVDShaderRegIndexMask];
						for(int j=0; j<4; ++j) {
							if (srcmask & (1 << (i*4 + j))) {
								regInfo.mReferenceMask |= 1 << ((src.mSwizzle >> (2*j)) & 3);
							}
						}
					}
				}
			}

			if (insn.mDestRegister && (insn.mDestRegister & kVDShaderRegTypeMask) == kVDShaderRegR0) {
				int rindex = insn.mDestRegister - kVDShaderRegR0;

				// remap temp register
				if (rindex >= mRegisterCount)
					return lexer.SetError("Internal compiler error: Invalid dest temp register index");

				int newrindex = tempRemap[rindex];
				if (newrindex < 0) {
					newrindex = newRegisterCount++;
					tempRemap[rindex] = newrindex;
				}

				rindex = newrindex;
				insn.mDestRegister = kVDShaderRegR0 + newrindex;

				RegisterInfo& dstRegInfo = regInfos[rindex];

				dstRegInfo.mWriteMask |= insn.mWriteMask;

				if (insn.mOp == kVDShaderOp_mov && !insn.mModifiers) {
					const VDShaderInsnSource& src1 = insn.mSources[0];

					for(int i=0; i<4; ++i) {
						if (insn.mWriteMask & (1 << i)) {
							dstRegInfo.mComponents[i].mReg = src1.mRegister;
							dstRegInfo.mComponents[i].mComp = (src1.mSwizzle >> (i+i)) & 3;
							dstRegInfo.mComponents[i].mMods = src1.mModifiers;
						}
					}
				} else {
					for(int i=0; i<4; ++i) {
						if (insn.mWriteMask & (1 << i))
							dstRegInfo.mComponents[i].mReg = 0;
					}
				}
			}
nop_insn:
			;
		}

		mConstants.swap(newConstants);
		mConstantBindings.swap(newConstantBindings);

		// dead store elimination
		newinsns.clear();
		for(Insns::const_iterator it(mInsns.begin()), itEnd(mInsns.end()); it!=itEnd; ++it) {
			VDShaderInsn insn = *it;

			if (insn.mOp == kVDShaderOp_nop) {
				eliminatedAnything = true;
				continue;
			}

			if (insn.mDestRegister && (insn.mDestRegister & kVDShaderRegTypeMask) == kVDShaderRegR0) {
				RegisterInfo& dstRegInfo = regInfos[insn.mDestRegister - kVDShaderRegR0];

				if (insn.mWriteMask & ~dstRegInfo.mReferenceMask) {
					insn.mWriteMask &= dstRegInfo.mReferenceMask;
					eliminatedAnything = true;
				}
			}

			if (insn.mWriteMask)
				newinsns.push_back(insn);
			else
				eliminatedAnything = true;
		}

		mInsns.swap(newinsns);
	} while(eliminatedAnything);


	return true;
}

///////////////////////////////////////////////////////////////////////////////
VDShaderParser::VDShaderParser() {
}

VDShaderParser::~VDShaderParser() {
	while(!mFunctions.empty()) {
		delete mFunctions.back();
		mFunctions.pop_back();
	}
}

namespace {
	class NameRemapper {
	public:
		NameRemapper(VDShaderNameTable& dstNameTable, const VDShaderNameTable& srcNameTable) : mDstNameTable(dstNameTable), mSrcNameTable(srcNameTable) {}

		int Remap(int srcNameToken) {
			if (!srcNameToken)
				return 0;

			NameRemap::iterator it(mNameRemap.find(srcNameToken));

			if (it != mNameRemap.end())
				return it->second;

			const char *s = mSrcNameTable.GetName(srcNameToken);
			int dstNameToken = mDstNameTable.AddName(s, strlen(s));

			mNameRemap.insert(NameRemap::value_type(srcNameToken, dstNameToken));

			return dstNameToken;
		}

		void RemapAnnotations(VDShaderVariable *anno, size_t count, char *data);

	protected:
		VDShaderNameTable& mDstNameTable;
		const VDShaderNameTable& mSrcNameTable;

		typedef std::map<int, int> NameRemap;
		NameRemap mNameRemap;
	};

	void NameRemapper::RemapAnnotations(VDShaderVariable *anno, size_t count, char *data) {
		while(count--) {
			anno->mDesc.mNameToken = Remap(anno->mDesc.mNameToken);

			if (anno->mDesc.mType == kTypeString && anno->mValueToken) {
				char *stringTokenData = data + anno->mValueToken - 1;

				int valueToken;
				memcpy(&valueToken, stringTokenData, sizeof(int));
				valueToken = Remap(valueToken);
				memcpy(stringTokenData, &valueToken, sizeof(int));
			}

			++anno;
		}
	}

	VDShaderAnnotations CopyAnnotations(vdfastvector<VDShaderVariable>& dst, vdfastvector<char>& dstData, const VDShaderVariable *srcTable, const char *srcData, const VDShaderAnnotations& srcAnno) {
		VDShaderAnnotations dstAnno;

		dstAnno.mStart = dst.size() + 1;
		dstAnno.mCount = srcAnno.mCount;

		for(int i=0; i<dstAnno.mCount; ++i) {
			VDShaderVariable var = srcTable[srcAnno.mStart - 1 + i];

			if (var.mValueToken) {
				const char *srcVal = srcData + (var.mValueToken - 1);
				var.mValueToken = dstData.size() + 1;
				dstData.insert(dstData.end(), srcVal, srcVal + var.mSize);
			}

			dst.push_back(var);
		}

		return dstAnno;
	}
}

bool VDShaderParser::Parse(const char *src, size_t len, const char *context, bool optimizationEnabled, IVDShaderLogOutput *pOutput, VDShaderEffect **ppEffect) {
	mpEffect = new VDShaderEffect;
	mbOptimizationEnabled = optimizationEnabled;

	mNextGlobalConstantRegister = 0;
	mCurrentScopeLevel = 0;
	mCurrentScopeLimit = 0;
	mpFunction = NULL;

	static const float kBaseConstants[]={
		0, 0, 0, 0
	};

	mConstants.assign(kBaseConstants, kBaseConstants + sizeof(kBaseConstants)/sizeof(kBaseConstants[0]));

	mLexer.Init(src, len, context, pOutput);

	while(TryParseGlobalStatement())
		;

	if (mpEffect->mTechniques.empty())
		InsertPlaceholderTechnique();

	if (mLexer.IsErrorFlagSet())
		return NULL;

	NameRemapper nameRemapper(mpEffect->mNameTable, mNameTable);

	for(Functions::iterator it(mFunctions.begin()), itEnd(mFunctions.end()); it!=itEnd; ++it) {
		VDShaderParserFunction *pParserFunc = *it;

		VDShaderFunction *func = new VDShaderFunction;

		func->mConstants.swap(pParserFunc->mConstants);
		func->mInsns.swap(pParserFunc->mInsns);
		func->mParameterRegisterCount = pParserFunc->mParameterRegisterCount;
		func->mParameters.swap(pParserFunc->mParameters);
		func->mRegisterCount = pParserFunc->mRegisterCount;
		func->mReturnType = pParserFunc->mReturnType;

		for(VDShaderFunction::Parameters::iterator itP(func->mParameters.begin()), itPEnd(func->mParameters.end()); itP!=itPEnd; ++itP) {
			itP->mNameToken = nameRemapper.Remap(itP->mNameToken);
			itP->mSemanticToken = nameRemapper.Remap(itP->mSemanticToken);
		}

		func->mNameToken = nameRemapper.Remap(pParserFunc->mNameToken);

		func->mSamplerBindings.swap(pParserFunc->mSamplerBindings);
		for(VDShaderFunction::SamplerBindings::iterator itS(func->mSamplerBindings.begin()), itSEnd(func->mSamplerBindings.end()); itS!=itSEnd; ++itS) {
			itS->mNameToken = nameRemapper.Remap(itS->mNameToken);
		}

		func->mSamplerCount = pParserFunc->mSamplerCount;

		func->mConstantBindings.swap(pParserFunc->mConstantBindings);
		for(VDShaderFunction::ConstantBindings::iterator itC(func->mConstantBindings.begin()), itCEnd(func->mConstantBindings.end()); itC!=itCEnd; ++itC) {
			VDShaderConstantBinding& binding = *itC;
			binding.mNameToken = nameRemapper.Remap(binding.mNameToken);
		}

		mpEffect->mFunctions.push_back(func);

		delete pParserFunc;
	}

	mFunctions.clear();
	while(!mFunctions.empty()) {
		delete mFunctions.back();
		mFunctions.pop_back();
	}

	const VDShaderVariable *pAnnotations = mAnnotations.data();
	const char *pAnnoData = mDataStorage.data();

	for(VDShaderEffect::Techniques::iterator it(mpEffect->mTechniques.begin()), itEnd(mpEffect->mTechniques.end()); it!=itEnd; ++it) {
		VDShaderTechnique& tech = **it;

		tech.mNameToken = nameRemapper.Remap(tech.mNameToken);
		tech.mAnnotations = CopyAnnotations(mpEffect->mAnnotations, mpEffect->mDataStorage, pAnnotations, pAnnoData, tech.mAnnotations);

		for(VDShaderTechnique::Passes::iterator itP(tech.mPasses.begin()), itPEnd(tech.mPasses.end()); itP!=itPEnd; ++itP) {
			VDShaderPass& pass = **itP;

			pass.mNameToken = nameRemapper.Remap(pass.mNameToken);
			pass.mAnnotations = CopyAnnotations(mpEffect->mAnnotations, mpEffect->mDataStorage, pAnnotations, pAnnoData, pass.mAnnotations);
		}
	}

	int numVars = mVarHeap.size();
	mpEffect->mVariables.resize(numVars);
	for(int i=0; i<numVars; ++i) {
		const VDShaderParserVariable& varSrc = mVarHeap[i];
		VDShaderVariable& var = mpEffect->mVariables[i];
		var.mDesc = varSrc;
		var.mDesc.mNameToken = nameRemapper.Remap(var.mDesc.mNameToken);
		var.mDesc.mSemanticToken = nameRemapper.Remap(var.mDesc.mSemanticToken);
		var.mDesc.mAnnotations = CopyAnnotations(mpEffect->mAnnotations, mpEffect->mDataStorage, pAnnotations, pAnnoData, var.mDesc.mAnnotations);

		var.mValueToken = mpEffect->mDataStorage.size() + 1;

		const char *src = (const char *)varSrc.m[0].v;

		switch(var.mDesc.mType) {
			case kTypeBool:
				mpEffect->mDataStorage.push_back(varSrc.m[0].v[0] != 0);
				var.mSize = 1;
				break;
			case kTypeFloat:
				mpEffect->mDataStorage.insert(mpEffect->mDataStorage.end(), src, src + sizeof(float));
				var.mSize = sizeof(float);
				break;
			case kTypeFloat2:
				mpEffect->mDataStorage.insert(mpEffect->mDataStorage.end(), src, src + sizeof(float)*2);
				var.mSize = sizeof(float)*2;
				break;
			case kTypeFloat3:
				mpEffect->mDataStorage.insert(mpEffect->mDataStorage.end(), src, src + sizeof(float)*3);
				var.mSize = sizeof(float)*3;
				break;
			case kTypeFloat4:
				mpEffect->mDataStorage.insert(mpEffect->mDataStorage.end(), src, src + sizeof(float)*4);
				var.mSize = sizeof(float)*4;
				break;
			case kTypeString:
				{
					int valueToken;
					memcpy(&valueToken, src, sizeof(int));
					valueToken = nameRemapper.Remap(valueToken);

					src = (const char *)&valueToken;
					mpEffect->mDataStorage.insert(mpEffect->mDataStorage.end(), src, src + sizeof(int));
				}
				var.mSize = sizeof(int);
				break;
			default:
				var.mSize = 0;
				var.mValueToken = 0;
				break;
		}
	}

	nameRemapper.RemapAnnotations(mpEffect->mAnnotations.data(), mpEffect->mAnnotations.size(), mpEffect->mDataStorage.data());

	for(VDShaderEffect::SamplerStates::iterator it(mpEffect->mSamplerStates.begin()), itEnd(mpEffect->mSamplerStates.end()); it!=itEnd; ++it) {
		VDShaderSamplerState& state = *it;

		state.mTextureToken = nameRemapper.Remap(state.mTextureToken);
	}

	*ppEffect = mpEffect.release();
	return true;
}

bool VDShaderParser::GetErrorInfo(VDShaderErrorInfo& errorInfo) {
	return mLexer.GetErrorInfo(errorInfo);
}

void VDShaderParser::InsertPlaceholderTechnique() {
	// try to find a function called 'main'
	int hMain = mNameTable.GetName("main", 4);

	Functions::const_iterator it(mFunctions.begin()), itEnd(mFunctions.end());
	VDShaderParserFunction *pFunc = NULL;
	for(; it != itEnd; ++it) {
		VDShaderParserFunction *func = *it;

		if (func->mNameToken == hMain) {
			pFunc = func;
			break;
		}
	}

	if (!pFunc) {
		mLexer.SetError("No techniques were declared and there is no main() function.");
		return;
	}

	// validate as pixel shader
	if (!ValidateFunctionAsPixelShader(*pFunc))
		return;

	// create technique
	vdautoptr<VDShaderTechnique> pTechnique(new VDShaderTechnique);
	pTechnique->mNameToken = 0;

	// create pass
	vdautoptr<VDShaderPass> pPass(new VDShaderPass);
	pPass->mNameToken = 0;
	pPass->mPixelShaderIndex = pFunc->mIndex;

	pTechnique->mPasses.push_back(pPass);
	pPass.release();

	mpEffect->mTechniques.push_back(pTechnique);
	pTechnique.release();
}

bool VDShaderParser::TryParseGlobalStatement() {
	int tok = mLexer.Token();

	if (!tok)
		return false;

	if (tok == kTokenIdent) {
		bool isExtern = false;
		bool isStatic = false;

		for(;;) {
			if (mLexer.mTokenLength == 6 && !memcmp(mLexer.mpToken, "extern", 6)) {
				tok = mLexer.Token();
				if (tok != kTokenIdent)
					return mLexer.SetError("Expected declaration after 'extern' keyword");

				isExtern = true;
			} else if (mLexer.mTokenLength == 6 && !memcmp(mLexer.mpToken, "static", 6)) {
				tok = mLexer.Token();
				if (tok != kTokenIdent)
					return mLexer.SetError("Expected declaration after 'extern' keyword");

				isStatic = true;
			} else
				break;
		}

		int type = GetTypeFromIdent();

		if (type) {
			tok = mLexer.Token();
			int name = GetNameToken();

			tok = mLexer.Token();
			if (tok == '(') {
				if (!ParseFunction(type, name))
					return false;
				return true;
			} else {
				mLexer.Push();
				if (!ParseNamedDeclarationList(type, name, isExtern, isStatic, true))
					return false;
				return true;
			}
		}

		if (mLexer.mTokenLength == 9 && !memcmp(mLexer.mpToken, "technique", 9))
			return ParseTechnique();
	}

	return mLexer.SetError("Expected declaration");
}

bool VDShaderParser::ParseTechnique() {
	vdautoptr<VDShaderTechnique> pTechnique(new VDShaderTechnique);

	int tok = mLexer.Token();

	pTechnique->mNameToken = 0;
	if (tok == kTokenIdent) {
		int name = GetNameToken();

		pTechnique->mNameToken = name;
		tok = mLexer.Token();
	}

	pTechnique->mAnnotations.mStart = 0;
	pTechnique->mAnnotations.mCount = 0;

	if (tok == '<') {
		if (!ParseAnnotations(pTechnique->mAnnotations))
			return false;

		tok = mLexer.Token();
	}

	if (tok != '{')
		return mLexer.SetError("Expected '{' after technique keyword");

	for(;;) {
		tok = mLexer.Token();

		if (tok == '}')
			break;

		if (tok != kTokenIdent || mLexer.mTokenLength != 4 || memcmp(mLexer.mpToken, "pass", 4))
			return mLexer.SetError("Expected pass or end of technique");

		if (!ParsePass(*pTechnique))
			return false;
	}

	mpEffect->mTechniques.push_back(pTechnique);
	pTechnique.release();

	return true;
}

bool VDShaderParser::ParsePass(VDShaderTechnique& tech) {
	vdautoptr<VDShaderPass> pPass(new VDShaderPass);

	int tok = mLexer.Token();

	pPass->mNameToken = 0;
	if (tok == kTokenIdent) {
		int name = GetNameToken();

		pPass->mNameToken = name;
		tok = mLexer.Token();
	}

	pPass->mAnnotations.mCount = 0;
	pPass->mAnnotations.mStart = 0;
	if (tok == '<') {
		if (!ParseAnnotations(pPass->mAnnotations))
			return false;

		tok = mLexer.Token();
	}

	if (tok != '{')
		return mLexer.SetError("Expected '{' after pass keyword");

	pPass->mPixelShaderIndex = -1;
	for(;;) {
		tok = mLexer.Token();

		if (tok == '}')
			break;

		if (tok == kTokenIdent) {
			if (mLexer.mTokenLength == 11 && !_memicmp(mLexer.mpToken, "PixelShader", 11)) {
				tok = mLexer.Token();

				if (tok != '=')
					return mLexer.SetError("Expected '=' after PixelShader keyword");

				if (!ParseExpression())
					return false;

				VDShaderParserVariable& v = mVarHeap.back();
				PromoteLValueToRValue(v);
				if (v.mType != kTypePixelShader)
					return mLexer.SetErrorF("Expression must be of type 'pixelshader', not '%s'", GetTypeName(v.mType));

				if (pPass->mPixelShaderIndex >= 0)
					return mLexer.SetError("Pixel shader already set for this pass");

				pPass->mPixelShaderIndex = v.mShader;

				mVarHeap.pop_back();

				tok = mLexer.Token();
				if (tok != ';')
					return mLexer.SetError("Expected ';' at end of pass statement");

				VDShaderParserFunction& func = *mFunctions[pPass->mPixelShaderIndex];
				if (!ValidateFunctionAsPixelShader(func))
					return false;

				continue;
			}
		}

		return mLexer.SetError("Expected pass statement or end of technique");
	}

	if (pPass->mPixelShaderIndex < 0)
		return mLexer.SetError("No pixel shader declared in pass");

	tech.mPasses.push_back(pPass);
	pPass.release();

	return true;
}

bool VDShaderParser::ParseFunction(int type, int name) {
	mpFunction = new VDShaderParserFunction;
	mpFunction->mNameToken = name;
	mpFunction->mReturnType = type;
	mpFunction->mIndex = mFunctions.size();
	mFunctions.push_back(mpFunction);

	int oldScopeLevel = mCurrentScopeLevel;
	mCurrentScopeLevel = (int)mVarHeap.size();

	// parse argument list
	int tok = mLexer.Token();
	if (tok != ')') {
		mLexer.Push();

		for(;;) {
			VDShaderParserVariable var;

			tok = mLexer.Token();
			if (tok != kTokenIdent)
				return mLexer.SetError("Expected parameter type");

			if (mLexer.mTokenLength == 7 && !memcmp(mLexer.mpToken, "uniform", 7)) {
				var.mbUniform = true;
				tok = mLexer.Token();
				if (tok != kTokenIdent)
					return mLexer.SetError("Expected parameter type");
			}

			int type = GetTypeFromIdent();
			if (!type)
				return mLexer.SetError("Expected parameter type");

			tok = mLexer.Token();
			if (tok != kTokenIdent)
				return mLexer.SetError("Expected parameter name");

			int paramName = GetNameToken();
			int paramSemantic = 0;

			tok = mLexer.Token();
			if (tok == ':') {
				tok = mLexer.Token();
				if (tok != kTokenIdent)
					return mLexer.SetError("Parameter semantic name must follow ':'");
				paramSemantic = GetNameToken();
				tok = mLexer.Token();
			}

			var.mNameToken = paramName;
			var.mSemanticToken = paramSemantic;
			var.mType = type;
			var.mbUnassigned = false;

			if (type == kTypeSampler) {
				var.mbIsConst = false;
				var.mRegister = kVDShaderRegS0 + mpFunction->mSamplerCount++;
				var.mSamplerBinding.mNameToken = paramName;
				var.mSamplerBinding.mStateToken = 0;

				VDShaderSamplerBinding binding = { paramName, 0 };
				mpFunction->mSamplerBindings.push_back(binding);
			} else {
				var.mRegister = mpFunction->AllocateParamRegisters(kRowCountForType[type]);
				var.mbIsConst = false;
			}

			mVarHeap.push_back(var);

			mpFunction->mParameters.push_back(var);

			if (tok == ')')
				break;
			else if (tok != ',')
				return mLexer.SetError("Expected ',' or ')'");
		}
	}

	mCurrentScopeLimit = (int)mVarHeap.size();

	mpFunction->mReturnSemantic = 0;

	tok = mLexer.Token();
	if (tok == ':') {
		tok = mLexer.Token();
		if (tok != kTokenIdent)
			return mLexer.SetError("Expected return semantic type after ':'");
		mpFunction->mReturnSemantic = GetNameToken();
		tok = mLexer.Token();
	}

	if (tok != '{')
		return mLexer.SetError("Expected function body");

	bool returnFound = false;
	if (!ParseBlock(returnFound))
		return false;

	if (!returnFound && mpFunction->mReturnType != kTypeVoid)
		return mLexer.SetError("Missing return statement");

	mVarHeap.resize(mCurrentScopeLevel);
	mCurrentScopeLevel = oldScopeLevel;
	mCurrentScopeLimit = mCurrentScopeLevel;

	if (mbOptimizationEnabled && !mpFunction->Optimize(mLexer, mNameTable))
		return false;

	if (!mpFunction->BakeHardcodedRegisters(mLexer))
		return false;

	mpFunction = NULL;

	return true;
}

bool VDShaderParser::ParseDeclarationList(int type, bool isExtern, bool isStatic, bool keepAnnotations) {
	int tok = mLexer.Token();

	if (tok != kTokenIdent)
		return mLexer.SetError("Expected type name");

	int name = GetNameToken();

	return ParseNamedDeclarationList(type, name, isExtern, isStatic, keepAnnotations);
}

bool VDShaderParser::ParseNamedDeclarationList(int type, int name, bool isExtern, bool isStatic, bool keepAnnotations) {
	int tok;

	if (type == kTypeVoid)
		return mLexer.SetError("Variables cannot be of type 'void'");

	for(;;) {
		int semantic = 0;
		int reg = 0;
		tok = mLexer.Token();

		for(;;) {
			if (tok != ':')
				break;

			tok = mLexer.Token();
			if (tok != kTokenIdent)
				return mLexer.SetError("Variable semantic type must follow ':'");

			if (mLexer.mTokenLength == 8 && !memcmp(mLexer.mpToken, "register", 8)) {
				tok = mLexer.Token();
				if (tok != '(')
					return mLexer.SetError("Expected '(' after register keyword");

				tok = mLexer.Token();
				if (tok != kTokenIdent)
					return mLexer.SetError("Expected constant or sampler register");

				const char *s = mLexer.mpToken;
				const char *t = s + mLexer.mTokenLength;
				bool regvalid = true;

				if (*s == 'c')
					reg = kVDShaderRegC0;
				else if (*s == 's')
					reg = kVDShaderRegS0;
				else
					regvalid = false;
				
				int regidx = 0;
				for(++s; s != t; ++s) {
					unsigned digit = (unsigned)*s - '0';

					if (digit >= 10) {
						regvalid = false;
						break;
					}

					if (regidx >= 4096) {
						regvalid = false;
						break;
					}

					regidx = (regidx * 10) + digit;
				}

				if (!regvalid || regidx >= 4096)
					return mLexer.SetErrorF("Invalid register '%.*s'", mLexer.mTokenLength, mLexer.mpToken);

				reg += regidx;

				tok = mLexer.Token();
				if (tok != ')')
					return mLexer.SetError("Expected ')' after register");
			} else {
				if (semantic)
					return mLexer.SetError("Only one semantic allowed");

				semantic = GetNameToken();
			}

			tok = mLexer.Token();
		}

		int index = (int)mVarHeap.size();

		VDShaderParserVariable var;
		var.mNameToken = name;
		var.mSemanticToken = semantic;
		var.mType = type;
		var.mbIsConst = false;
		var.mbUnassigned = false;
		var.mRegister = reg;
		var.mbExtern = isExtern;
		var.mbStatic = isStatic;
		memset(var.m, 0, sizeof var.m);

		if (mpFunction) {
			// function scope
			if (type == kTypeSampler)
				return mLexer.SetError("Sampler variables may not be declared at function scope");

			var.mRegister = mpFunction->AllocateRegisters(kRowCountForType[type]);
		} else {
			var.mbUnassigned = true;

			// global scope
			if (type == kTypeSampler) {
				var.mSamplerBinding.mNameToken = name;
				var.mSamplerBinding.mStateToken = 0;
			}
		}

		if (type == kTypeTexture)
			var.mTextureBinding = name;

		if (tok == '<') {
			if (mpFunction)
				return mLexer.SetError("Local variables cannot have annotations");

			if (!ParseAnnotations(var.mAnnotations))
				return false;

			if (!keepAnnotations) {
				var.mAnnotations.mCount = 0;
				var.mAnnotations.mStart = 0;
			}

			tok = mLexer.Token();
		}

		mVarHeap.push_back(var);
		mCurrentScopeLimit = mVarHeap.size();

		if (tok == '=') {
			tok = mLexer.Token();

			if (tok == '{') {
				if (!ParseInitializerList(type, false))
					return false;

				tok = mLexer.Token();

				if (tok != '}')
					return mLexer.SetError("Expected '}' after initializer list");
			} else {
				mLexer.Push();

				if (!ParseAssignmentExpression())
					return false;
			}

			VDShaderParserVariable& var2 = mVarHeap.back();
			PromoteLValueToRValue(var2);

			int varType = mVarHeap[index].mType;
			if (!ConvertToType(var2, varType))
				return mLexer.SetErrorF("Cannot convert initializer from type '%s' to '%s'", GetTypeName(var2.mType), GetTypeName(varType));

			if (varType == kTypeSampler) {
				if (!var2.mSamplerBinding.mNameToken)
					var2.mSamplerBinding.mNameToken = name;
			}

			var2.mNameToken = name;
			var2.mSemanticToken = semantic;
			var2.mAnnotations = var.mAnnotations;
			var2.mbExtern = isExtern;
			var2.mbStatic = isStatic;

			if (!mpFunction && !isStatic) {
				var2.mbIsConst = false;
				var2.mbUnassigned = true;
			}

			mVarHeap[index] = var2;
			mVarHeap.pop_back();

			tok = mLexer.Token();
		} else {
			// make uninitialized globals extern by default
			mVarHeap.back().mbExtern = true;

			if (isStatic)
				return mLexer.SetError("Static constants must be initialized");
		}

		if (tok == ';')
			return true;

		if (tok != ',') {
			mLexer.SetError("Expected ',' or ';'");
			return false;
		}
	}
}

bool VDShaderParser::ParseBlock(bool& returnFound) {
	for(;;) {
		int tok = mLexer.Token();

		if (tok == '}')
			break;

		mLexer.Push();
		if (!ParseStatement(returnFound))
			return false;
	}

	return true;
}

bool VDShaderParser::ParseStatement(bool& returnFound) {
	int tok = mLexer.Token();

	if (tok == '{') {
		int oldScopeLevel = mCurrentScopeLevel;
		mCurrentScopeLevel = mVarHeap.size();
		mCurrentScopeLimit = mCurrentScopeLevel;

		bool result = ParseBlock(returnFound);

		mVarHeap.resize(mCurrentScopeLevel);
		mCurrentScopeLimit = mCurrentScopeLevel;
		mCurrentScopeLevel = oldScopeLevel;

		return result;
	} else if (tok == kTokenIdent) {
		int type = GetTypeFromIdent();

		if (type)
			return ParseDeclarationList(type, false, false, false);

		if (mLexer.mTokenLength == 2 && !memcmp(mLexer.mpToken, "if", 2))
			return ParseIfStatement();

		if (mLexer.mTokenLength == 6 && !memcmp(mLexer.mpToken, "return", 6)) {
			returnFound = true;
			return ParseReturnStatement();
		}

		mLexer.Push();

		if (!ParseExpression())
			return false;

		mVarHeap.pop_back();

		tok = mLexer.Token();
		if (tok != ';')
			return mLexer.SetError("Expected end of statement");

		return true;
	}

	return mLexer.SetError("Expected statement");
}

bool VDShaderParser::ParseIfStatement() {
	int tok = mLexer.Token();

	if (tok != '(')
		return mLexer.SetError("Expected '(' after if statement");

	if (!ParseExpression())
		return false;

	VDShaderParserVariable varCond = mVarHeap.back();
	mVarHeap.pop_back();

	PromoteLValueToRValue(varCond);

	if (!ConvertToType(varCond, kTypeBool))
		return false;

	tok = mLexer.Token();
	if (tok != ')')
		return mLexer.SetError("Expected ')' after if statement condition");

	VarHeap baseVars(mVarHeap);
	VarHeap falseVars(mVarHeap);

	bool returnFound = false;
	if (!ParseStatement(returnFound))
		return false;

	VarHeap trueVars(mVarHeap);

	if (returnFound)
		return mLexer.SetError("Return statements in conditional statements are not yet supported.");

	// check for else

	mVarHeap = baseVars;

	tok = mLexer.Token();
	if (tok == kTokenIdent && mLexer.mTokenLength == 4 && !memcmp(mLexer.mpToken, "else", 4)) {
		if (!ParseStatement(returnFound))
			return false;

		if (returnFound)
			return mLexer.SetError("Return statements in conditional statements are not yet supported.");

		falseVars = mVarHeap;

		mVarHeap = baseVars;
	} else {
		mLexer.Push();
	}

	if (varCond.mbIsConst) {
		if (varCond.m[0].v[0] != 0)
			mVarHeap = trueVars;
		else
			mVarHeap = falseVars;

		return true;
	}

	// diff and merge
	//
	//	true		false		action
	//	same		same		no action
	//	modified	same		v = cmp(cond, v_true, v_base)
	//	same		modified	v = cmp(cond, v_base, v_false)
	//	modified	modified	v = cmp(cond, v_true, v_false)

	int n = baseVars.size();
	for(int i=0; i<n; ++i) {
		VDShaderParserVariable& v_true = trueVars[i];
		VDShaderParserVariable& v_false = falseVars[i];

		if (v_true.mbIsConst == v_false.mbIsConst) {
			if (v_true.mbIsConst) {
				bool diff = false;

				for(int i=0; i<4; ++i) {
					if (v_true.m[i] != v_false.m[i]) {
						diff = true;
						break;
					}
				}

				if (!diff)
					continue;
			} else {
				if (v_true.mRegister == v_false.mRegister)
					continue;
			}
		}

		mpFunction->PromoteToConstantRegisters(v_true);
		mpFunction->PromoteToConstantRegisters(v_false);

		int t = v_true.mType;
		int rows = kRowCountForType[t];
		if (rows == 0)
			return mLexer.SetErrorF("Cannot support conditionals on type '%s'", GetTypeName(t));

		int r0 = mpFunction->AllocateRegisters(rows);
		int mask = kColWriteMaskForType[t];

		for(int j=0; j<rows; ++j)
			mpFunction->cmp_swizzle_mods(r0 + j, mask, varCond.mRegister, 0, kVDShaderMod_Negate, v_false.mRegister + j, 0xE4, 0, v_true.mRegister + j, 0xE4, 0);

		VDShaderParserVariable& v_result = mVarHeap[i];
		v_result.mRegister = r0;
		v_result.mbIsConst = false;
		v_result.mbUnassigned = false;
	}

	return true;
}

bool VDShaderParser::ParseReturnStatement() {
	if (!ParseExpression())
		return false;

	VDShaderParserVariable& var = mVarHeap.back();

	PromoteLValueToRValue(var);

	if (!ConvertToType(var, mpFunction->mReturnType))
		return mLexer.SetErrorF("Cannot convert return value from type '%s' to type '%s'", GetTypeName(var.mType), GetTypeName(mpFunction->mReturnType));

	mpFunction->PromoteToConstantRegisters(var);

	int rows = kRowCountForType[mpFunction->mReturnType];
	for(int i=0; i<rows; ++i) {
		mpFunction->mov(kVDShaderRegO0 + i, 0x0F, var.mRegister + i);
	}
	mVarHeap.pop_back();

	int tok = mLexer.Token();
	if (tok != ';')
		return mLexer.SetError("Expected end of return statement");

	return true;
}

bool VDShaderParser::ParseAnnotations(VDShaderAnnotations& annos) {
	annos.mStart = mAnnotations.size() + 1;
	annos.mCount = 0;

	for(;;) {
		int tok = mLexer.Token();

		if (tok == '>')
			break;

		int type = 0;
		
		if (tok == kTokenIdent)
			type = GetTypeFromIdent();

		if (!type)
			return mLexer.SetError("Expected annotation declaration");

		int size = 0;

		switch(type) {
			case kTypeBool:
				size = 1;
				break;

			case kTypeFloat:
				size = sizeof(float);
				break;

			case kTypeFloat2:
				size = sizeof(float) * 2;
				break;

			case kTypeFloat3:
				size = sizeof(float) * 3;
				break;

			case kTypeFloat4:
				size = sizeof(float) * 4;
				break;

			case kTypeString:
				size = sizeof(int);
				break;

			default:
				return mLexer.SetError("Annotations must be of type float or string");
		}

		tok = mLexer.Token();
		if (tok != kTokenIdent)
			return mLexer.SetError("Expected annotation name");

		int name = mNameTable.AddName(mLexer.mpToken, mLexer.mTokenLength);

		tok = mLexer.Token();
		if (tok != '=' || !ParseExpression())
			return mLexer.SetError("Expected annotation value");

		VDShaderParserVariable& var = mVarHeap.back();

		if (!var.mbIsConst)
			return mLexer.SetError("Annotation values must be constant");

		if (!ConvertToType(var, type))
			return mLexer.SetError("Cannot convert initializer value to annotation type");

		if (var.mAnnotations.mCount)
			return mLexer.SetError("Annotations cannot have annotations");

		if (var.mSemanticToken)
			return mLexer.SetError("Annotations cannot have semantics");

		VDShaderVariable& anno = mAnnotations.push_back();
		anno.mDesc = var;
		anno.mDesc.mNameToken = name;
		anno.mSize = size;
		anno.mValueToken = mDataStorage.size() + 1;

		const char *val = (const char *)var.m[0].v;

		switch(type) {
			case kTypeBool:
				mDataStorage.push_back(var.m[0].v[0] != 0);
				break;
			case kTypeFloat:
			case kTypeFloat2:
			case kTypeFloat3:
			case kTypeFloat4:
				mDataStorage.insert(mDataStorage.end(), val, val + size);
				break;
			case kTypeString:
				val = (const char *)&var.mStringValue;
				mDataStorage.insert(mDataStorage.end(), val, val + sizeof(int));
				break;
		}

		++annos.mCount;

		mVarHeap.pop_back();
		tok = mLexer.Token();

		if (tok != ';')
			return mLexer.SetError("Expected ';'");
	}

	if (!annos.mCount)
		annos.mStart = 0;

	return true;
}

namespace {
	enum ExpressionOp {
		kOpNone,
		kOpAdd,
		kOpSub,
		kOpMul,
		kOpDiv,
		kOpSwizzle,
		kOpArrayIndex,
		kOpEqual,
		kOpNotEqual,
		kOpLessThan,
		kOpGreaterThan,
		kOpLessEqual,
		kOpGreaterEqual,
		kOpAnd,
		kOpOr
	};
}

bool VDShaderParser::ParseExpression() {
	return ParseCommaExpression();
}

bool VDShaderParser::ParseCommaExpression() {
	for(;;) {
		if (!ParseAssignmentExpression())
			return false;

		int tok = mLexer.Token();
		if (tok != ',')
			break;
	}

	mLexer.Push();
	return true;
}

bool VDShaderParser::ParseAssignmentExpression() {
	if (!ParseLogicalOrExpression())
		return false;

	int tok = mLexer.Token();
	int subOp = 0;
	switch(tok) {
		case '?':
			return ParseConditionalExpression();
		case '=':
			break;
		case kTokenAddEq:
			subOp = kOpAdd;
			break;
		case kTokenSubEq:
			subOp = kOpSub;
			break;
		case kTokenMulEq:
			subOp = kOpMul;
			break;
		case kTokenDivEq:
			subOp = kOpDiv;
			break;
		default:
			mLexer.Push();
			return true;
	}

	if (subOp)
		mVarHeap.push_back(mVarHeap.back());

	if (!ParseAssignmentExpression())
		return false;

	if (subOp) {
		if (!ReduceOp(subOp))
			return false;
	}

	VDShaderParserVariable& lvarref = *(mVarHeap.end() - 2);

	if (!lvarref.mbIsLValue)
		return mLexer.SetError("Left argument to assignment operator must be an lvalue");

	VDShaderParserVariable& lvar = mVarHeap[lvarref.mLVRef.mLVIndex];
	VDShaderParserVariable& rvar = mVarHeap.back();

	// invert swizzle and compute write mask

	const VDShaderParserSwizzle& dstSwizzle = lvarref.mLVRef.mSwizzle;
	uint8 dstWriteMask = 0;
	uint8 srcSwizzle = 0xE4;

	for(int i=0; i<dstSwizzle.mLength; ++i) {
		int comp = dstSwizzle.mSwizzle[i];
		uint8 compBit = 1 << comp;

		if (dstWriteMask & compBit)
			return mLexer.SetError("Invalid write mask");

		dstWriteMask |= compBit;

		int swizzleShift = 2*comp;
		srcSwizzle = (srcSwizzle & ~(3 << swizzleShift)) | (i << swizzleShift);
	}

	int dstType = lvar.mType;
	int dstCols = kColCountForType[dstType];
	uint8 fullWriteMask = kColWriteMaskForType[dstType];
	bool writeMaskActive = (dstWriteMask & fullWriteMask) != fullWriteMask;
	bool swizzleActive = 0 != ((srcSwizzle ^ 0xE4) & ((1 << (dstCols+dstCols)) - 1));

	PromoteLValueToRValue(rvar);
	if (!ConvertToType(rvar, lvar.mType))
		return mLexer.SetErrorF("Cannot convert from type '%s' to type '%s'", GetTypeName(rvar.mType), GetTypeName(lvar.mType));

	if (writeMaskActive || swizzleActive) {
		int newReg = mpFunction->AllocateRegisters(1);

		mpFunction->PromoteToConstantRegisters(lvar);
		mpFunction->PromoteToConstantRegisters(rvar);

		int validMask = (1 << dstCols) - 1;
		int preserveMask = validMask ^ dstWriteMask;

		if (preserveMask)
			mpFunction->mov(newReg, preserveMask, lvar.mRegister);

		mpFunction->mov_swizzle(newReg, dstWriteMask, rvar.mRegister, srcSwizzle);

		lvar.mRegister = newReg;
	} else {
		const int name = lvar.mNameToken;
		const int sema = lvar.mSemanticToken;
		const VDShaderAnnotations annos = lvar.mAnnotations;
		const bool isExtern = lvar.mbExtern;
		const bool isStatic = lvar.mbStatic;

		lvar = rvar;
		lvar.mNameToken = name;
		lvar.mSemanticToken = sema;
		lvar.mAnnotations = annos;
		lvar.mbExtern = isExtern;
		lvar.mbStatic = isStatic;
	}

	lvarref = rvar;
	mVarHeap.pop_back();

	return true;
}

bool VDShaderParser::ParseConditionalExpression() {
	if (!ParseExpression())
		return false;

	int tok = mLexer.Token();
	if (tok != ':')
		return mLexer.SetError("Expected ':' in conditional ?: expression");

	if (!ParseAssignmentExpression())
		return false;

	VDShaderParserVariable& var1 = *(mVarHeap.end() - 3);
	VDShaderParserVariable& var2 = *(mVarHeap.end() - 2);
	VDShaderParserVariable& var3 = mVarHeap.back();

	if (!PromoteLValueToRValue(var1))
		return false;

	if (!PromoteLValueToRValue(var2))
		return false;

	if (!PromoteLValueToRValue(var3))
		return false;

	// type check
	if (var2.mType != var3.mType)
		return mLexer.SetError("Second and third arguments to ternary ?: operator must be of the same type");

	bool typesValid = false;
	switch(var1.mType) {
		case kTypeBool:
			switch(var2.mType) {
			case kTypeBool:
			case kTypeBool2:
			case kTypeBool3:
			case kTypeBool4:
			case kTypeFloat:
			case kTypeFloat2:
			case kTypeFloat3:
			case kTypeFloat4:
				typesValid = true;
				break;
			}
			break;

		case kTypeBool2:
			typesValid = (var2.mType == kTypeBool2 || var2.mType == kTypeFloat2);
			break;

		case kTypeBool3:
			typesValid = (var2.mType == kTypeBool3 || var2.mType == kTypeFloat3);
			break;

		case kTypeBool4:
			typesValid = (var2.mType == kTypeBool4 || var2.mType == kTypeFloat4);
			break;
	}

	if (!typesValid)
		return mLexer.SetErrorF("Cannot use ternary ?: operator with types: %s, %s, %s", GetTypeName(var1.mType), GetTypeName(var2.mType), GetTypeName(var3.mType));

	mpFunction->PromoteToConstantRegisters(var2);
	mpFunction->PromoteToConstantRegisters(var3);

	unsigned char writeMask = kColWriteMaskForType[var2.mType];

	if (var1.mbIsConst) {
		if (var1.mType == kTypeBool)
			var1.mRegister = (var1.m[0].v[0] != 0) ? var2.mRegister : var3.mRegister;
		else {
			int t0 = mpFunction->AllocateRegisters(1);
			unsigned char condWriteMask = 0;

			if (var1.m[0].v[0] != 0) condWriteMask += 1;
			if (var1.m[0].v[1] != 0) condWriteMask += 2;
			if (var1.m[0].v[2] != 0) condWriteMask += 4;
			if (var1.m[0].v[3] != 0) condWriteMask += 8;

			condWriteMask &= writeMask;

			if (condWriteMask)
				mpFunction->mov(t0, condWriteMask, var1.mRegister);

			if (condWriteMask != writeMask)
				mpFunction->mov(t0, writeMask ^ condWriteMask, var2.mRegister);

			var1.mRegister = t0;
		}

		var1.mbIsConst = false;
	} else {
		int t0 = mpFunction->AllocateRegisters(1);

		mpFunction->cmp_swizzle_mods(t0, writeMask, var1.mRegister, (var1.mType == kTypeBool ? 0x00 : 0xE4), kVDShaderMod_Negate, var3.mRegister, 0xE4, 0, var2.mRegister, 0xE4, 0);

		var1.mRegister = t0;
	}

	var1.mNameToken = 0;
	var1.mSemanticToken = 0;
	var1.mAnnotations.mCount = 0;
	var1.mAnnotations.mStart = 0;
	var1.mType = var2.mType;

	mVarHeap.pop_back();
	mVarHeap.pop_back();

	return true;
}

bool VDShaderParser::ParseLogicalOrExpression() {
	size_t opBase = mOpStack.size();
	bool expectValue = true;

	for(;;) {
		int tok = mLexer.Token();

		if (expectValue) {
			if (!ParseUnaryExpression(tok))
				return false;

			expectValue = false;
		} else {
			int op = 0;

			if (tok == kTokenOr)			op = ( 3 << 8) + kOpOr;
			else if (tok == kTokenAnd)		op = ( 4 << 8) + kOpAnd;
			else if (tok == kTokenEQ)		op = ( 5 << 8) + kOpEqual;
			else if (tok == kTokenNE)		op = ( 5 << 8) + kOpNotEqual;
			else if (tok == '<')			op = ( 6 << 8) + kOpLessThan;
			else if (tok == '>')			op = ( 6 << 8) + kOpGreaterThan;
			else if (tok == kTokenLE)		op = ( 6 << 8) + kOpLessEqual;
			else if (tok == kTokenGE)		op = ( 6 << 8) + kOpGreaterEqual;
			else if (tok == '+')			op = ( 7 << 8) + kOpAdd;
			else if (tok == '-')			op = ( 7 << 8) + kOpSub;
			else if (tok == '*')			op = ( 8 << 8) + kOpMul;
			else if (tok == '/')			op = ( 8 << 8) + kOpDiv;
			else if (tok == '.')			op = (10 << 8) + kOpSwizzle;
			else if (tok == '[')			op = (10 << 8) + kOpArrayIndex;

			// reduce as necessary
			while(mOpStack.size() > opBase) {
				int otherop = mOpStack.back();

				if (op > otherop)
					break;

				mOpStack.pop_back();

				ExpressionOp baseop = (ExpressionOp)(otherop & 0xff);

				if (!ReduceOp(baseop))
					return false;
			}

			expectValue = true;

			if (!op)
				break;

			int opcode = op & 0xff;
			if (opcode == kOpSwizzle) {
				if (!ParseSwizzleExpression())
					return false;

				expectValue = false;
			} else if (opcode == kOpArrayIndex) {
				if (!ParseArrayIndexExpression())
					return false;

				expectValue = false;
			} else {
				mOpStack.push_back(op);
			}
		}
	}

	mLexer.Push();
	return true;
}

bool VDShaderParser::ReduceOp(int baseop) {
	// coerce
	VDShaderParserVariable& var1 = *(mVarHeap.end() - 2);
	VDShaderParserVariable& var2 = mVarHeap.back();

	if (!PromoteLValueToRValue(var1))
		return false;

	if (!PromoteLValueToRValue(var2))
		return false;

	int commonType = GetBinaryResultType(var1.mType, var2.mType);
	if (!commonType)
		return mLexer.SetErrorF("Cannot operate between types '%s' and '%s'", GetTypeName(var1.mType), GetTypeName(var2.mType));

	ConvertToType(var1, commonType);
	ConvertToType(var2, commonType);

	if (!var1.mbIsConst || !var2.mbIsConst) {
		mpFunction->PromoteToConstantRegisters(var1);
		mpFunction->PromoteToConstantRegisters(var2);
	}

	unsigned char writeMask = (unsigned char)((1 << kColCountForType[commonType]) - 1);

	int rowCount = kRowCountForType[commonType];
	int colCount = kColCountForType[commonType];
	int srcBase1 = var1.mRegister;
	int srcBase2 = var2.mRegister;
	int dstBase = 0;
	
	if (!var1.mbIsConst)
		dstBase = mpFunction->AllocateRegisters(rowCount);

	switch(baseop) {
		case kOpAdd:
			if (!var1.mbIsConst) {
				for(int row=0; row<rowCount; ++row)
					mpFunction->add(dstBase+row, writeMask, srcBase1+row, srcBase2+row);
			} else {
				for(int row=0; row<rowCount; ++row)
					var1.m[row] += var2.m[row];
			}
			break;
		case kOpSub:
			if (!var1.mbIsConst) {
				for(int row=0; row<rowCount; ++row)
					mpFunction->sub(dstBase+row, writeMask, srcBase1+row, srcBase2+row);
			} else {
				for(int row=0; row<rowCount; ++row)
					var1.m[row] -= var2.m[row];
			}
			break;
		case kOpMul:
			if (!var1.mbIsConst) {
				for(int row=0; row<rowCount; ++row)
					mpFunction->mul(dstBase+row, writeMask, srcBase1+row, srcBase2+row);
			} else {
				for(int row=0; row<rowCount; ++row)
					var1.m[row] *= var2.m[row];
			}
			break;
		case kOpDiv:
			if (!var1.mbIsConst) {
				int tmp = mpFunction->AllocateRegisters(1);

				for(int row=0; row<rowCount; ++row) {
					for(unsigned char bit=0; bit<4; ++bit) {
						if (writeMask & (1 << bit))
							mpFunction->rcp(tmp, 1 << bit, srcBase2+row, 0x55*bit);
					}

					mpFunction->mul(dstBase+row, writeMask, srcBase1+row, tmp);
				}
			} else {
				switch(colCount) {
				case 1:
					for(int row=0; row<rowCount; ++row)
						var1.m[row] = div1(var1.m[row], var2.m[row]);
					break;
				case 2:
					for(int row=0; row<rowCount; ++row)
						var1.m[row] = div2(var1.m[row], var2.m[row]);
					break;
				case 3:
					for(int row=0; row<rowCount; ++row)
						var1.m[row] = div3(var1.m[row], var2.m[row]);
					break;
				case 4:
					for(int row=0; row<rowCount; ++row)
						var1.m[row] = div4(var1.m[row], var2.m[row]);
					break;
				}
			}
			break;
		case kOpEqual:
			if (!var1.mbIsConst) {
				const VDShaderConstant c = { 0, 1, 0, 0 };
				int c0 = mpFunction->AddConstant(c);
				int t0 = mpFunction->AllocateRegisters(2);
				int t1 = t0 + 1;

				mpFunction->sub(t0, writeMask, srcBase1, srcBase2);
				mpFunction->cmp_swizzle_mods(t1, writeMask, t0, 0xE4, 0, c0, 0x55, 0, c0, 0x00, 0);
				mpFunction->cmp_swizzle_mods(dstBase, writeMask, t0, 0xE4, kVDShaderMod_Negate, t1, 0, 0, c0, 0x00, 0);
			} else {
				for(int i=0; i<colCount; ++i)
					var1.m[0].v[i] = (var1.m[0].v[i] == var2.m[0].v[i]) ? 1.0f : 0.0f;
			}
			commonType = kTypeBool + (colCount - 1);
			break;
		case kOpNotEqual:
			if (!var1.mbIsConst) {
				const VDShaderConstant c = { 0, 1, 0, 0 };
				int c0 = mpFunction->AddConstant(c);
				int t0 = mpFunction->AllocateRegisters(2);
				int t1 = t0 + 1;

				mpFunction->sub(t0, writeMask, srcBase1, srcBase2);
				mpFunction->cmp_swizzle_mods(t1, writeMask, t0, 0xE4, 0, c0, 0x00, 0, c0, 0x55, 0);
				mpFunction->cmp_swizzle_mods(dstBase, writeMask, t0, 0xE4, kVDShaderMod_Negate, t1, 0, 0, c0, 0x55, 0);
			} else {
				for(int i=0; i<colCount; ++i)
					var1.m[0].v[i] = (var1.m[0].v[i] != var2.m[0].v[i]) ? 1.0f : 0.0f;
			}
			commonType = kTypeBool + (colCount - 1);
			break;
		case kOpLessThan:
			if (!var1.mbIsConst) {
				const VDShaderConstant c = { 0, 1, 0, 0 };
				int c0 = mpFunction->AddConstant(c);
				int t0 = mpFunction->AllocateRegisters(1);

				mpFunction->sub(t0, writeMask, srcBase1, srcBase2);
				mpFunction->cmp_swizzle_mods(dstBase, writeMask, t0, 0xE4, 0, c0, 0, 0, c0, 0x55, 0);
			} else {
				for(int i=0; i<colCount; ++i)
					var1.m[0].v[i] = (var1.m[0].v[i] < var2.m[0].v[i]) ? 1.0f : 0.0f;
			}
			commonType = kTypeBool + (colCount - 1);
			break;
		case kOpGreaterEqual:
			if (!var1.mbIsConst) {
				const VDShaderConstant c = { 0, 1, 0, 0 };
				int c0 = mpFunction->AddConstant(c);
				int t0 = mpFunction->AllocateRegisters(1);

				mpFunction->sub(t0, writeMask, srcBase1, srcBase2);
				mpFunction->cmp_swizzle_mods(dstBase, writeMask, t0, 0xE4, 0, c0, 0x55, 0, c0, 0x00, 0);
			} else {
				for(int i=0; i<colCount; ++i)
					var1.m[0].v[i] = (var1.m[0].v[i] >= var2.m[0].v[i]) ? 1.0f : 0.0f;
			}
			commonType = kTypeBool + (colCount - 1);
			break;
		case kOpGreaterThan:
			if (!var1.mbIsConst) {
				const VDShaderConstant c = { 0, 1, 0, 0 };
				int c0 = mpFunction->AddConstant(c);
				int t0 = mpFunction->AllocateRegisters(1);

				mpFunction->sub(t0, writeMask, srcBase2, srcBase1);
				mpFunction->cmp_swizzle_mods(dstBase, writeMask, t0, 0xE4, 0, c0, 0x00, 0, c0, 0x55, 0);
			} else {
				for(int i=0; i<colCount; ++i)
					var1.m[0].v[i] = (var1.m[0].v[i] > var2.m[0].v[i]) ? 1.0f : 0.0f;
			}
			commonType = kTypeBool + (colCount - 1);
			break;
		case kOpLessEqual:
			if (!var1.mbIsConst) {
				const VDShaderConstant c = { 0, 1, 0, 0 };
				int c0 = mpFunction->AddConstant(c);
				int t0 = mpFunction->AllocateRegisters(1);

				mpFunction->sub(t0, writeMask, srcBase2, srcBase1);
				mpFunction->cmp_swizzle_mods(dstBase, writeMask, t0, 0xE4, 0, c0, 0x55, 0, c0, 0x00, 0);
			} else {
				for(int i=0; i<colCount; ++i)
					var1.m[0].v[i] = (var1.m[0].v[i] <= var2.m[0].v[i]) ? 1.0f : 0.0f;
			}
			commonType = kTypeBool + (colCount - 1);
			break;
		case kOpAnd:
			if (!var1.mbIsConst) {
				int c0 = mpFunction->AddConstantZero();
				int t0 = mpFunction->AllocateRegisters(1);

				// r0 = r1 && r2
				//
				// def c0, 0, 1, 0, 0
				// cmp r0, -r1, c0.x, r2

				mpFunction->cmp_swizzle_mods(dstBase, writeMask, srcBase1, 0xE4, kVDShaderMod_Negate, c0, 0x00, 0, srcBase2, 0xE4, 0);
			} else {
				for(int i=0; i<colCount; ++i)
					var1.m[0].v[i] = (var1.m[0].v[i] != 0 && var2.m[0].v[i] != 0);
			}
			commonType = kTypeBool + (colCount - 1);
			break;
		case kOpOr:
			if (!var1.mbIsConst) {
				const VDShaderConstant c = { 0, 1, 0, 0 };
				int c0 = mpFunction->AddConstant(c);
				int t0 = mpFunction->AllocateRegisters(1);

				// r0 = r1 || r2
				//
				// def c0, 0, 1, 0, 0
				// cmp r0, -r1, r2, c0.y

				mpFunction->cmp_swizzle_mods(dstBase, writeMask, srcBase1, 0xE4, kVDShaderMod_Negate, srcBase2, 0xE4, 0, c0, 0x55, 0);
			} else {
				for(int i=0; i<colCount; ++i)
					var1.m[0].v[i] = (var1.m[0].v[i] != 0 || var2.m[0].v[i] != 0);
			}
			commonType = kTypeBool + (colCount - 1);
			break;
	}

	var1.mType = commonType;
	var1.mRegister = dstBase;
	var1.mNameToken = 0;
	var1.mSemanticToken = 0;
	var1.mAnnotations.mCount = 0;
	var1.mAnnotations.mStart = 0;

	mVarHeap.pop_back();
	return true;
}

bool VDShaderParser::ParseSwizzleExpression() {
	VDShaderParserVariable& v = mVarHeap.back();

	int maxComponents;
	switch(v.mType) {
		case kTypeFloat:
		case kTypeBool:
			maxComponents = 1;
			break;
		case kTypeFloat2:
		case kTypeBool2:
			maxComponents = 2;
			break;
		case kTypeFloat3:
		case kTypeBool3:
			maxComponents = 3;
			break;
		case kTypeFloat4:
		case kTypeBool4:
			maxComponents = 4;
			break;
		default:
			return mLexer.SetErrorF("Cannot swizzle type '%s'", GetTypeName(v.mType));
	}

	// parse swizzle
	VDShaderParserSwizzle swizzle;

	if (!ParseSwizzle(v.mType, maxComponents, swizzle))
		return false;

	if (!ApplySwizzle(v, swizzle))
		return false;

	return true;
}

bool VDShaderParser::ParseArrayIndexExpression() {
	if (!ParseExpression())
		return false;

	int tok = mLexer.Token();
	if (tok != ']')
		return mLexer.SetError("Expected ']' after array index");

	VDShaderParserVariable& base = *(mVarHeap.end() - 2);
	VDShaderParserVariable& index = mVarHeap.back();

	PromoteLValueToRValue(base);
	PromoteLValueToRValue(index);

	if (!ConvertToType(index, kTypeFloat))
		return mLexer.SetError("Index type must be convertable to int");

	if (!index.mbIsConst)
		return mLexer.SetError("Variable indexing not yet supported");

	switch(base.mType) {
		case kTypeFloat:
		case kTypeFloat2:
		case kTypeFloat3:
		case kTypeFloat4:
			break;
		default:
			return mLexer.SetError("Indexing not supported on this type");
	}

	int idx = (int)index.m[0].v[0];
	int cols = kColCountForType[base.mType];

	if (!(idx >= 0 && idx < cols))
		return mLexer.SetError("Array index out of bounds");

	if (base.mbIsConst) {
		base.m[0].v[0] = base.m[0].v[idx];
	} else {
		mpFunction->PromoteToConstantRegisters(base);

		int r = mpFunction->AllocateRegisters(1);

		mpFunction->mov_swizzle(r, 0x0F, base.mRegister, 0x55*idx);
		base.mRegister = r;
	}

	base.mType = kTypeFloat;
	base.mNameToken = 0;
	base.mSemanticToken = 0;
	base.mAnnotations.mCount = 0;
	base.mAnnotations.mStart = 0;

	mVarHeap.pop_back();
	return true;
}

bool VDShaderParser::ParseUnaryExpression(int tok) {
	if (tok == '+') {
		if (!ParseValue(mLexer.Token()))
			return false;

		VDShaderParserVariable& var = mVarHeap.back();

		if (!PromoteLValueToRValue(var))
			return false;

		if (!kRowCountForType[var.mType])
			return mLexer.SetError("Unary '+' can only be applied to numeric types");

		return true;
	} else if (tok == '-') {
		if (!ParseValue(mLexer.Token()))
			return false;

		VDShaderParserVariable& v = mVarHeap.back();

		if (!PromoteLValueToRValue(v))
			return false;

		if (!kRowCountForType[v.mType])
			return mLexer.SetError("Unary '-' can only be applied to numeric types");

		if (v.mbIsConst) {
			v.m[0] = -v.m[0];
			v.m[1] = -v.m[1];
			v.m[2] = -v.m[2];
			v.m[3] = -v.m[3];
		} else {
			int rows = kRowCountForType[v.mType];
			unsigned char writeMask = (unsigned char)((1 << kColCountForType[v.mType]) - 1);
			int r = mpFunction->AllocateRegisters(rows);

			for(int i=0; i<rows; ++i)
				mpFunction->mov_mods(r+i, writeMask, v.mRegister, 0xE4, kVDShaderMod_Negate);

			v.mRegister = r;
		}

		return true;
	}

	return ParseValue(tok);
}

bool VDShaderParser::ParseValue(int tok) {
	if (tok == '(') {
		if (!ParseExpression())
			return false;

		tok = mLexer.Token();
		if (tok != ')')
			return mLexer.SetError("Missing ')' in expression");
		return true;
	}

	if (tok == kTokenIdent) {
		int type = GetTypeFromIdent();
		if (type) {
			if (!ParseCtorExpression(type))
				return false;

			return true;
		}

		if (mLexer.mTokenLength == 13 && !memcmp(mLexer.mpToken, "sampler_state", 13)) {
			if (!ParseSamplerState())
				return false;

			return true;
		}

		if (mLexer.mTokenLength == 7 && !memcmp(mLexer.mpToken, "compile", 7)) {
			if (!ParseCompileExpression())
				return false;

			return true;
		}

		if (mLexer.mTokenLength == 4 && !memcmp(mLexer.mpToken, "true", 4)) {
			VDShaderParserVariable v;

			v.m[0].v[0] = 1.0f;
			v.m[0].v[1] = 1.0f;
			v.m[0].v[2] = 1.0f;
			v.m[0].v[3] = 1.0f;
			v.mType = kTypeBool;
			v.mbIsConst = true;
			mVarHeap.push_back(v);
			return true;
		}

		if (mLexer.mTokenLength == 5 && !memcmp(mLexer.mpToken, "false", 5)) {
			VDShaderParserVariable v;

			v.m[0].v[0] = 0.0f;
			v.m[0].v[1] = 0.0f;
			v.m[0].v[2] = 0.0f;
			v.m[0].v[3] = 0.0f;
			v.mType = kTypeBool;
			v.mbIsConst = true;
			mVarHeap.push_back(v);
			return true;
		}

		int varName = GetNameToken();
		const VDShaderIntrinsicEntry *pBaseIntrinsic = VDShaderParserGetIntrinsicByName(GetName(varName));

		tok = mLexer.Token();
		if (tok == '(') {
			int argCount = 0;
			int argBase = mVarHeap.size();

			tok = mLexer.Token();
			if (tok != ')') {
				mLexer.Push();

				for(;;) {
					++argCount;
					if (!ParseAssignmentExpression())
						return false;

					PromoteLValueToRValue(mVarHeap.back());

					tok = mLexer.Token();
					if (tok == ')')
						break;

					if (tok != ',')
						return mLexer.SetError("Syntax error in argument list");
				}
			}

			if (pBaseIntrinsic) {
				if (!argCount) {
					VDShaderParserVariable var;
					var.mType = 0;
					var.mbIsConst = false;
					mVarHeap.push_back(var);
				}

				if (!VDShaderParserCompileIntrinsic(mLexer, *mpFunction, pBaseIntrinsic, &mVarHeap[argBase], argCount))
					return false;

				if (argCount > 1)
					mVarHeap.resize(argBase + 1);

				return true;
			} else {
				Functions::const_iterator it(mFunctions.begin()), itEnd(mFunctions.end());
				for(; it!=itEnd; ++it) {
					VDShaderParserFunction& calledFunc = **it;
					if (calledFunc.mNameToken == varName) {
						// check argument count
						int expectedArgCount = calledFunc.mParameters.size();
						if (argCount != expectedArgCount)
							return mLexer.SetErrorF("Expected %d arguments to function '%s', found %d\n", expectedArgCount, GetName(varName), argCount);

						// coerce arguments
						for(int i=0; i<argCount; ++i) {
							VDShaderParserVariable& arg = mVarHeap[argBase + i];
							const VDShaderVariableDesc& argVar = calledFunc.mParameters[i];

							if (!ConvertToType(arg, argVar.mType))
								return mLexer.SetErrorF("Cannot convert argument %d from type '%s' to type '%s'", i+1, GetTypeName(arg.mType), GetTypeName(argVar.mType));
						}

						// inline the function
						VDShaderParserVariable returnValue;
						if (!mpFunction->Inline(mLexer, calledFunc, argCount ? &mVarHeap[argBase] : NULL, returnValue))
							return false;

						returnValue.mbIsConst = false;

						mVarHeap.resize(argBase + 1);
						mVarHeap.back() = returnValue;

						return true;
					}
				}

				return mLexer.SetErrorF("Undefined function '%s'", GetName(varName));
			}
		} else {
			mLexer.Push();

			if (pBaseIntrinsic)
				return mLexer.SetErrorF("Expected argument list for intrinsic '%s'", pBaseIntrinsic->mpName);
		}

		int varIndex = FindVariable(varName, false);
		if (varIndex < 0)
			return mLexer.SetErrorF("Undefined variable '%s'", GetName(varName));

		const VDShaderParserVariable& varRef = mVarHeap[varIndex];

		VDShaderParserVariable varCopy(varRef);

		if (varCopy.mType != kTypeSampler) {
			if (!varCopy.mbIsConst && !varCopy.mRegister) {
				if (!varCopy.mNameToken)
					return mLexer.SetErrorF("Reference to uninitialized variable '%s'", GetName(varName));
			}
		}

		varCopy.mbIsLValue = true;
		varCopy.mLVRef.mLVIndex = varIndex;
		varCopy.mLVRef.mSwizzle.mLength = kColCountForType[varCopy.mType];
		varCopy.mLVRef.mSwizzle.mPackedSwizzle = 0xE4;
		varCopy.mLVRef.mSwizzle.mSwizzle[0] = 0;
		varCopy.mLVRef.mSwizzle.mSwizzle[1] = 1;
		varCopy.mLVRef.mSwizzle.mSwizzle[2] = 2;
		varCopy.mLVRef.mSwizzle.mSwizzle[3] = 3;
		mVarHeap.push_back(varCopy);
		return true;
	} else if (tok == kTokenString) {
		int stok = mNameTable.AddName(mLexer.mpToken, mLexer.mTokenLength);

		VDShaderParserVariable constVar;
		constVar.mType = kTypeString;
		constVar.mbIsConst = true;
		constVar.mStringValue = stok;

		mVarHeap.push_back(constVar);
		return true;		
	} else if (tok == kTokenFloat || tok == kTokenInt) {
		VDShaderParserVariable constVar;
		constVar.mType = kTypeFloat;
		constVar.mbIsConst = true;
		memset(constVar.m, 0, sizeof constVar.m);

		if (tok == kTokenFloat)
			constVar.m[0].v[0] = mLexer.mFloatVal;
		else
			constVar.m[0].v[0] = (float)mLexer.mIntVal;

		mVarHeap.push_back(constVar);
		return true;
	}

	return mLexer.SetError("Expected expression value");
}

bool VDShaderParser::ParseSwizzle(int type, int maxComponents, VDShaderParserSwizzle& swizzle) {
	int tok = mLexer.Token();

	if (tok != kTokenIdent)
		return mLexer.SetError("Expected swizzle");

	int swizzleLength = mLexer.mTokenLength;

	if (swizzleLength > 4)
		return mLexer.SetError("Invalid swizzle");

	for(int i=0; i<swizzleLength; ++i) {
		int component;
		switch(mLexer.mpToken[i]) {
			case 'x':	case 'r':	component = 0; break;
			case 'y':	case 'g':	component = 1; break;
			case 'z':	case 'b':	component = 2; break;
			case 'w':	case 'a':	component = 3; break;
			default:
				return mLexer.SetError("Invalid swizzle");
		}

		if (component > maxComponents)
			return mLexer.SetErrorF("Invalid swizzle component for type '%s'", GetTypeName(type));

		swizzle.mSwizzle[i] = component;
	}

	for(int i=swizzleLength; i<4; ++i)
		swizzle.mSwizzle[i] = swizzle.mSwizzle[swizzleLength - 1];

	swizzle.mLength = swizzleLength;
	swizzle.mPackedSwizzle	=  swizzle.mSwizzle[0]
							+ (swizzle.mSwizzle[1] << 2)
							+ (swizzle.mSwizzle[2] << 4)
							+ (swizzle.mSwizzle[3] << 6);

	return true;
}

bool VDShaderParser::ParseSamplerState() {
	int tok = mLexer.Token();

	if (tok != '{')
		return mLexer.SetError("'sampler_state' must be followed by '{'");

	enum {
		kSamplerStateTexture,
		kSamplerStateAddressU,
		kSamplerStateAddressV,
		kSamplerStateAddressW,
		kSamplerStateMinFilter,
		kSamplerStateMagFilter,
		kSamplerStateMipFilter,
		kSamplerStateBorderColor,
		kSamplerStateMaxMipLevel,
		kSamplerStateSRGBTexture,
		kSamplerStateMaxAnisotropy,
		kSamplerStateMipMapLODBias
	};

	static const char *const kAddressingModes[]={
		"clamp",
		"wrap",
		NULL
	};

	static const char *const kFilterModes[]={
		"point",
		"linear",
		NULL
	};

	static const char *const kMipFilterModes[]={
		"none",
		"point",
		"linear",
		NULL
	};

	static const char *const kBoolStates[]={
		"false",
		"true",
		NULL
	};

	static const char *const kNullStates[]={
		"false",
		"true",
		NULL
	};

	VDShaderSamplerState samplerState = {};

	for(;;) {
		tok = mLexer.Token();
		if (tok == '}')
			break;

		if (tok != kTokenIdent)
			return mLexer.SetError("Expected sampler state");

		int state = -1;
		int checkType = 0;
		switch(mLexer.mTokenLength) {
			case 7:
				if (!_memicmp(mLexer.mpToken, "texture", 7)) {
					state = kSamplerStateTexture;
					checkType = kTypeTexture;
				}
				break;
			case 8:
				if (!_memicmp(mLexer.mpToken, "addressu", 8)) {
					state = kSamplerStateAddressU;
				} else if (!_memicmp(mLexer.mpToken, "addressv", 8)) {
					state = kSamplerStateAddressV;
				} else if (!_memicmp(mLexer.mpToken, "addressw", 8)) {
					state = kSamplerStateAddressW;
				}
				break;
			case 9:
				if (!_memicmp(mLexer.mpToken, "minfilter", 9)) {
					state = kSamplerStateMinFilter;
				} else if (!_memicmp(mLexer.mpToken, "magfilter", 9)) {
					state = kSamplerStateMagFilter;
				} else if (!_memicmp(mLexer.mpToken, "mipfilter", 9)) {
					state = kSamplerStateMipFilter;
				}
				break;
			case 11:
				if (!_memicmp(mLexer.mpToken, "bordercolor", 11)) {
					state = kSamplerStateBorderColor;
				} else if (!_memicmp(mLexer.mpToken, "maxmiplevel", 11)) {
					state = kSamplerStateMaxMipLevel;
				} else if (!_memicmp(mLexer.mpToken, "srgbtexture", 11)) {
					state = kSamplerStateSRGBTexture;
				}
				break;
			case 13:
				if (!_memicmp(mLexer.mpToken, "maxanisotropy", 13)) {
					state = kSamplerStateMaxAnisotropy;
				} else if (!_memicmp(mLexer.mpToken, "mipmaplodbias", 13)) {
					state = kSamplerStateMipMapLODBias;
				}
				break;
		}

		if (state < 0)
			return mLexer.SetErrorF("Unknown sampler state '%.*s'", mLexer.mTokenLength, mLexer.mpToken);

		// parse sampler state

		tok = mLexer.Token();
		if (tok != '=')
			return mLexer.SetError("Expected '=' after sampler state name");

		// parse sampler value
		int value;
		int basetok = mLexer.Token();
		if (basetok == kTokenIdent) {
			mLexer.Push();
			switch(state) {
				case kSamplerStateTexture:
					if (!ParseEnum(kNullStates, value))
						return false;
					break;
				case kSamplerStateAddressU:
				case kSamplerStateAddressV:
				case kSamplerStateAddressW:
					if (!ParseEnum(kAddressingModes, value))
						return false;
					break;
				case kSamplerStateMinFilter:
				case kSamplerStateMagFilter:
					if (!ParseEnum(kFilterModes, value))
						return false;
					break;
				case kSamplerStateMipFilter:
					if (!ParseEnum(kMipFilterModes, value))
						return false;
					break;
				case kSamplerStateSRGBTexture:
					if (!ParseEnum(kBoolStates, value))
						return false;
					break;
				case kSamplerStateMaxAnisotropy:
				case kSamplerStateMipMapLODBias:
					return mLexer.SetError("Expected expression");
			}
		} else if (basetok == '(' || basetok == '<') {
			if (!ParseExpression())
				return false;

			PromoteLValueToRValue(mVarHeap.back());
			value = mVarHeap.back().mTextureBinding;

			mVarHeap.pop_back();

			int matchtok = (basetok == '(') ? ')' : '>';
			tok = mLexer.Token();
			if (tok != matchtok)
				return mLexer.SetErrorF("Expected '%c' after sampler state expression", matchtok);
		} else if (basetok == kTokenFloat) {
			value = (int)mLexer.mFloatVal;
		} else if (basetok == kTokenInt) {
			value = mLexer.mIntVal;
		} else
			return mLexer.SetError("Expected sampler state value");

		switch(state) {
			case kSamplerStateTexture:
				samplerState.mTextureToken = value;
				break;
			case kSamplerStateAddressU:
				samplerState.mbAddressUWrapSet = true;
				samplerState.mbAddressUWrap = (value != 0);
				break;
			case kSamplerStateAddressV:
				samplerState.mbAddressVWrapSet = true;
				samplerState.mbAddressVWrap = (value != 0);
				break;
			case kSamplerStateAddressW:
				break;
			case kSamplerStateMinFilter:
				samplerState.mbMinFilterSet = true;
				samplerState.mbMinFilter = (value != 0);
				break;
			case kSamplerStateMagFilter:
				samplerState.mbMagFilterSet = true;
				samplerState.mbMagFilter = (value != 0);
				break;
			case kSamplerStateMipFilter:
				samplerState.mbMipFilterSet = true;
				samplerState.mbMipFilter = (value != 0);
				break;
			case kSamplerStateSRGBTexture:
			case kSamplerStateMaxAnisotropy:
			case kSamplerStateMipMapLODBias:
				break;
		}

		tok = mLexer.Token();
		if (tok != ';')
			return mLexer.SetError("Expected ';' after sampler state declaration");
	}

	mpEffect->mSamplerStates.push_back(samplerState);

	VDShaderParserVariable var;
	var.mType = kTypeSampler;
	var.mSamplerBinding.mNameToken = 0;
	var.mSamplerBinding.mStateToken = mpEffect->mSamplerStates.size();
	var.mbIsConst = true;
	mVarHeap.push_back(var);

	return true;
}

bool VDShaderParser::ParseEnum(const char *const *enumValues, int& value) {
	int tok = mLexer.Token();
	if (tok != kTokenIdent)
		return mLexer.SetError("Expected state enumeration value");

	int i = 0;
	while(const char *s = *enumValues++) {
		size_t klen = strlen(s);

		if (mLexer.mTokenLength == klen && !_memicmp(mLexer.mpToken, s, klen)) {
			value = i;
			return true;
		}

		++i;
	}

	return mLexer.SetErrorF("Unknown state value '%.*s'", mLexer.mTokenLength, mLexer.mpToken);
}

bool VDShaderParser::ParseCompileExpression() {
	int tok = mLexer.Token();

	if (tok != kTokenIdent)
		return mLexer.SetError("Expected shader profile after 'compile' keyword");

	tok = mLexer.Token();
	if (tok != kTokenIdent)
		return mLexer.SetError("Function name expected");

	int nameToken = GetNameToken();

	VDShaderParserFunction *pFunc = NULL;
	for(Functions::const_iterator it(mFunctions.begin()), itEnd(mFunctions.end()); it!=itEnd; ++it) {
		VDShaderParserFunction& calledFunc = **it;
		if (calledFunc.mNameToken == nameToken) {
			pFunc = &calledFunc;
			break;
		}
	}

	if (!pFunc)
		return mLexer.SetErrorF("Unknown function '%s'", GetName(nameToken));

	// start new function
	mpFunction = new VDShaderParserFunction;
	mpFunction->mNameToken = pFunc->mNameToken;
	mpFunction->mReturnType = pFunc->mReturnType;
	mpFunction->mReturnSemantic = pFunc->mReturnSemantic;
	mpFunction->mIndex = mFunctions.size();
	mFunctions.push_back(mpFunction);

	VDShaderFunction::Parameters::const_iterator itPBegin(pFunc->mParameters.begin()), itPEnd(pFunc->mParameters.end()), itP(itPBegin);

	// copy over and allocate all non-uniform parameters
	int uniformArgCount = 0;
	for(itP = itPBegin; itP != itPEnd; ++itP) {
		const VDShaderVariableDesc& parm = *itP;
		if (parm.mbUniform) {
			++uniformArgCount;
			continue;
		}

		VDShaderVariableDesc newParm(parm);

		if (int rows = kRowCountForType[parm.mType])
			newParm.mRegister = mpFunction->AllocateParamRegisters(rows);

		mpFunction->mParameters.push_back(newParm);
	}

	tok = mLexer.Token();
	if (tok != '(')
		return mLexer.SetError("Expected '(' after function name");

	int varBase = mVarHeap.size();
	int parmIndex = 0;

	tok = mLexer.Token();
	itP = itPBegin;
	bool endOfArgs = true;

	if (tok != ')') {
		mLexer.Push();
		endOfArgs = false;
	}

	int nonUniformIndex = 0;
	while(itP != itPEnd) {
		const VDShaderVariableDesc& parmvar = *itP++;
		int rows = kRowCountForType[parmvar.mType];
		int cols = kColCountForType[parmvar.mType];
		unsigned char mask = (1 << cols) - 1;

		// if we have a non-uniform parameter, alias it against our argument
		if (!parmvar.mbUniform) {
			const VDShaderVariableDesc& inputParm = mpFunction->mParameters[nonUniformIndex++];

			VDShaderParserVariable nonUniParm;
			nonUniParm.mNameToken = parmvar.mNameToken;
			nonUniParm.mSemanticToken = parmvar.mSemanticToken;
			nonUniParm.mAnnotations = parmvar.mAnnotations;
			nonUniParm.mType = parmvar.mType;
			nonUniParm.mRegister = inputParm.mRegister;
			nonUniParm.mbIsConst = false;
			mVarHeap.push_back(nonUniParm);
			continue;
		}

		if (endOfArgs)
			return mLexer.SetErrorF("Function has %d uniform arguments; only %d provided", uniformArgCount, parmIndex);

		// uniform argument - add to constants
		if (!ParseAssignmentExpression())
			return false;

		VDShaderParserVariable& arg = mVarHeap.back();
		PromoteLValueToRValue(arg);

		if (!arg.mbIsConst && arg.mType != kTypeSampler)
			return mLexer.SetError("Function argument must be a constant expression");

		if (!ConvertToType(arg, parmvar.mType))
			return mLexer.SetErrorF("Cannot convert parameter %d from type '%s' to type '%s'", parmIndex+1, GetTypeName(arg.mType), GetTypeName(parmvar.mType));

		VDShaderParserVariable argVar;
		argVar.mbIsConst = false;
		argVar.mbUniform = false;
		argVar.mType = arg.mType;
		for(int row=0; row<rows; ++row) {
			int r = mpFunction->AddConstant(arg.m[row]);

			if (!row)
				argVar.mRegister = r;
		}

		++parmIndex;

		tok = mLexer.Token();
		if (tok == ')')
			endOfArgs = true;
		else if (tok != ',')
			return mLexer.SetError("Expected ',' or ')' in function argument list");
	}

	if (!endOfArgs)
		return mLexer.SetErrorF("Too many arguments for function '%s'", GetName(pFunc->mNameToken));

	// specialize the function
	VDShaderVariableDesc retVar;
	if (!mpFunction->Inline(mLexer, *pFunc, itPBegin != itPEnd ? &mVarHeap[varBase] : NULL, retVar))
		return false;

	int retrows = kRowCountForType[retVar.mType];
	int retcols = kColCountForType[retVar.mType];
	for(int row=0; row<retrows; ++row)
		mpFunction->mov(kVDShaderRegO0 + row, (1 << retcols) - 1, retVar.mRegister + row);

	if (mbOptimizationEnabled && !mpFunction->Optimize(mLexer, mNameTable))
		return false;

	if (!mpFunction->BakeHardcodedRegisters(mLexer))
		return false;

	mVarHeap.resize(varBase);

	// construct pixel shader variable
	VDShaderParserVariable v;
	v.mType = kTypePixelShader;
	v.mbIsConst = true;
	v.mShader = mpFunction->mIndex;
	mVarHeap.push_back(v);

	mpFunction = NULL;
	return true;
}

bool VDShaderParser::ParseCtorExpression(int type) {
	if (type == kTypeVoid)
		return mLexer.SetError("Cannot create value of type 'void'");

	int tok = mLexer.Token();
	if (tok != '(')
		return mLexer.SetError("Expected '(' after type");

	if (!ParseInitializerList(type, true))
		return false;

	tok = mLexer.Token();
	if (tok != ')')
		return mLexer.SetError("Expected ')' after ctor-expression");

	return true;
}

bool VDShaderParser::ParseInitializerList(int type, bool requireFullInitialization) {
	int rows = kRowCountForType[type];
	int cols = kColCountForType[type];

	VDShaderParserVariable newVar;
	newVar.mType = type;
	memset(newVar.m, 0, sizeof newVar.m);

	uint32 constMask = 0;
	bool nonConst = false;

	for(int row=0; row<rows; ++row) {
		int col = 0;

		while(col < cols) {
			if (row+col) {
				int tok = mLexer.Token();
				if (tok != ',') {
					if (!requireFullInitialization)
						break;

					return mLexer.SetError("Expected ',' in vector constructor argument list");
				}
			}

			if (!ParseAssignmentExpression())
				return false;

			VDShaderParserVariable argVar(mVarHeap.back());
			mVarHeap.pop_back();

			PromoteLValueToRValue(argVar);
			mpFunction->PromoteToConstantRegisters(argVar);

			int argRows = kRowCountForType[argVar.mType];
			int argCols = kColCountForType[argVar.mType];

			if (argRows > 1)
				return mLexer.SetError("Cannot initialize with a multi-row matrix");

			if (argCols + col > 4)
				return mLexer.SetErrorF("Too many elements in row %d", row);

			if (argVar.mbIsConst) {
				memcpy(&newVar.m[row].v[col], argVar.m[0].v, sizeof(float)*argCols);
				constMask |= ((1 << argCols) - 1) << (col + row*4);
			} else {
				if (!nonConst)
					newVar.mRegister = mpFunction->AllocateRegisters(rows);

				mpFunction->mov_swizzle(newVar.mRegister + row, ((1 << argCols) - 1) << col, argVar.mRegister, 0xE4 << (col+col));
				nonConst = true;
			}

			col += argCols;
		}
	}

	// if we have a mix of const and non-const values, merge the const values into the temp registers
	if (nonConst) {
		newVar.mbIsConst = false;

		for(int row=0; row<rows; ++row) {
			unsigned char mask = (constMask >> (4*row)) & 15;

			if (mask) {
				int r = mpFunction->AddConstant(newVar.m[row]);

				mpFunction->mov(newVar.mRegister + row, mask, r);
			}
		}
	} else {
		newVar.mbIsConst = true;
	}

	mVarHeap.push_back(newVar);

	int tok = mLexer.Token();
	if (tok == ',')
		return mLexer.SetError("Too many initializers");

	mLexer.Push();

	return true;
}

int VDShaderParser::GetTypeFromIdent() const {
	const char *s = mLexer.mpToken;

	switch(mLexer.mTokenLength) {
		case 4:
			if (!memcmp(s, "void", 4))
				return kTypeVoid;
			else if (!memcmp(s, "bool", 4))
				return kTypeBool;
			else if (!memcmp(s, "half", 4))
				return kTypeFloat;		// map half -> float
			break;
		case 5:
			if (!memcmp(s, "half", 4)) {
				if (s[4] == '2')
					return kTypeFloat2;
				else if (s[4] == '3')
					return kTypeFloat3;
				else if (s[4] == '4')
					return kTypeFloat4;
			} else if (!memcmp(s, "float", 5))
				return kTypeFloat;
			else if (!memcmp(s, "bool", 4)) {
				if (s[4] == '2')
					return kTypeBool2;
				else if (s[4] == '3')
					return kTypeBool3;
				else if (s[4] == '4')
					return kTypeBool4;
			}
			break;
		case 6:
			if (!memcmp(s, "float", 5)) {
				if (s[5] == '2')
					return kTypeFloat2;
				else if (s[5] == '3')
					return kTypeFloat3;
				else if (s[5] == '4')
					return kTypeFloat4;
			} else if (!memcmp(s, "string", 6))
				return kTypeString;
			break;
		case 7:
			if (!memcmp(s, "sampler", 7))
				return kTypeSampler;
			else if (!memcmp(s, "texture", 7))
				return kTypeTexture;
			break;
		case 8:
			if (!memcmp(s, "float", 5) && s[6] == 'x') {
				unsigned rows = (unsigned)(s[5] - '0');
				unsigned cols = (unsigned)(s[7] - '0');

				if ((rows - 2) < 3 && (cols - 2) < 3)
					return kTypeFloat2x2 + (rows*3) + cols - 8;
			}
			break;
		case 11:
			if (!memcmp(s, "pixelshader", 11))
				return kTypePixelShader;
			break;
		case 12:
			if (!memcmp(s, "vertexshader", 12))
				return kTypeVertexShader;
			break;
	}

	return kTypeInvalid;
}

int VDShaderParser::GetNameToken() {
	int token = mNameTable.AddName(mLexer.mpToken, mLexer.mTokenLength);

	return token;
}

const char *VDShaderParser::GetTypeName(int type) const {
	return VDShaderGetTypeName(type);
}

const char *VDShaderParser::GetName(int name) const {
	return mNameTable.GetName(name);
}

int VDShaderParser::FindVariable(int name, bool localOnly) {
	int lvindex = FindLValue(name, localOnly);

	return lvindex;
}

int VDShaderParser::FindLValue(int name, bool localOnly) {
	int startOffset = 0;

	if (localOnly)
		startOffset = mCurrentScopeLevel;

	VDASSERT((size_t)mCurrentScopeLimit <= mVarHeap.size());

	VarHeap::const_iterator it(mVarHeap.begin() + startOffset), itEnd(mVarHeap.begin() + mCurrentScopeLimit);

	while(it != itEnd) {
		--itEnd;

		const VDShaderVariableDesc& var = *itEnd;

		if (var.mNameToken == name)
			return itEnd - mVarHeap.begin();
	}

	return -1;
}

bool VDShaderParserConvertVarToType(IVDShaderErrorOutput& lexer, VDShaderParserFunction *func, VDShaderParserVariable& var, int dstType) {
	int srcType = var.mType;

	if (srcType == dstType)
		return true;

	int srcRows = kRowCountForType[srcType];
	int srcCols = kColCountForType[srcType];
	unsigned char srcswizzle = 0xE4;
	if (srcCols == 1) {
		srcswizzle = 0;
		srcCols = 4;
	}

	int dstRows = kRowCountForType[dstType];
	int dstCols = kColCountForType[dstType];

	if (!srcRows || !dstRows)
		return false;

	int minRows = srcRows < dstRows ? srcRows : dstRows;
	int minCols = srcCols < dstCols ? srcCols : dstCols;

	if (!var.mbIsConst) {
		if (!func)
			return lexer.SetError("Constant expression required for global initializer");

		func->PromoteToConstantRegisters(var);

		int srcBase = var.mRegister;
		int dstBase = func->AllocateRegisters(dstRows);

		unsigned char copymask = (1 << minCols) - 1;
		unsigned char zeromask = ((1 << dstCols) - 1) & ~copymask;

		int czero = func->AddConstantZero();

		for(int row=0; row<minRows; ++row) {
			if (copymask)
				func->mov_swizzle(dstBase+row, copymask, srcBase+row, srcswizzle);

			if (zeromask)
				func->mov(dstBase+row, zeromask, czero);
		}

		for(int row=minRows; row<dstRows; ++row)
			func->mov(dstBase+row, 0xF, czero);

		var.mRegister = dstBase;
	} else {
		if (srcswizzle == 0) {
			for(int row=0; row<minRows; ++row) {
				VDShaderConstant& c = var.m[row];
				for(int col=1; col<dstCols; ++col)
					c.v[col] = c.v[0];
			}
		} else if (dstCols < srcCols) {
			for(int row=0; row<minRows; ++row)
				memset(&var.m[row].v[dstCols], 0, sizeof(float)*(srcCols - dstCols));
		}

		if (dstRows < srcRows)
			memset(&var.m[dstRows].v, 0, sizeof(float)*4*(srcRows - dstRows));
	}

	var.mType = dstType;
	return true;
}

bool VDShaderParser::ApplySwizzle(VDShaderParserVariable& var, const VDShaderParserSwizzle& swizzle) {
	int maxComponents = kColCountForType[var.mType];
	if (!maxComponents || kRowCountForType[var.mType] != 1)
		return mLexer.SetErrorF("Cannot swizzle type '%s'", GetTypeName(var.mType));

	if (var.mbIsLValue) {
		const VDShaderParserSwizzle oldSwizzle(var.mLVRef.mSwizzle);
		uint8 packedSwizzle = 0;

		for(int i=0; i<swizzle.mLength; ++i) {
			uint8 srcComponent = swizzle.mSwizzle[i];

			if (srcComponent >= oldSwizzle.mLength)
				return mLexer.SetError("Swizzle contains component not available in swizzled value");

			uint8 dstComponent = oldSwizzle.mSwizzle[srcComponent];
			var.mLVRef.mSwizzle.mSwizzle[i] = dstComponent;
			packedSwizzle += dstComponent << (2*i);
		}

		var.mLVRef.mSwizzle.mLength = swizzle.mLength;
		var.mLVRef.mSwizzle.mPackedSwizzle = packedSwizzle;
	} else {
		uint8 mask = 0xFF >> (8 - 2*std::min<int>(maxComponents, swizzle.mLength));

		PromoteLValueToRValue(var);

		const uint8 packedSwizzle =  swizzle.mSwizzle[0]
								  + (swizzle.mSwizzle[1] << 2)
								  + (swizzle.mSwizzle[2] << 4)
								  + (swizzle.mSwizzle[3] << 6);

		if (((packedSwizzle ^ 0xE4) & mask) || swizzle.mLength > maxComponents) {
			if (var.mbIsConst) {
				var.m[0] = nsVDShaderArith::swizzle(var.m[0], packedSwizzle);
			} else {
				if (!mpFunction)
					return mLexer.SetError("Constant expression required for global initializer");

				mpFunction->PromoteToConstantRegisters(var);

				const int rA = mpFunction->AllocateRegisters(1);
				const uint8 writeMask = (1 << swizzle.mLength) - 1;

				mpFunction->mov_swizzle(rA, writeMask, var.mRegister, packedSwizzle);

				if (writeMask != 0x0F) {
					const int rZ = mpFunction->AddConstantZero();
					mpFunction->mov(rA, 0x0F ^ writeMask, rZ);
				}

				var.mRegister = rA;
			}
		}
	}

	var.mType = var.mType - maxComponents + swizzle.mLength;

	if (!var.mbUnassigned)
		var.mNameToken = 0;

	var.mSemanticToken = 0;
	var.mAnnotations.mStart = 0;
	var.mAnnotations.mCount = 0;

	return true;
}

bool VDShaderParser::PromoteLValueToRValue(VDShaderParserVariable& var) {
	if (!var.mbIsLValue)
		return true;

	VDShaderParserSwizzle swizzle = var.mLVRef.mSwizzle;
	int oldType = var.mType;
	var = mVarHeap[var.mLVRef.mLVIndex];
	var.mType = oldType;

	if (var.mType != kTypeSampler && var.mType != kTypeTexture) {
		if (!var.mbIsConst && !var.mRegister && !var.mbUnassigned)
			return mLexer.SetErrorF("Reference to uninitialized variable '%s'", GetName(var.mNameToken));
	}

	if (kColCountForType[var.mType] && kRowCountForType[var.mType] == 1) {
		if (!ApplySwizzle(var, swizzle))
			return false;
	}

	return true;
}

bool VDShaderParser::ConvertToType(VDShaderParserVariable& var, int dstType) {
	return VDShaderParserConvertVarToType(mLexer, mpFunction, var, dstType);
}

bool VDShaderParser::ValidateFunctionAsPixelShader(const VDShaderParserFunction& func) {
	// check return types
	if (func.mReturnType != kTypeFloat4)
		return mLexer.SetErrorF("'%s': Pixel shader must return float4", GetName(func.mNameToken));

	const char *returnSemantic = GetName(func.mReturnSemantic);
	if (strcmp(returnSemantic, "COLOR0") && strcmp(returnSemantic, "COLOR"))
		return mLexer.SetErrorF("'%s': Pixel shader return value must have COLOR/COLOR0 semantic", GetName(func.mNameToken));

	// check arguments
	int texcoordToken = mNameTable.GetName("TEXCOORD", 8);
	int texcoord0Token = mNameTable.GetName("TEXCOORD0", 9);

	for(VDShaderParserFunction::Parameters::const_iterator it(func.mParameters.begin()), itEnd(func.mParameters.end()); it!=itEnd; ++it) {
		const VDShaderVariableDesc& var = *it;

		if (!var.mSemanticToken)
			return mLexer.SetErrorF("Invalid pixel shader function '%s': missing semantic on parameter '%s'", GetName(func.mNameToken), GetName(var.mNameToken));

		if (var.mSemanticToken != texcoord0Token && var.mSemanticToken != texcoordToken)
			return mLexer.SetErrorF("Unsupported semantic for pixel shader; must be 'TEXCOORD' or 'TEXCOORD0'");
	}

	if (!func.Verify(mLexer))
		return false;

	return true;
}
