#include "stdafx.h"
#include "preprocessor.h"

using namespace nsVDShaderTokens;

VDShaderPreprocessor::VDShaderPreprocessor() {
}

VDShaderPreprocessor::~VDShaderPreprocessor() {
}

void VDShaderPreprocessor::Init(const char *src, size_t len, const char *context, IVDShaderLogOutput *pOutput) {
	mLexer.Init(src, len, context, pOutput);
	mLexer.SetReturnWhitespace(true);
	mbTokenQueued = false;
}

int VDShaderPreprocessor::Token() {
	if (mbTokenQueued) {
		mbTokenQueued = false;
		return mLastToken;
	}

	if (!mMacroExpansion.empty()) {
		const ExpandedToken& extok = mMacroExpansion[mMacroExpansionIdx++];

		if (mMacroExpansionIdx >= (int)mMacroExpansion.size())
			mMacroExpansion.clear();

		mLastToken = extok.mToken;
		mpToken = extok.mpToken;
		mTokenLength = extok.mTokenLength;
		mIntVal = extok.mIntVal;

		return mLastToken;
	}

	int tok;
	
	do {
		tok = mLexer.Token();

		if (tok == '\n') {
			mbStartOfLine = true;
			continue;
		}

		if (tok == '#') {
			if (!mbStartOfLine) {
				SetError("Preprocessor statement must be at start of line");
				return 0;
			}

			tok = InternalToken();
			if (tok == '\n') {
				mbStartOfLine = true;
				continue;
			}

			if (tok != kTokenIdent) {
				SetError("Preprocessor statement expected");
				return 0;
			}

			if (mLexer.mTokenLength == 6 && !memcmp(mLexer.mpToken, "define", 6)) {
				if (!ParseDefine())
					return 0;
			} else {
				SetErrorF("Unknown preprocessor directive '#%.*s'", mLexer.mTokenLength, mLexer.mpToken);
				return 0;
			}

			tok = ' ';	// kluge
			continue;
		}

		mbStartOfLine = false;

		if (tok == kTokenIdent) {
			VDStringA tokstr;
			tokstr.assign(mLexer.mpToken, mLexer.mpToken + mLexer.mTokenLength);

			Macros::const_iterator it(mMacros.find(tokstr));
			if (it != mMacros.end()) {
				const MacroDefinition& def = it->second;

				VDASSERT(mMacroExpansion.empty());

				mMacroExpansionIdx = 0;

				for(MacroDefinition::Tokens::const_iterator it(def.mTokens.begin()), itEnd(def.mTokens.end()); it != itEnd; ++it) {
					const StoredToken& stok = *it;

					if (stok.mToken == ' ')
						continue;

					ExpandedToken xtok;
					xtok.mToken = stok.mToken;
					xtok.mpToken = def.mTokenHeap.data() + stok.mTokenOffset;
					xtok.mTokenLength = stok.mTokenLength;
					xtok.mIntVal = stok.mIntVal;
					mMacroExpansion.push_back(xtok);
				}

				return Token();
			}
		}

	} while(tok == ' ' || tok == '\n');

	mpToken = mLexer.mpToken;
	mTokenLength = mLexer.mTokenLength;
	mIntVal = mLexer.mIntVal;
	mLastToken = tok;

	return tok;
}

void VDShaderPreprocessor::Push() {
	mbTokenQueued = true;
}

bool VDShaderPreprocessor::SetError(const char *s) {
	return mLexer.SetError(s);
}

bool VDShaderPreprocessor::SetErrorF(const char *format, ...) {
	va_list val;

	va_start(val, format);
	mLexer.SetErrorVF(format, val);
	va_end(val);

	return false;
}

bool VDShaderPreprocessor::SetErrorVF(const char *format, va_list val) {
	mLexer.SetErrorVF(format, val);
	return false;
}

void VDShaderPreprocessor::ClearErrors() {
	mLexer.ClearErrors();
}

bool VDShaderPreprocessor::ParseDefine() {
	int tok = InternalToken();

	if (tok != kTokenIdent) {
		SetError("Expected macro name");
		return 0;
	}

	VDStringA name;
	name.assign(mLexer.mpToken, mLexer.mpToken + mLexer.mTokenLength);

	tok = mLexer.Token();
	if (tok == '(') {
		SetError("Macros with arguments are not supported yet");
		return 0;
	}

	if (tok != ' ')
		mLexer.Push();

	// collect tokens until EOL
	MacroDefinition def;

	for(;;) {
		tok = mLexer.Token();
		if (!tok) {
			SetError("EOF reached in the middle of a macro definition");
			return 0;
		}

		if (tok == '\n')
			break;

		StoredToken token = {0};
		token.mToken = tok;
		token.mTokenOffset = def.mTokenHeap.size();
		token.mTokenLength = mLexer.mTokenLength;
		def.mTokenHeap.insert(def.mTokenHeap.end(), mLexer.mpToken, mLexer.mpToken + mLexer.mTokenLength);

		if (tok == kTokenInt)
			token.mIntVal = mLexer.mIntVal;
		else if (tok == kTokenFloat)
			token.mFloatVal = mLexer.mFloatVal;

		def.mTokens.push_back(token);
	}

	mMacros[name] = def;

	mbStartOfLine = true;
	return true;
}

int VDShaderPreprocessor::InternalToken() {
	for(;;) {
		int tok = mLexer.Token();

		if (tok != ' ' && tok != '\n')
			return tok;
	}
}
