#include "stdafx.h"
#include <vector>
#include <float.h>
#include "parser.h"
#include "log.h"
#include "disasm.h"
#include "scalar.h"

void VDShaderTransformFunctionSimplify(VDShaderFunction& func);

///////////////////////////////////////////////////////////////////////////////
void Disassemble(IVDShaderLogOutput &out, const VDShaderInterpreterTransformedInsn& insn) {
	switch(insn.op) {
		case kVDShaderInterpOp_end:
			out.WriteLogOutput("end");
			break;
		case kVDShaderInterpOp_mov:
			out.WriteLogOutput("mov");
			break;
		case kVDShaderInterpOp_add:
			out.WriteLogOutput("add");
			break;
		case kVDShaderInterpOp_sub:
			out.WriteLogOutput("sub");
			break;
		case kVDShaderInterpOp_mul:
			out.WriteLogOutput("mul");
			break;
		case kVDShaderInterpOp_rsq:
			out.WriteLogOutput("rsq");
			break;
		case kVDShaderInterpOp_rcp:
			out.WriteLogOutput("rcp");
			break;
		case kVDShaderInterpOp_frc:
			out.WriteLogOutput("frc");
			break;
		case kVDShaderInterpOp_neg:
			out.WriteLogOutput("neg");
			break;
		case kVDShaderInterpOp_min:
			out.WriteLogOutput("min");
			break;
		case kVDShaderInterpOp_max:
			out.WriteLogOutput("max");
			break;
		case kVDShaderInterpOp_abs:
			out.WriteLogOutput("abs");
			break;
		case kVDShaderInterpOp_log:
			out.WriteLogOutput("log");
			break;
		case kVDShaderInterpOp_exp:
			out.WriteLogOutput("exp");
			break;
		case kVDShaderInterpOp_cmp:
			out.WriteLogOutput("cmp");
			break;
		case kVDShaderInterpOp_texld:
			out.WriteLogOutput("texld");
			break;
		case kVDShaderInterpOp_unpackr:
			out.WriteLogOutput("unpackr");
			break;
		case kVDShaderInterpOp_unpackg:
			out.WriteLogOutput("unpackg");
			break;
		case kVDShaderInterpOp_unpackb:
			out.WriteLogOutput("unpackb");
			break;
		case kVDShaderInterpOp_unpacka:
			out.WriteLogOutput("unpacka");
			break;
		case kVDShaderInterpOp_beginloop:
			out.WriteLogOutput("beginloop");
			break;
		default:
			VDASSERT(false);
	}

	if (insn.dst) {
		switch(insn.dst & kVDShaderRegTypeMask) {
			case kVDShaderRegR0:
				VDShaderWriteLogOutputF(out, " r%d", insn.dst - kVDShaderRegR0);
				break;

			case kVDShaderRegC0:
				VDShaderWriteLogOutputF(out, " c%d", insn.dst - kVDShaderRegC0);
				break;

			case kVDShaderRegO0:
				VDShaderWriteLogOutputF(out, " o%d", insn.dst - kVDShaderRegO0);
				break;

			default:
				out.WriteLogOutput(" <invalid>");
				break;
		}
	}

	for(int i=0; i<3; ++i) {
		const char *comma = insn.dst || i ? "," : "";
		int src = insn.src[i];

		if (!src)
			continue;

		switch(src & kVDShaderRegTypeMask) {
			case kVDShaderRegR0:
				VDShaderWriteLogOutputF(out, "%s r%d", comma, src - kVDShaderRegR0);
				break;

			case kVDShaderRegC0:
				VDShaderWriteLogOutputF(out, "%s c%d", comma, src - kVDShaderRegC0);
				break;

			case kVDShaderRegO0:
				VDShaderWriteLogOutputF(out, "%s o%d", comma, src - kVDShaderRegO0);
				break;

			case kVDShaderRegV0:
				VDShaderWriteLogOutputF(out, "%s v%d", comma, src - kVDShaderRegV0);
				break;

			case kVDShaderRegS0:
				VDShaderWriteLogOutputF(out, "%s s%d", comma, src - kVDShaderRegS0);
				break;

			default:
				VDShaderWriteLogOutputF(out, "%s <invalid>", comma);
				break;
		}
	}

	out.WriteLogOutput("\n");
}

void Disassemble(IVDShaderLogOutput& out, const VDShaderScalarForm& shader) {
	VDASSERT(shader.mConstants.size() == shader.mConstantBindings.size());

	const int numConstants = shader.mConstants.size();

	for(int i=0; i<numConstants; ++i) {
		int srcIdx = shader.mConstantBindings[i];

		if (srcIdx >= 0)
			VDShaderWriteLogOutputF(out, "link c%d, src.c%d.%c\n", i, srcIdx >> 2, "xyzw"[srcIdx & 3]);
	}

	for(VDShaderScalarForm::ConstantUploads::const_iterator it(shader.mConstantUploads.begin()), itEnd(shader.mConstantUploads.end()); it != itEnd; ++it) {
		const VDShaderScalarForm::ConstantUploadEntry& conUp = *it;
		int dstIdx = conUp.mDstIdx;
		int srcIdx = shader.mConstantBindings[dstIdx];

		VDASSERT(srcIdx >= 0);

		VDShaderWriteLogOutputF(out, "upload c%d, %ld, %g\n", dstIdx, (long)conUp.mSrcOffset, shader.mConstants[dstIdx]);
	}

	for(int i=0; i<numConstants; ++i) {
		if (shader.mConstantBindings[i] < 0)
			VDShaderWriteLogOutputF(out, "def c%d, %g\n", i, shader.mConstants[i]);
	}

	for(vdfastvector<VDShaderInterpreterTransformedInsn>::const_iterator it(shader.mInsns.begin()), itEnd(shader.mInsns.end()); it!=itEnd; ++it) {
		const VDShaderInterpreterTransformedInsn& insn = *it;

		Disassemble(out, insn);
	}
}

///////////////////////////////////////////////////////////////////////////////

bool VDShaderScalarConvertToSSAForm(VDShaderScalarForm& shader, IVDShaderLogOutput *out) {
	vdfastvector<int> tempRemap(shader.mTempCount, 0);
	vdfastvector<int> tempRemapInput(shader.mTempCount, 0);
	vdfastvector<bool> tempLoopInputs(shader.mTempCount, false);
	vdfastvector<bool> tempLoopWritten(shader.mTempCount, false);
	vdfastvector<int> tempBackMap;

	int tempNext = kVDShaderRegR0;

	VDShaderScalarForm::Insns::iterator it(shader.mInsns.begin()), itEnd(shader.mInsns.end());
	for(; it != itEnd; ++it)
	{
		VDShaderInterpreterTransformedInsn& insn = *it;

		if (insn.op == kVDShaderInterpOp_beginloop) {
			std::fill(tempLoopInputs.begin(), tempLoopInputs.end(), false);
			std::fill(tempLoopWritten.begin(), tempLoopWritten.end(), false);
			continue;
		}

		for(int i=0; i<3; ++i) {
			int& src = insn.src[i];

			if ((src & kVDShaderRegTypeMask) == kVDShaderRegR0) {
				int idx = src - kVDShaderRegR0;

				if (idx >= shader.mTempCount) {
					if (out)
						VDShaderWriteLogOutputF(*out, "Internal compiler error [SSA]: Invalid source temp register index.");

					return false;
				}

				if (tempRemap[idx] < 0) {
					if (out)
						VDShaderWriteLogOutputF(*out, "Internal compiler error [SSA]: Reading uninitialized temp register.");

					return false;
				}

				src = tempRemap[idx];

				if (!tempLoopWritten[idx] && !tempLoopInputs[idx]) {
					tempLoopInputs[idx] = true;
					tempRemapInput[idx] = src;
				}
			}
		}

		if ((insn.dst & kVDShaderRegTypeMask) == kVDShaderRegR0) {
			int dstIdx = insn.dst - kVDShaderRegR0;

			if (dstIdx >= shader.mTempCount) {
				if (out)
					VDShaderWriteLogOutputF(*out, "Internal compiler error [SSA]: Invalid source temp register index.");

				return false;
			}

			int newDst = tempNext++;
			tempBackMap.push_back(dstIdx);

			tempRemap[dstIdx] = newDst;
			tempLoopWritten[dstIdx] = true;

			insn.dst = newDst;
		}
	}

	// back up and remap loop carried dependencies
	VDShaderScalarForm::Insns::iterator it2(shader.mInsns.end()), it2Begin(shader.mInsns.begin());
	while(it2 != it2Begin) {
		VDShaderInterpreterTransformedInsn& insn = *--it2;

		if (insn.op == kVDShaderInterpOp_beginloop)
			break;

		if ((insn.dst & kVDShaderRegTypeMask) == kVDShaderRegR0) {
			int dstIdx = tempBackMap[insn.dst - kVDShaderRegR0];

			if (tempLoopWritten[dstIdx] && tempLoopInputs[dstIdx])
				insn.dst = tempRemapInput[dstIdx];
		}
	}

	shader.mTempCount = tempNext - kVDShaderRegR0;

	return true;
}

void VDShaderScalarBakeSamplers(VDShaderScalarForm& shader, const VDShaderInterpreterSampler *const *pSamplers, int numSamplers) {
	for(VDShaderScalarForm::Insns::iterator it(shader.mInsns.begin()), itEnd(shader.mInsns.end()); it != itEnd; ++it) {
		VDShaderInterpreterTransformedInsn& insn = *it;

		if (insn.op != kVDShaderInterpOp_texld)
			continue;

		// determine sampler index
		VDASSERT((insn.src[2] & kVDShaderRegTypeMask) == kVDShaderRegS0);
		int sampIdx = insn.src[2] & kVDShaderRegIndexMask;

		// check if sampler is valid
		if (sampIdx >= numSamplers || !pSamplers[sampIdx]) {
			// invalid sampler... load zero
			insn.op = kVDShaderInterpOp_mov;
			insn.src[0] = kVDShaderRegC0 + shader.mConstants.size();
			insn.src[1] = 0;
			insn.src[2] = 0;
			shader.mConstants.push_back(0);

			shader.mConstantBindings.push_back(-1);
			continue;
		}

		// begin transforming texld
		int ureg = insn.src[0];
		int vreg = insn.src[1];

		// if we're wrapping U or V, do a frac on the coord
		const VDShaderInterpreterSampler& samp = *pSamplers[sampIdx];
		VDShaderInterpreterTransformedInsn tmp;

		if (!samp.mbClampU) {
			tmp.op = kVDShaderInterpOp_frc;
			tmp.dst = kVDShaderRegR0 + (shader.mTempCount++);
			tmp.src[0] = ureg;
			tmp.src[1] = 0;
			tmp.src[2] = 0;
			it = shader.mInsns.insert(it, tmp) + 1;
			ureg = tmp.dst;
		}

		if (!samp.mbClampV) {
			tmp.op = kVDShaderInterpOp_frc;
			tmp.dst = kVDShaderRegR0 + (shader.mTempCount++);
			tmp.src[0] = vreg;
			tmp.src[1] = 0;
			tmp.src[2] = 0;
			it = shader.mInsns.insert(it, tmp) + 1;
			vreg = tmp.dst;
		}

		// multiply by width and height
		tmp.op = kVDShaderInterpOp_mul;
		tmp.dst = kVDShaderRegR0 + (shader.mTempCount++);
		tmp.src[0] = ureg;
		tmp.src[1] = kVDShaderRegC0 + shader.mConstants.size();
		tmp.src[2] = 0;
		it = shader.mInsns.insert(it, tmp) + 1;
		ureg = tmp.dst;
		shader.mConstants.push_back(samp.mFWidth);
		shader.mConstantBindings.push_back(-1);

		tmp.op = kVDShaderInterpOp_mul;
		tmp.dst = kVDShaderRegR0 + (shader.mTempCount++);
		tmp.src[0] = vreg;
		tmp.src[1] = kVDShaderRegC0 + shader.mConstants.size();
		tmp.src[2] = 0;
		it = shader.mInsns.insert(it, tmp) + 1;
		vreg = tmp.dst;
		shader.mConstants.push_back(samp.mFHeight);
		shader.mConstantBindings.push_back(-1);

		// clamp if we're clamping
		if (samp.mbClampU) {
			tmp.op = kVDShaderInterpOp_max;
			tmp.dst = kVDShaderRegR0 + (shader.mTempCount++);
			tmp.src[0] = ureg;
			tmp.src[1] = kVDShaderRegC0 + shader.mConstants.size();
			tmp.src[2] = 0;
			it = shader.mInsns.insert(it, tmp) + 1;
			ureg = tmp.dst;
			shader.mConstants.push_back(0);
			shader.mConstantBindings.push_back(-1);

			tmp.op = kVDShaderInterpOp_min;
			tmp.dst = kVDShaderRegR0 + (shader.mTempCount++);
			tmp.src[0] = ureg;
			tmp.src[1] = kVDShaderRegC0 + shader.mConstants.size();
			tmp.src[2] = 0;
			it = shader.mInsns.insert(it, tmp) + 1;
			ureg = tmp.dst;
			shader.mConstants.push_back((float)samp.mULimit);
			shader.mConstantBindings.push_back(-1);
		}

		if (samp.mbClampV) {
			tmp.op = kVDShaderInterpOp_max;
			tmp.dst = kVDShaderRegR0 + (shader.mTempCount++);
			tmp.src[0] = vreg;
			tmp.src[1] = kVDShaderRegC0 + shader.mConstants.size();
			tmp.src[2] = 0;
			it = shader.mInsns.insert(it, tmp) + 1;
			vreg = tmp.dst;
			shader.mConstants.push_back(0);
			shader.mConstantBindings.push_back(-1);

			tmp.op = kVDShaderInterpOp_min;
			tmp.dst = kVDShaderRegR0 + (shader.mTempCount++);
			tmp.src[0] = vreg;
			tmp.src[1] = kVDShaderRegC0 + shader.mConstants.size();
			tmp.src[2] = 0;
			it = shader.mInsns.insert(it, tmp) + 1;
			vreg = tmp.dst;
			shader.mConstants.push_back((float)samp.mVLimit);
			shader.mConstantBindings.push_back(-1);
		}

		// fixup texld and continue
		it->src[0] = ureg;
		it->src[1] = vreg;

		itEnd = shader.mInsns.end();
	}
}

void VDShaderScalarBakeInterpolatorStepping(VDShaderScalarForm& shader, const VDShaderInterpreterInterpolator *interps, uint32 interpCount, float interpStepFactor) {
	int interpStepBase = shader.mConstants.size() + kVDShaderRegC0;

	for(uint32 i=0; i<interpCount; ++i) {
		const VDShaderInterpreterInterpolator& interp = interps[i];

		shader.mConstants.push_back(interp.dudx[0] * interpStepFactor);
		shader.mConstants.push_back(interp.dudx[1] * interpStepFactor);
		shader.mConstants.push_back(interp.dudx[2] * interpStepFactor);
		shader.mConstants.push_back(interp.dudx[3] * interpStepFactor);
		shader.mConstantBindings.push_back(-1);
		shader.mConstantBindings.push_back(-1);
		shader.mConstantBindings.push_back(-1);
		shader.mConstantBindings.push_back(-1);
	}

	for(VDShaderScalarForm::Insns::iterator it(shader.mInsns.begin()), itEnd(shader.mInsns.end()); it != itEnd; ++it) {
		VDShaderInterpreterTransformedInsn& insn = *it;

		for(int i=0; i<3; ++i) {
			int src = insn.src[i];

			if ((src & kVDShaderRegTypeMask) != kVDShaderRegV0)
				continue;

			if (!(src & 1))
				continue;

			insn.src[i] = interpStepBase + ((src & kVDShaderRegIndexMask) >> 1);
		}
	}
}

bool VDShaderScalarInsnFoldConstants(VDShaderScalarForm& shader, VDShaderInterpreterTransformedInsn& insn) {
	if (!insn.dst)
		return false;

	const float *pc[3] = { NULL };
	for(int i=0; i<3; ++i) {
		if (insn.src[i]) {
			if ((insn.src[i] & kVDShaderRegTypeMask) == kVDShaderRegC0) {
				int idx = insn.src[i] - kVDShaderRegC0;
				if (shader.mConstantBindings[idx] < 0)
					pc[i] = &shader.mConstants[idx];
			}
		}
	}

	const float *pc1 = pc[0];
	const float *pc2 = pc[1];
	const float *pc3 = pc[2];

	float result;
	switch(insn.op) {
		case kVDShaderInterpOp_add:
			if (!pc1 || !pc2) {
				if (pc1 && !pc2) {
					std::swap(pc1, pc2);
					std::swap(insn.src[0], insn.src[1]);
				}

				if (!pc1 && pc2 && *pc2 == 0.0f) {
					insn.op = kVDShaderInterpOp_mov;
					insn.src[1] = 0;
					return true;
				}

				return false;
			}

			result = *pc1 + *pc2;
			break;

		case kVDShaderInterpOp_sub:
			if (!pc1 || !pc2)
				return false;
			result = *pc1 - *pc2;
			break;

		case kVDShaderInterpOp_mul:
			if (!pc1 || !pc2) {
				if (pc1 && !pc2) {
					std::swap(pc1, pc2);
					std::swap(insn.src[0], insn.src[1]);
				}

				if (!pc1 && pc2) {
					if (*pc2 == 1.0f) {
						insn.op = kVDShaderInterpOp_mov;
						insn.src[1] = 0;
						return true;
					} else if (*pc2 == 0.0f) {
						insn.op = kVDShaderInterpOp_mov;
						insn.src[0] = shader.mConstants.size() + kVDShaderRegC0;
						insn.src[1] = 0;
						shader.mConstants.push_back(0.0f);
						shader.mConstantBindings.push_back(-1);
						return true;
					}
				}

				return false;
			}

			result = *pc1 * *pc2;
			break;
		case kVDShaderInterpOp_rsq:
			if (!pc1)
				return false;
			result = 1.0f / sqrtf(fabs(*pc1));
			break;
		case kVDShaderInterpOp_rcp:
			if (!pc1)
				return false;
			result = 1.0f / *pc1;
			break;
		case kVDShaderInterpOp_frc:
			if (!pc1)
				return false;
			result = *pc1 - floorf(*pc1);
			break;
		case kVDShaderInterpOp_neg:
			if (!pc1)
				return false;
			result = -*pc1;
			break;
		case kVDShaderInterpOp_min:
			if (!pc1 || !pc2)
				return false;
			result = (*pc1 >= *pc2) ? *pc2 : *pc1;
			break;
		case kVDShaderInterpOp_max:
			if (!pc1 || !pc2)
				return false;
			result = (*pc1 <= *pc2) ? *pc2 : *pc1;
			break;
		case kVDShaderInterpOp_abs:
			if (!pc1)
				return false;
			result = fabsf(*pc1);
			break;
		case kVDShaderInterpOp_log:
			if (!pc1)
				return false;
			{
				float f = fabsf(*pc1);
				result = (f != 0.0f) ? logf(f) * 1.4426950408889634073599246810019f : -FLT_MAX;
			}
			break;
		case kVDShaderInterpOp_exp:
			if (!pc1)
				return false;

			// The intrinsic version of exp() does not handle -Inf correctly, which we need
			// for pow(x, y) = exp2(y * log2(x)) to work correctly.
			if (*pc1 < -FLT_MAX)
				result = 0;
			else
				result = expf(*pc1 * 0.69314718055994530941723212145818f);
			break;
		case kVDShaderInterpOp_cmp:
			if (!pc1)
				return false;

			if (pc2 && pc3) {
				result = (*pc1 >= 0.0f) ? *pc2 : *pc3;
			} else {
				insn.op = kVDShaderInterpOp_mov;
				insn.src[0] = (*pc1 >= 0.0f) ? insn.src[1] : insn.src[2];
				return true;
			}
			break;
		case kVDShaderInterpOp_unpackb:
			if (!pc1)
				return false;
			result = ((const uint8 *)pc1)[0] / 255.0f;
			break;
		case kVDShaderInterpOp_unpackg:
			if (!pc1)
				return false;
			result = ((const uint8 *)pc1)[1] / 255.0f;
			break;
		case kVDShaderInterpOp_unpackr:
			if (!pc1)
				return false;
			result = ((const uint8 *)pc1)[2] / 255.0f;
			break;
		case kVDShaderInterpOp_unpacka:
			if (!pc1)
				return false;
			result = ((const uint8 *)pc1)[3] / 255.0f;
			break;
		default:
			return false;
	}

	insn.op = kVDShaderInterpOp_mov;
	insn.src[0] = kVDShaderRegC0 + shader.mConstants.size();
	insn.src[1] = 0;
	shader.mConstants.push_back(result);
	shader.mConstantBindings.push_back(-1);
	return true;
}

bool VDShaderScalarFoldConstants(VDShaderScalarForm& shader) {
	bool renamed = false;

	for(VDShaderScalarForm::Insns::iterator it(shader.mInsns.begin()), itEnd(shader.mInsns.end()); it != itEnd; ++it) {
		VDShaderInterpreterTransformedInsn& insn = *it;

		while(VDShaderScalarInsnFoldConstants(shader, insn))
			renamed = true;
	}

	return renamed;
}

void VDShaderScalarApplyCopyPropagation(VDShaderScalarForm& shader) {
	vdfastvector<int> tempRemap(shader.mTempCount);
	vdfastvector<bool> tempWrittenInLoop(shader.mTempCount, false);

	for(int i=0; i<shader.mTempCount; ++i)
		tempRemap[i] = kVDShaderRegR0 + i;

	// Pre-iterate through the shader and find all registers that are
	// written in the loop.
	bool inLoop = false;
	for(VDShaderScalarForm::Insns::iterator it(shader.mInsns.begin()), itEnd(shader.mInsns.end()); it != itEnd; ++it) {
		VDShaderInterpreterTransformedInsn& insn = *it;

		if (insn.op == kVDShaderInterpOp_beginloop)
			inLoop = true;

		if (inLoop && (insn.dst & kVDShaderRegTypeMask) == kVDShaderRegR0)
			tempWrittenInLoop[insn.dst - kVDShaderRegR0] = true;
	}

	// Apply copy propagation.
	for(VDShaderScalarForm::Insns::iterator it(shader.mInsns.begin()), itEnd(shader.mInsns.end()); it != itEnd; ++it) {
		VDShaderInterpreterTransformedInsn& insn = *it;

		if (insn.op == kVDShaderInterpOp_beginloop) {
			// At the beginning of the loop, any register that we attempt to read
			// that was written at the end of the loop is a loop carried dependency
			// that we cannot remap, as it is written by both the prologue and in
			// the loop.
			for(int i=0; i<shader.mTempCount; ++i) {
				if (tempWrittenInLoop[i])
					tempRemap[i] = kVDShaderRegR0 + i;
			}
		}

		// remap source registers
		for(int i=0; i<3; ++i) {
			if ((insn.src[i] & kVDShaderRegTypeMask) == kVDShaderRegR0)
				insn.src[i] = tempRemap[insn.src[i] - kVDShaderRegR0];
		}

		// map dest register if this is a mov insn
		if (insn.op == kVDShaderInterpOp_mov) {
			if ((insn.dst & kVDShaderRegTypeMask) == kVDShaderRegR0)
				tempRemap[insn.dst - kVDShaderRegR0] = insn.src[0];

			// check for special case of a useless mov instruction
			if (insn.dst == insn.src[0])
				insn.dst = 0;
		}
	}
}

///////////////////////////////////////////////////////////////////////////////

void VDShaderScalarApplyDeadStoreElimination(VDShaderScalarForm& shader) {
	int numInsns = shader.mInsns.size();

	vdfastvector<int> regWriteHead(shader.mTempCount, -1);
	vdfastvector<int> regWriteNext(numInsns, -1);
	vdfastvector<int> insnQueue;
	vdfastvector<bool> markedInsns(numInsns, false);

	// Iterate forwards through the instruction stream construct a list of
	// (a) instructions that write each temporary, and (b) instructions that
	// write outputs.
	for(int i=0; i<numInsns; ++i) {
		VDShaderInterpreterTransformedInsn& insn = shader.mInsns[i];

		if (!insn.dst) {
			// special case: ignore mov instructions that go nowhere
			if (insn.op != kVDShaderInterpOp_mov) {
				insnQueue.push_back(i);
				markedInsns[i] = true;
			}
			continue;
		}

		int dstidx = insn.dst & kVDShaderRegIndexMask;
		switch(insn.dst & kVDShaderRegTypeMask) {
			case kVDShaderRegR0:
				regWriteNext[i] = regWriteHead[dstidx];
				regWriteHead[dstidx] = i;
				break;
			case kVDShaderRegO0:
				insnQueue.push_back(i);
				markedInsns[i] = true;
				break;
		}
	}

	// Follow reference graph and mark all instructions that 
	// Iterate backwards through the instruction stream and eliminate all
	// writes that aren't referenced.
	while(!insnQueue.empty()) {
		int i = insnQueue.back();
		insnQueue.pop_back();

		VDShaderInterpreterTransformedInsn& insn = shader.mInsns[i];

		// queue instructions that write sources
		for(int j=0; j<3; ++j) {
			int r = insn.src[j];

			if ((r & kVDShaderRegTypeMask) == kVDShaderRegR0) {
				int ridx = r - kVDShaderRegR0;

				for(int k = regWriteHead[ridx]; k >= 0; k = regWriteNext[k]) {
					if (k != i && !markedInsns[k]) {
						markedInsns[k] = true;
						insnQueue.push_back(k);
					}
				}
			}
		}
	}

	// Delete all unmarked instructions.
	VDShaderScalarForm::Insns::iterator itDst(shader.mInsns.begin());
	for(int i=0; i<numInsns; ++i) {
		const VDShaderInterpreterTransformedInsn& insn = shader.mInsns[i];

		if (markedInsns[i]) {
			if (&*itDst != &insn)
				*itDst = insn;

			++itDst;
		}
	}

	shader.mInsns.erase(itDst, shader.mInsns.end());
}

///////////////////////////////////////////////////////////////////////////////

void VDShaderScalarHoistLoopInvariants(VDShaderScalarForm& shader) {
	vdfastvector<int> lastTempWrite(shader.mTempCount, -1);
	vdfastvector<VDShaderInterpreterTransformedInsn> newInsns;
	vdfastvector<VDShaderInterpreterTransformedInsn> newInsns2;

	// find last write for every temp register
	int insnIdx = 0;
	for(VDShaderScalarForm::Insns::const_iterator it(shader.mInsns.begin()), itEnd(shader.mInsns.end()); it != itEnd; ++it, ++insnIdx) {
		const VDShaderInterpreterTransformedInsn& insn = *it;

		if ((insn.dst & kVDShaderRegTypeMask) == kVDShaderRegR0) {
			int dstIdx = insn.dst - kVDShaderRegR0;

			lastTempWrite[dstIdx] = insnIdx;
		}
	}

	// resort loop invariant insns to the beginning
	insnIdx = 0;
	for(VDShaderScalarForm::Insns::const_iterator it(shader.mInsns.begin()), itEnd(shader.mInsns.end()); it != itEnd; ++it, ++insnIdx) {
		const VDShaderInterpreterTransformedInsn& insn = *it;

		if ((insn.dst & kVDShaderRegTypeMask) != kVDShaderRegO0 && insn.op != kVDShaderInterpOp_beginloop) {
			int refLimit = -1;

			for(int i=0; i<3; ++i) {
				int src = insn.src[i];

				if ((src & kVDShaderRegTypeMask) == kVDShaderRegR0) {
					int srcWriteIdx = lastTempWrite[src & kVDShaderRegIndexMask];

					if (refLimit < srcWriteIdx)
						refLimit = srcWriteIdx;
				}
			}

			if (refLimit < insnIdx) {
				newInsns.push_back(insn);
				continue;
			}
		}

		newInsns2.push_back(insn);
		if ((insn.dst & kVDShaderRegTypeMask) == kVDShaderRegR0) {
			int dstIdx = insn.dst - kVDShaderRegR0;

			lastTempWrite[dstIdx] = INT_MAX;
		}
	}

	shader.mInsns = newInsns;
	shader.mInsns.insert(shader.mInsns.end(), newInsns2.begin(), newInsns2.end());
}

///////////////////////////////////////////////////////////////////////////////

void VDShaderScalarPackConstants(VDShaderScalarForm& shader) {
	vdfastvector<int> constantMap(shader.mConstants.size(), -1);
	vdfastvector<float> newConstants;
	vdfastvector<int> newBindings;

	for(VDShaderScalarForm::Insns::iterator it(shader.mInsns.begin()), itEnd(shader.mInsns.end()); it != itEnd; ++it) {
		VDShaderInterpreterTransformedInsn& insn = *it;

		for(int i=0; i<3; ++i) {
			int r = insn.src[i];

			if ((r & kVDShaderRegTypeMask) == kVDShaderRegC0) {
				int idx = r - kVDShaderRegC0;
				int newIdx = constantMap[idx];

				if (newIdx < 0) {
					// scan for constant
					uint32 constBits = (uint32&)shader.mConstants[idx];
					bool found = false;

					if (shader.mConstantBindings[idx] < 0) {
						int idx2 = 0;
						for(vdfastvector<float>::const_iterator it2(newConstants.begin()), it2End(newConstants.end()); it2 != it2End; ++it2, ++idx2) {
							if (newBindings[idx2] < 0 && (uint32&)*it2 == constBits) {
								newIdx = it2 - newConstants.begin();
								found = true;
								break;
							}
						}
					}

					if (!found) {
						newIdx = newConstants.size();
						newConstants.push_back();
						(uint32&)newConstants.back() = constBits;

						newBindings.push_back(shader.mConstantBindings[idx]);
					}

					constantMap[idx] = newIdx;
				}

				insn.src[i] = newIdx + kVDShaderRegC0;
			}
		}
	}

	shader.mConstants.swap(newConstants);
	shader.mConstantBindings.swap(newBindings);
}

bool VDShaderScalarPackTemps(VDShaderScalarForm& shader, IVDShaderLogOutput *out) {
	const int numInsns = shader.mInsns.size();

	vdfastvector<int> tempMap(shader.mTempCount, -1);
	vdfastvector<bool> tempWritten(shader.mTempCount, false);
	int newTempCount = 0;

	// Iterate forwards through the instruction stream and mark all temp
	// reads that aren't hit by a write. These are loop carried dependencies
	// and must be considered live at the end of the loop.
	bool inLoop = false;
	for(int i=0; i<numInsns; ++i) {
		VDShaderInterpreterTransformedInsn& insn = shader.mInsns[i];

		if (insn.op == kVDShaderInterpOp_beginloop)
			inLoop = true;
		else if (!inLoop)
			continue;

		if ((insn.dst & kVDShaderRegTypeMask) == kVDShaderRegR0)
			tempWritten[insn.dst - kVDShaderRegR0] = true;

		for(int j=0; j<3; ++j) {
			if ((insn.src[j] & kVDShaderRegTypeMask) == kVDShaderRegR0) {
				int idx = insn.src[j] - kVDShaderRegR0;

				if (!tempWritten[idx] && tempMap[idx] < 0)
					tempMap[idx] = newTempCount++;
			}
		}
	}

	vdfastvector<int> freeTemps;

	for(int i=numInsns - 1; i >= 0; --i) {
		VDShaderInterpreterTransformedInsn& insn = shader.mInsns[i];

		// remap and kill destination register
		if ((insn.dst & kVDShaderRegTypeMask) == kVDShaderRegR0) {
			int dstidx = insn.dst - kVDShaderRegR0;
			int dstidx2 = tempMap[dstidx];

			if (dstidx2 < 0) {
				if (out) {
					out->WriteLogOutput("Internal Compiler Error: Dead store detected during register allocation pass.\n");
					out->WriteLogOutput("  Bad instruction:");
					Disassemble(*out, insn);
				}
				return false;
			}

			insn.dst = kVDShaderRegR0 + dstidx2;

			// We shouldn't kill the destination register if it's the same as one of the sources. Not
			// only is it unnecessary, but it can also break loop carried dependencies.
			if (insn.dst != insn.src[0] && insn.dst != insn.src[1] && insn.dst != insn.src[2]) {
				freeTemps.push_back(dstidx2);
				tempMap[dstidx] = -1;
			}
		}

		// allocate and remap source registers
		for(int j=0; j<3; ++j) {
			if ((insn.src[j] & kVDShaderRegTypeMask) == kVDShaderRegR0) {
				int srcIdx = insn.src[j] - kVDShaderRegR0;
				int srcIdx2 = tempMap[srcIdx];

				if (srcIdx2 < 0) {
					if (freeTemps.empty())
						freeTemps.push_back(newTempCount++);

					srcIdx2 = freeTemps.back();
					freeTemps.pop_back();

					tempMap[srcIdx] = srcIdx2;
				}

				insn.src[j] = kVDShaderRegR0 + srcIdx2;
			}
		}
	}

	return true;
}

bool VDShaderScalarBuildConstantUploadList(VDShaderScalarForm& shader, const VDShaderInterpConstant *pConstants, int numConstants, IVDShaderLogOutput *out) {
	vdfastvector<ptrdiff_t> srcOffsets;
	vdfastvector<bool> srcOffsetAssigned;
	int regLimit = 0;

	for(int i=0; i<numConstants; ++i) {
		const VDShaderInterpConstant& c = pConstants[i];
		int ci = c.mConstantIndex * 4;

		if (ci >= regLimit) {
			regLimit = ci + 4;

			srcOffsets.resize(regLimit, 0);
			srcOffsetAssigned.resize(regLimit, false);
		}

		for(int j=0; j<4; ++j) {
			srcOffsets[ci + j] = c.mOffset + sizeof(float)*j;
			srcOffsetAssigned[ci + j] = true;
		}
	}

	int numConstantBindings = shader.mConstantBindings.size();
	VDASSERT(numConstantBindings == shader.mConstants.size());

	shader.mConstantUploads.clear();
	for(int i=0; i<numConstantBindings; ++i) {
		const int regIdx = shader.mConstantBindings[i];

		if ((unsigned)regIdx >= (unsigned)regLimit)
			continue;

		if (srcOffsetAssigned[regIdx]) {
			VDShaderScalarForm::ConstantUploadEntry cent;
			cent.mDstIdx = i;
			cent.mSrcOffset = srcOffsets[regIdx];
			shader.mConstantUploads.push_back(cent);
		}
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////

VDShaderScalarForm::VDShaderScalarForm() {
}

VDShaderScalarForm::~VDShaderScalarForm() {
}

void VDShaderScalarForm::Clear() {
	mInsns.clear();
	mConstants.clear();
	mConstantBindings.clear();
	mTempCount = 0;
}

///////////////////////////////////////////////////////////////////////////////

VDShaderScalarTransformer::VDShaderScalarTransformer(VDShaderScalarForm& dst)
	: mShader(dst)
{
}

VDShaderScalarTransformer::~VDShaderScalarTransformer() {
}

bool VDShaderScalarTransformer::LoadShader(const VDShaderFunction& func0, const VDShaderBindingElement *pElements, int numElements, const VDShaderInterpreterSampler *const *pSamplers, int numSamplers, const VDShaderInterpConstant *pConstants, int numConstants, const VDShaderInterpreterInterpolator *interps, uint32 interpCount, float interpStepFactor, IVDShaderLogOutput *out, bool showDisassembly) {
	VDShaderFunction func(func0);

	// transform complex vector instructions to simpler ones
	VDShaderTransformFunctionSimplify(func);

	if (out && showDisassembly)
		VDShaderDisassembleFunction(*out, func, NULL);

	// load constants
	int numSrcConstants = func.mConstants.size() * 4;
	const float *srcConstants = (const float *)func.mConstants.data();
	mShader.mConstants.assign(srcConstants, srcConstants + numSrcConstants);

	const int cZero = mShader.mConstants.size() + kVDShaderRegC0;
	const int cOne = cZero + 1;
	mShader.mConstants.push_back(0.0f);
	mShader.mConstants.push_back(1.0f);

	mShader.mConstantBindings.resize(numSrcConstants + 2, -1);

	for(VDShaderFunction::ConstantBindings::const_iterator it(func.mConstantBindings.begin()), itEnd(func.mConstantBindings.end()); it != itEnd; ++it) {
		const VDShaderConstantBinding& binding = *it;
		
		int srcRegIdx = (binding.mRegister & kVDShaderRegIndexMask)*4;

		int size = 4*binding.mSize;
		for(int j=0; j<size; ++j)
			mShader.mConstantBindings[srcRegIdx + j] = srcRegIdx + j;
	}

	// transform instructions
	int nextDstReg = kVDShaderRegR0 + func.mRegisterCount*4;

	int t0_x = nextDstReg + 0;
	int t0_y = nextDstReg + 1;
	nextDstReg += 4;

	mov(t0_x, kVDShaderRegV0);
	mov(t0_y, kVDShaderRegV0 + 2);

	beginloop();

	for(VDShaderFunction::Insns::const_iterator it(func.mInsns.begin()), itEnd(func.mInsns.end()); it!=itEnd; ++it) {
		const VDShaderInsn& insn = *it;

		struct DecodedInsn {
			int op;
			int dst[4];
			int src[16];
		};

		DecodedInsn dinsn;

		bool writeConflict = false;

		for(int i=0; i<4; ++i) {
			const VDShaderInsnSource& src = insn.mSources[i];

			if (!src.mRegister)
				continue;

			int srcRegIndex = src.mRegister & kVDShaderRegIndexMask;
			int srcBase = i*4;

			switch(src.mRegister & kVDShaderRegTypeMask) {
				case kVDShaderRegR0:
					dinsn.src[srcBase+0] = kVDShaderRegR0 + srcRegIndex*4 + ((src.mSwizzle >> 0) & 3);
					dinsn.src[srcBase+1] = kVDShaderRegR0 + srcRegIndex*4 + ((src.mSwizzle >> 2) & 3);
					dinsn.src[srcBase+2] = kVDShaderRegR0 + srcRegIndex*4 + ((src.mSwizzle >> 4) & 3);
					dinsn.src[srcBase+3] = kVDShaderRegR0 + srcRegIndex*4 + ((src.mSwizzle >> 6) & 3);
					break;

				case kVDShaderRegC0:
					dinsn.src[srcBase+0] = kVDShaderRegC0 + srcRegIndex*4 + ((src.mSwizzle >> 0) & 3);
					dinsn.src[srcBase+1] = kVDShaderRegC0 + srcRegIndex*4 + ((src.mSwizzle >> 2) & 3);
					dinsn.src[srcBase+2] = kVDShaderRegC0 + srcRegIndex*4 + ((src.mSwizzle >> 4) & 3);
					dinsn.src[srcBase+3] = kVDShaderRegC0 + srcRegIndex*4 + ((src.mSwizzle >> 6) & 3);
					break;

				case kVDShaderRegS0:
					break;

				case kVDShaderRegV0:
					dinsn.src[srcBase+0] = t0_x + srcRegIndex*4 + ((src.mSwizzle >> 0) & 3);
					dinsn.src[srcBase+1] = t0_x + srcRegIndex*4 + ((src.mSwizzle >> 2) & 3);
					dinsn.src[srcBase+2] = t0_x + srcRegIndex*4 + ((src.mSwizzle >> 4) & 3);
					dinsn.src[srcBase+3] = t0_x + srcRegIndex*4 + ((src.mSwizzle >> 6) & 3);
					break;

				default:
					VDASSERT(false);
					break;
			}

			if (src.mModifiers & kVDShaderMod_Negate) {
				for(int j=0; j<4; ++j) {
					neg(nextDstReg, dinsn.src[srcBase+j]);
					dinsn.src[srcBase+j] = nextDstReg++;
				}
			}

			if (src.mRegister == insn.mDestRegister)
				writeConflict = true;
		}

		int dstReg = insn.mDestRegister;
		int dstRegIndex = dstReg & kVDShaderRegIndexMask;
		int writeMask = insn.mWriteMask;

		switch(dstReg & kVDShaderRegTypeMask) {
			case kVDShaderRegO0:
				dstRegIndex = kVDShaderRegO0 + dstRegIndex*4;
				break;

			case kVDShaderRegR0:
				dstRegIndex = kVDShaderRegR0 + dstRegIndex*4;
				break;

			default:
				VDASSERT(false);
				break;
		}

		// force alt write if we have modifiers
		if (insn.mModifiers)
			writeConflict = true;

		int writeBase = nextDstReg;

		for(int i=0; i<4; ++i) {
			if (writeMask & (1 << i)) {
				if (writeConflict)
					dinsn.dst[i] = nextDstReg++;
				else
					dinsn.dst[i] = dstRegIndex + i;
			} else
				dinsn.dst[i] = 0;
		}

		switch(insn.mOp) {
			case kVDShaderOp_mov:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i])
						mov(dinsn.dst[i], dinsn.src[i]);
				}
				break;
			case kVDShaderOp_add:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i])
						add(dinsn.dst[i], dinsn.src[i], dinsn.src[i+4]);
				}
				break;
			case kVDShaderOp_cmp:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i])
						cmp(dinsn.dst[i], dinsn.src[i], dinsn.src[i+4], dinsn.src[i+8]);
				}
				break;
			case kVDShaderOp_sub:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i])
						sub(dinsn.dst[i], dinsn.src[i], dinsn.src[i+4]);
				}
				break;
			case kVDShaderOp_mul:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i])
						mul(dinsn.dst[i], dinsn.src[i], dinsn.src[i+4]);
				}
				break;
			case kVDShaderOp_mad:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i]) {
						int t = nextDstReg++;
						mul(t, dinsn.src[i], dinsn.src[i+4]);
						add(dinsn.dst[i], t, dinsn.src[i+8]);
					}
				}
				break;
			case kVDShaderOp_min:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i])
						min(dinsn.dst[i], dinsn.src[i], dinsn.src[i+4]);
				}
				break;
			case kVDShaderOp_max:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i])
						max(dinsn.dst[i], dinsn.src[i], dinsn.src[i+4]);
				}
				break;
			case kVDShaderOp_dp3:
				{
					int t = nextDstReg++;
					int t2 = nextDstReg++;
					int t3 = nextDstReg++;
					int t4 = nextDstReg++;
					int t5 = 0;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i]) {
							if (!t5) {
								t5 = dinsn.dst[i];

								mul(t, dinsn.src[0], dinsn.src[4]);
								mul(t2, dinsn.src[1], dinsn.src[5]);
								add(t3, t, t2);
								mul(t4, dinsn.src[2], dinsn.src[6]);
								add(t5, t3, t4);
							} else
								mov(dinsn.dst[i], t5);
						}
					}
				}
				break;
			case kVDShaderOp_dp4:
				{
					int t = nextDstReg++;
					int t2 = nextDstReg++;
					int t3 = nextDstReg++;
					int t4 = nextDstReg++;
					int t5 = nextDstReg++;
					int t6 = nextDstReg++;
					int t7 = 0;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i]) {
							if (!t7) {
								t7 = dinsn.dst[i];

								mul(t, dinsn.src[0], dinsn.src[4]);
								mul(t2, dinsn.src[1], dinsn.src[5]);
								add(t3, t, t2);
								mul(t4, dinsn.src[2], dinsn.src[6]);
								add(t5, t3, t4);
								mul(t6, dinsn.src[3], dinsn.src[7]);
								add(t7, t5, t6);
							} else
								mov(dinsn.dst[i], t7);
						}
					}
				}
				break;
			case kVDShaderOp_rsq:
				{
					int t = 0;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i]) {
							if (!t) {
								t = dinsn.dst[i];

								rsq(t, dinsn.src[0]);
							} else
								mov(dinsn.dst[i], t);
						}
					}
				}
				break;
			case kVDShaderOp_rcp:
				{
					int t = 0;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i]) {
							if (!t) {
								t = dinsn.dst[i];

								rcp(t, dinsn.src[0]);
							} else
								mov(dinsn.dst[i], t);
						}
					}
				}
				break;
			case kVDShaderOp_frc:
				{
					int t = 0;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i]) {
							if (!t) {
								t = dinsn.dst[i];

								frc(t, dinsn.src[0]);
							} else
								mov(dinsn.dst[i], t);
						}
					}
				}
				break;

			case kVDShaderOp_abs:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i])
						abs(dinsn.dst[i], dinsn.src[i]);
				}
				break;

			case kVDShaderOp_crs:
				// crs r0, r1, r2 becomes:
				//
				//		mul r0, r1.zxy, r2.yzx
				//		mad r0, r1.yzx, r2.zxy, -r0

				if (dinsn.dst[0]) {
					mul(nextDstReg+0, dinsn.src[1], dinsn.src[6]);
					mul(nextDstReg+1, dinsn.src[2], dinsn.src[5]);
					sub(dinsn.dst[0], nextDstReg+0, nextDstReg+1);
					nextDstReg += 2;
				}

				if (dinsn.dst[1]) {
					mul(nextDstReg+0, dinsn.src[2], dinsn.src[4]);
					mul(nextDstReg+1, dinsn.src[0], dinsn.src[6]);
					sub(dinsn.dst[1], nextDstReg+0, nextDstReg+1);
					nextDstReg += 2;
				}

				if (dinsn.dst[2]) {
					mul(nextDstReg+0, dinsn.src[0], dinsn.src[5]);
					mul(nextDstReg+1, dinsn.src[1], dinsn.src[4]);
					sub(dinsn.dst[2], nextDstReg+0, nextDstReg+1);
					nextDstReg += 2;
				}

				break;

			case kVDShaderOp_dp2add:
				{
					int t0 = nextDstReg++;
					int t1 = nextDstReg++;
					int t2 = nextDstReg++;
					int t3 = nextDstReg++;

					mul(t0, dinsn.src[0], dinsn.src[4]);
					mul(t1, dinsn.src[1], dinsn.src[5]);
					add(t2, t0, t1);
					add(t3, t2, dinsn.src[8]);

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i])
							mov(dinsn.dst[i], t3);
					}
				}
				break;

			case kVDShaderOp_log:
				{
					int t = 0;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i]) {
							if (!t) {
								t = dinsn.dst[i];

								log(t, dinsn.src[0]);
							} else
								mov(dinsn.dst[i], t);
						}
					}
				}
				break;

			case kVDShaderOp_exp:
				{
					int t = 0;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i]) {
							if (!t) {
								t = dinsn.dst[i];

								exp(t, dinsn.src[0]);
							} else
								mov(dinsn.dst[i], t);
						}
					}
				}
				break;

			case kVDShaderOp_texld:
				{
					int t = nextDstReg++;
					texld(t, dinsn.src[0], dinsn.src[1], insn.mSources[1].mRegister);

					if (dinsn.dst[0])
						unpackr(dinsn.dst[0], t);

					if (dinsn.dst[1])
						unpackg(dinsn.dst[1], t);

					if (dinsn.dst[2])
						unpackb(dinsn.dst[2], t);

					// For now, we unpack 1.0 as alpha instead of the actual alpha data.
					if (dinsn.dst[3])
						mov(dinsn.dst[3], cOne);
				}
				break;

			default:
				VDASSERT(false);
		}

		if (writeConflict) {
			for(int i=0; i<4; ++i) {
				if (writeMask & (1 << i)) {
					if (insn.mModifiers & kVDShaderInsnMod_Saturate) {
						max(nextDstReg, writeBase, cZero);
						min(dstRegIndex + i, nextDstReg, cOne);
						++nextDstReg;
					} else
						mov(dstRegIndex + i, writeBase);

					++writeBase;
				}
			}
		}
	}

	// update interpolators
	add(t0_x, t0_x, kVDShaderRegV0 + 1);
	add(t0_y, t0_y, kVDShaderRegV0 + 3);

	mShader.mTempCount = nextDstReg - kVDShaderRegR0;

	if (out && showDisassembly) {
		out->WriteLogOutput("//\n");
		out->WriteLogOutput("// Initial conversion result:\n");
		out->WriteLogOutput("//\n");

		Disassemble(*out, mShader);
	}

	if (!VDShaderScalarConvertToSSAForm(mShader, out))
		return false;

	if (out && showDisassembly) {
		out->WriteLogOutput("//\n");
		out->WriteLogOutput("// SSA conversion result:\n");
		out->WriteLogOutput("//\n");

		Disassemble(*out, mShader);
	}

	// bake sampler parameters
	VDShaderScalarBakeSamplers(mShader, pSamplers, numSamplers);

	if (out && showDisassembly) {
		out->WriteLogOutput("//\n");
		out->WriteLogOutput("// Bake samplers output:\n");
		out->WriteLogOutput("//\n");

		Disassemble(*out, mShader);
	}

	// bake interpolator steps
	if (interps) {
		VDShaderScalarBakeInterpolatorStepping(mShader, interps, interpCount, interpStepFactor);

		if (out && showDisassembly) {
			out->WriteLogOutput("//\n");
			out->WriteLogOutput("// Bake interpolator stepping output:\n");
			out->WriteLogOutput("//\n");

			Disassemble(*out, mShader);
		}
	}

	// apply copy propagation
	VDShaderScalarApplyCopyPropagation(mShader);

	if (out && showDisassembly) {
		out->WriteLogOutput("//\n");
		out->WriteLogOutput("// Copy propagation output:\n");
		out->WriteLogOutput("//\n");

		Disassemble(*out, mShader);
	}

	// apply constant folding
	do {
		VDShaderScalarApplyCopyPropagation(mShader);
	} while(VDShaderScalarFoldConstants(mShader));

	if (out && showDisassembly) {
		out->WriteLogOutput("//\n");
		out->WriteLogOutput("// Constant folding output:\n");
		out->WriteLogOutput("//\n");

		Disassemble(*out, mShader);
	}

	// apply dead store elimination
	VDShaderScalarApplyDeadStoreElimination(mShader);

	if (out && showDisassembly) {
		out->WriteLogOutput("//\n");
		out->WriteLogOutput("// Dead store elimination output:\n");
		out->WriteLogOutput("//\n");

		Disassemble(*out, mShader);
	}

	// hoist out loop invariants
	VDShaderScalarHoistLoopInvariants(mShader);

	if (out && showDisassembly) {
		out->WriteLogOutput("//\n");
		out->WriteLogOutput("// Loop invariant hoist output:\n");
		out->WriteLogOutput("//\n");

		Disassemble(*out, mShader);
	}

	// eliminate redundant and unused constants
	VDShaderScalarPackConstants(mShader);

	if (out && showDisassembly) {
		out->WriteLogOutput("//\n");
		out->WriteLogOutput("// Constant pack output:\n");
		out->WriteLogOutput("//\n");

		Disassemble(*out, mShader);
	}

	// pack temporary registers
	if (!VDShaderScalarPackTemps(mShader, out))
		return false;

	if (out && showDisassembly) {
		out->WriteLogOutput("//\n");
		out->WriteLogOutput("// Pack temps output:\n");
		out->WriteLogOutput("//\n");

		Disassemble(*out, mShader);
	}

	// assign uploads
	if (!VDShaderScalarBuildConstantUploadList(mShader, pConstants, numConstants, out))
		return false;

	if (out && showDisassembly) {
		out->WriteLogOutput("//\n");
		out->WriteLogOutput("// Generated by VirtualDub vector-to-scalar recompiler\n");
		out->WriteLogOutput("//\n");

		Disassemble(*out, mShader);
	}

	int numConstants = 0;
	int numTemps = 0;
	int numOutputs = 0;
	int numInputs = 0;

	for(VDShaderScalarForm::Insns::const_iterator it(mShader.mInsns.begin()), itEnd(mShader.mInsns.end()); it != itEnd; ++it) {
		const VDShaderInterpreterTransformedInsn& insn = *it;

		const int registers[] = { insn.dst, insn.src[0], insn.src[1], insn.src[2] };

		for(int i=0; i<sizeof(registers)/sizeof(registers[0]); ++i) {
			int r = registers[i];

			if (r >= 0) {
				int idx = r & kVDShaderRegIndexMask;

				switch(r & kVDShaderRegTypeMask) {
					case kVDShaderRegR0:
						if (idx >= numTemps)
							numTemps = idx + 1;
						break;

					case kVDShaderRegC0:
						if (idx >= numConstants)
							numConstants = idx + 1;
						break;

					case kVDShaderRegO0:
						if (idx >= numOutputs)
							numOutputs = idx + 1;
						break;

					case kVDShaderRegV0:
						if (idx >= numInputs)
							numInputs = idx + 1;
						break;
				}
			}
		}
	}

	mShader.mInputCount = numInputs;

	return true;
}

void VDShaderScalarTransformer::mov(int dst, int src) {
	push_insn(kVDShaderInterpOp_mov, dst, src, 0);
}

void VDShaderScalarTransformer::add(int dst, int src1, int src2) {
	push_insn(kVDShaderInterpOp_add, dst, src1, src2);
}

void VDShaderScalarTransformer::sub(int dst, int src1, int src2) {
	push_insn(kVDShaderInterpOp_sub, dst, src1, src2);
}

void VDShaderScalarTransformer::mul(int dst, int src1, int src2) {
	push_insn(kVDShaderInterpOp_mul, dst, src1, src2);
}

void VDShaderScalarTransformer::min(int dst, int src1, int src2) {
	push_insn(kVDShaderInterpOp_min, dst, src1, src2);
}

void VDShaderScalarTransformer::max(int dst, int src1, int src2) {
	push_insn(kVDShaderInterpOp_max, dst, src1, src2);
}

void VDShaderScalarTransformer::rsq(int dst, int src) {
	push_insn(kVDShaderInterpOp_rsq, dst, src, 0);
}

void VDShaderScalarTransformer::rcp(int dst, int src) {
	push_insn(kVDShaderInterpOp_rcp, dst, src, 0);
}

void VDShaderScalarTransformer::frc(int dst, int src) {
	push_insn(kVDShaderInterpOp_frc, dst, src, 0);
}

void VDShaderScalarTransformer::log(int dst, int src) {
	push_insn(kVDShaderInterpOp_log, dst, src, 0);
}

void VDShaderScalarTransformer::exp(int dst, int src) {
	push_insn(kVDShaderInterpOp_exp, dst, src, 0);
}

void VDShaderScalarTransformer::neg(int dst, int src) {
	push_insn(kVDShaderInterpOp_neg, dst, src, 0);
}

void VDShaderScalarTransformer::abs(int dst, int src) {
	push_insn(kVDShaderInterpOp_abs, dst, src, 0);
}

void VDShaderScalarTransformer::cmp(int dst, int src1, int src2, int src3) {
	push_insn(kVDShaderInterpOp_cmp, dst, src1, src2, src3);
}

void VDShaderScalarTransformer::texld(int dst, int srcu, int srcv, int srcsamp) {
	push_insn(kVDShaderInterpOp_texld, dst, srcu, srcv, srcsamp);
}

void VDShaderScalarTransformer::unpackr(int dst, int src) {
	push_insn(kVDShaderInterpOp_unpackr, dst, src, 0);
}

void VDShaderScalarTransformer::unpackg(int dst, int src) {
	push_insn(kVDShaderInterpOp_unpackg, dst, src, 0);
}

void VDShaderScalarTransformer::unpackb(int dst, int src) {
	push_insn(kVDShaderInterpOp_unpackb, dst, src, 0);
}

void VDShaderScalarTransformer::unpacka(int dst, int src) {
	push_insn(kVDShaderInterpOp_unpacka, dst, src, 0);
}

void VDShaderScalarTransformer::beginloop() {
	push_insn(kVDShaderInterpOp_beginloop, 0, 0, 0);
}

void VDShaderScalarTransformer::push_insn(int op, int dst, int src1, int src2, int src3) {
	VDShaderInterpreterTransformedInsn tinsn;
	tinsn.op = op;
	tinsn.dst = dst;
	tinsn.src[0] = src1;
	tinsn.src[1] = src2;
	tinsn.src[2] = src3;
	mShader.mInsns.push_back(tinsn);
}
